/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * xax500's "about xax500" dialog box.
 */

#include <Xm/PanedW.h>
#include <Xm/Label.h>
#include <Xm/Protocols.h>

#include <lber.h>
#include <ldap.h>

#include "dtypes.h"
#include "globals.h"
#include "util.h"
#include "help.h"

extern char *versionInfo[];
extern char *buildInfo[];
extern char *aboutInfo[];


/*
 * Destroy the about box.
 */
void aboutClose(w)
Widget	w;
{
    XtDestroyWidget(aboutDlg);
}


/*
 * Create an "about" box with version information, authors, copyright, etc.
 */
void
doAboutDlg()
{
    Widget			aboutPane, aboutPix, aboutDismissButton,
				aboutHelpButton, aboutActionArea, aboutLabel;
    static ActionAreaItem	aboutActions[2];
    Atom			wm_delete_window;
    XmString			s, str;
    int				i;

    aboutDlg = XtVaAppCreateShell("aboutDlg", "Xax500",
	    applicationShellWidgetClass,
	    XtDisplay(toplevel),
	    NULL);

    /*
     * Arrange for the "close" function on the WM menu to *not*
     * destroy the application's main window.
     */
    wm_delete_window = XmInternAtom(XtDisplay(aboutDlg), "WM_DELETE_WINDOW",
	    False);
    XmAddWMProtocolCallback(aboutDlg, wm_delete_window, aboutClose, NULL);

    aboutPane = XtVaCreateManagedWidget("aboutPane",
	    xmPanedWindowWidgetClass,
	    aboutDlg,
	    XmNsashWidth, 1,
	    XmNsashHeight, 1,
	    NULL);

    aboutPix = create_pixmap_button(aboutPane, "xax500_image", NULL,
	    "aboutPix");

    XtManageChild(aboutPix);

    /*
     * The version, "about", and build information are stored in
     * strings in version.c (generated from Version.c)
     */
    str = XmStringSegmentCreate("",
	    XmSTRING_DEFAULT_CHARSET,
	    XmSTRING_DIRECTION_L_TO_R, False);
    for (i = 0; versionInfo[i] != NULL; i++) {
	s = XmStringSegmentCreate(versionInfo[i], XmSTRING_DEFAULT_CHARSET,
		XmSTRING_DIRECTION_L_TO_R, True);
	str = XmStringConcat(str, s);
	XmStringFree(s);
    }
    for (i = 0; aboutInfo[i] != NULL; i++) {
	s = XmStringSegmentCreate(aboutInfo[i],
		XmSTRING_DEFAULT_CHARSET,
		XmSTRING_DIRECTION_L_TO_R, True);
	str = XmStringConcat(str, s);
	XmStringFree(s);
    }
    for (i = 0; buildInfo[i] != NULL; i++) {
	s = XmStringSegmentCreate(buildInfo[i],
		XmSTRING_DEFAULT_CHARSET,
		XmSTRING_DIRECTION_L_TO_R, True);
	str = XmStringConcat(str, s);
	XmStringFree(s);
    }
    aboutLabel = XtVaCreateManagedWidget("aboutLabel",
	    xmLabelWidgetClass, aboutPane,
	    XmNlabelString, str,
	    NULL);
    XmStringFree(str);

    /* Create an action area */
    aboutActions[0].label = "aboutDismissButton";
    aboutActions[0].data = (caddr_t) aboutDlg;
    aboutActions[0].widget = &aboutDismissButton;
    aboutActions[0].callback = aboutClose;
    aboutActions[1].label = "aboutHelpButton";
    aboutActions[1].data = (caddr_t) HLP_TOPIC_TOP;
    aboutActions[1].widget = &aboutHelpButton;
    aboutActions[1].callback = doHelpDlg;

    aboutActionArea = CreateActionArea(aboutPane, aboutActions, 2);
    XtRealizeWidget(aboutDlg);
    XtMapWidget(aboutDlg);

    /* Arrange for the about dialog to be non-resizable */
    XtAddEventHandler(aboutDlg, StructureNotifyMask, False,
	    (XtEventHandler) configureSize, NULL);
}
