/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Generic dialog support
 */

#include <stdio.h>

#include <Xm/SelectioB.h>
#include <Xm/MessageB.h>
#include <Xm/List.h>

#include "lber.h"
#include "ldap.h"
#include "dtypes.h"
#include "globals.h"
#include "dialog.h"
#include "errtxt.h"

/* Forward declarations */
#ifndef NEEDPROTOS
static int	modalYNCInternal();
#else /* !NEEDPROTOS */
static int	modalYNCInternal(Widget w, char *str, Boolean do_cancel);
#endif /* !NEEDPROTOS */

/*
 * Create a message dialog.  I know, 11 arguments is really gross.
 */
Widget
createDlg(w, type, dlgName, text, okt, okc,
	 cancelt, cancelc, helpt, helpc, modality)
Widget		w;
int		type;
char		*dlgName;
char		*text;
char		*okt;
void		(*okc)();
char		*cancelt;
void		(*cancelc)();
char		*helpt;
void		(*helpc)();
unsigned char	modality;
{
    Widget	dlg;
    XmString	dlgStr, okStr, cancelStr, helpStr;
    int		i = 0;
    Arg		args[10];

    /* Create the dialog */
    XtSetArg(args[i], XmNdialogType, type); i++;
    XtSetArg(args[i], XmNdialogStyle, modality); i++;
    dlg = XmCreateMessageDialog(w, dlgName, args, i);

    /* Alter it */
    if (text != NULL) {
	dlgStr = XmStringCreateSimple(text);
	XtVaSetValues(dlg, XmNmessageString, dlgStr, NULL);
	XmStringFree(dlgStr);
    }
    if (okt != NULL) {
	okStr = XmStringCreateSimple(okt);
	XtVaSetValues(dlg, XmNokLabelString, okStr, NULL);
	XmStringFree(okStr);
	if (okc != NULL) {
	    XtAddCallback(dlg, XmNokCallback, okc, NULL);
	}
    } else {
	XtUnmanageChild(XmMessageBoxGetChild(dlg, XmDIALOG_OK_BUTTON));
    }
    if (cancelt != NULL) {
	cancelStr = XmStringCreateSimple(cancelt);
	XtVaSetValues(dlg, XmNcancelLabelString, cancelStr, NULL);
	XmStringFree(cancelStr);
	if (cancelc != NULL) {
	    XtAddCallback(dlg, XmNcancelCallback, cancelc, NULL);
	}
    } else {
	XtUnmanageChild( XmMessageBoxGetChild(dlg, XmDIALOG_CANCEL_BUTTON));
    }
    if (helpt != NULL) {
	helpStr = XmStringCreateSimple(helpt);
	XtVaSetValues(dlg, XmNhelpLabelString, helpStr, NULL);
	XmStringFree(helpStr);
	if (helpc != NULL) {
	    XtAddCallback(dlg, XmNhelpCallback, helpc, NULL);
	}
    } else {
	XtUnmanageChild( XmMessageBoxGetChild(dlg, XmDIALOG_HELP_BUTTON));
    }
		
    return(dlg);
}




/*
 * Create an error dialog.
 */
Widget
errorDlg(w, text)
Widget	w;
char	*text;
{
    Widget	dlg;

    dlg = createDlg(w, XmDIALOG_ERROR, "errorDlg", text, "OK",
	    NULL, NULL, NULL, NULL, NULL,
	    XmDIALOG_FULL_APPLICATION_MODAL);
    XtManageChild(dlg);
    XtPopup(XtParent(dlg), XtGrabNone);
    return(dlg);
}


/*
 * Given an enumerated error, produce an error dialog.  XXX all routines
 * should call this to display errors, so non-English text can be
 * used.
 */
Widget
errorDlgEnum(w, id)
Widget	w;
int	id;
{
    extern errorTextStruct errorTexts[];

    if ((id < MIN_ERROR_ID) || (id >= MAX_ERROR_ID)) {
	return(errorDlg(w, "Error: illegal error message id"));
    } else {
	if (*(errorTexts[id].errorMsg) == NULL) {
	    return(errorDlg(w,
		"An error occurred, but there is no error message available."));
	} else {
	    return(errorDlg(w, *(errorTexts[id].errorMsg)));
	}
    }
}



/*
 * Create an info dialog.
 */
Widget
infoDlg(w, text)
Widget	w;
char	*text;
{
    Widget	dlg;

    dlg = createDlg(w, XmDIALOG_INFORMATION, "infoDlg", text, "OK",
	    NULL, NULL, NULL, NULL, NULL,
	    XmDIALOG_FULL_APPLICATION_MODAL);
    XtManageChild(dlg);
    XtPopup(XtParent(dlg), XtGrabNone);
    return(dlg);
}



void
updateText(w, text)
Widget	w;
char	*text;
{
    XmString	s;

    if (w == NULL) {
	return;
    }
    s = XmStringCreateLocalized(text);
    XtVaSetValues(w, XmNlabelString, s, NULL);
    XmStringFree(s);
}



static void
response(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    int *answer = (int *) client_data;
    XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct *) call_data;

    switch (cbs->reason) {
    case XmCR_OK:
	*answer = dlgYES;
	break;
    case XmCR_CANCEL:
	*answer = dlgNO;
	break;
    case XmCR_HELP:
	*answer = dlgCANCEL;
	break;
    default:
	return;
    }
}

int
modalYesNoCancel(w, str)
Widget	w;
char	*str;
{
    return(modalYNCInternal(w, str, True));
}


int
modalYesNo(w, str)
Widget	w;
char	*str;
{
    return(modalYNCInternal(w, str, False));
}




static int
modalYNCInternal(w, str, do_cancel)
Widget	w;
char	*str;
Boolean	do_cancel;
{
    Widget	dialog;
    XmString	text;
    static int	answer;

    dialog = XmCreateQuestionDialog(w, "yncDialog", NULL, 0);
    XtVaSetValues(dialog,
	    XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
	    NULL);
    if (do_cancel) {
	XtAddCallback(dialog, XmNhelpCallback, response, &answer);
    } else {
	XtUnmanageChild( XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    }
    XtAddCallback(dialog, XmNokCallback, response, &answer);
    XtAddCallback(dialog, XmNcancelCallback, response, &answer);
    answer = 0;
    text = XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(dialog, XmNmessageString, text, NULL);
    XmStringFree(text);
    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);

    while (answer == 0) {
	XtAppProcessEvent(app, XtIMAll);
    }
    XtDestroyWidget(dialog);
    return(answer);
}



static void
list_response(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    char				**value = (char **) client_data;
    XmSelectionBoxCallbackStruct	*cbs; 

    cbs = (XmSelectionBoxCallbackStruct *) call_data;

    if (cbs->reason == XmCR_CANCEL) {
	*value = "";
    } else {
	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, value);
    }
}
	


int
modalSelectionDlg(w, name, vals)
Widget		w;
char		*name;	/* resource name of widget */
char		**vals;	/* null-terminated array of choices */
{
    int		i, item;
    char	*value;
    Widget	dlg, list;
    XmString	xms;

    dlg = XmCreateSelectionDialog(w, name, NULL, 0);
    XtVaSetValues(dlg,
	    XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
	    NULL);
    XtUnmanageChild(XmSelectionBoxGetChild(dlg, XmDIALOG_SELECTION_LABEL));
    XtUnmanageChild(XmSelectionBoxGetChild(dlg, XmDIALOG_TEXT));
    XtUnmanageChild(XmSelectionBoxGetChild(dlg, XmDIALOG_APPLY_BUTTON));
    list = XmSelectionBoxGetChild(dlg, XmDIALOG_LIST);
	
    if (vals != NULL) {
	for (i = 0; vals[i] != NULL; i++) {
	    xms = XmStringCreateLocalized(vals[i]);
	    XmListAddItem(list, xms, 0);
	    XmStringFree(xms);
	}
    }
    XmListSelectPos(list, 1, True);
    XtAddCallback(dlg, XmNokCallback, list_response, &value);
    XtAddCallback(dlg, XmNcancelCallback, list_response, &value);
    value = NULL;
    XtManageChild(dlg);
    XtPopup(XtParent(dlg), XtGrabNone);

    while (value == NULL) {
	XtAppProcessEvent(app, XtIMAll);
    }
    if (strlen(value) == 0) {
	XtFree(value);
	return(-1);
    }
    item = -1;
    for (i = 0; vals[i] != NULL; i++) {
	if (!strcmp(value, vals[i])) {
	    item = i;
	    break;
	}
    }
    XtFree(value);
    XtDestroyWidget(dlg);
    return(item);
}
	

static void
prompt_response(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    char				**value = (char **) client_data;
    XmSelectionBoxCallbackStruct	*cbs; 

    cbs = (XmSelectionBoxCallbackStruct *) call_data;

    if (cbs->reason == XmCR_CANCEL) {
	*value = "";
    } else {
	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, value);
    }
}


char *
modalPromptDialog(w, name, val, help_cb, help_cbval)
Widget		w;
char		*name;
char		*val;
XtPointer	help_cb;
XtPointer	help_cbval;
{
    Widget	dlg, text;
    char	*value;

    dlg =  XmCreatePromptDialog(w, name, NULL, 0);
    XtVaSetValues(dlg,
	    XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
	    NULL);
    text = XmSelectionBoxGetChild(dlg, XmDIALOG_TEXT);
    if (val != NULL) {
	XtVaSetValues(text, XmNvalue, val, NULL);
    }
    XtAddCallback(dlg, XmNokCallback, prompt_response, &value);
    XtAddCallback(dlg, XmNcancelCallback, prompt_response, &value);
    if (help_cb != NULL) {
	XtAddCallback(dlg, XmNhelpCallback, help_cb, help_cbval);
    } else {
	XtUnmanageChild(XmSelectionBoxGetChild(dlg, XmDIALOG_HELP_BUTTON));
    }
    value = NULL;
    XtManageChild(dlg);
    XtPopup(XtParent(dlg), XtGrabNone);

    while (value == NULL) {
	XtAppProcessEvent(app, XtIMAll);
    }

    XtDestroyWidget(dlg);
    return(value);
}


void
destroyWorking(wp)
Widget	*wp;
{
    if (*wp != NULL && XtIsRealized(*wp)) {
	XtDestroyWidget(*wp);
	*wp = NULL;
    }
}




Widget
create_working_dialog(parent, text, ok_text, cancel_text, help_text,
		      ok_cb, cancel_cb, help_cb, cb_data, user_data)
Widget		parent;
char		*text;
char		*ok_text;
char		*cancel_text;
char		*help_text;
XtCallbackProc	ok_cb;
XtCallbackProc	cancel_cb;
XtCallbackProc	help_cb;
XtPointer	cb_data;
XtPointer	user_data;
{
    XmString	Dtext, Dok_text, Dcancel_text, Dhelp_text;
    Widget	workingDlg;
    Arg		args[10];
    int		i;

    i = 0;
    if (text != NULL) {
	Dtext = XmStringCreateSimple(text);
	XtSetArg(args[i], XmNmessageString, Dtext); i++;
    }
    if (user_data != NULL) {
	XtSetArg(args[i], XmNuserData, user_data); i++;
    }
    XtSetArg(args[i], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); i++;
    workingDlg = XmCreateWorkingDialog(parent, "workingDlg", args, i);

    /* XXX - resource */
    XtVaSetValues(XtParent(workingDlg), XmNtitle, "Working...", NULL);

    /* OK button (or not) */
    if (ok_text != NULL) {
	Dok_text = XmStringCreateSimple(ok_text);
	XtSetArg(args[i], XmNokLabelString, Dok_text); i++;
	if (ok_cb != NULL) {
	    XtAddCallback(workingDlg, XmNokCallback, ok_cb, cb_data);
	}
    } else {
	XtUnmanageChild(XmMessageBoxGetChild(workingDlg, XmDIALOG_OK_BUTTON));
    }
		
    /* Cancel button (or not) */
    if (cancel_text != NULL) {
	Dcancel_text = XmStringCreateSimple(cancel_text);
	XtSetArg(args[i], XmNcancelLabelString, Dcancel_text); i++;
	if (cancel_cb != NULL) {
	    XtAddCallback(workingDlg, XmNcancelCallback, cancel_cb, cb_data);
	}
    } else {
	XtUnmanageChild(XmMessageBoxGetChild(workingDlg, 
		XmDIALOG_CANCEL_BUTTON));
    }

    /* Help button (or not) */
    if (help_text != NULL) {
	Dhelp_text = XmStringCreateSimple(help_text);
	XtSetArg(args[i], XmNhelpLabelString, Dhelp_text); i++;
	if (help_cb != NULL) {
	    XtAddCallback(workingDlg, XmNhelpCallback, help_cb, cb_data);
	}
    } else {
	XtUnmanageChild(XmMessageBoxGetChild(workingDlg, XmDIALOG_HELP_BUTTON));
    }

    if (ok_text == NULL && cancel_text == NULL && help_text == NULL) {
	/* No separator */
	/* XXX how do I disable the separator? */
    }
		

    XtSetValues(workingDlg, args, i);

    if (text != NULL) {
	XmStringFree(Dtext);
    }
    if (ok_text != NULL) {
	XmStringFree(Dok_text);
    }
    if (cancel_text != NULL) {
	XmStringFree(Dcancel_text);
    }
    if (help_text != NULL) {
	XmStringFree(Dhelp_text);
    }

    XtManageChild(workingDlg);
    XtPopup(XtParent(workingDlg), XtGrabNone);
    XmUpdateDisplay(workingDlg);
    return(workingDlg);
}
