/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */


/*
 * Display entries found in do_read()
 */

#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleBG.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>
#include <Xm/PanedW.h>
#include <Xm/SeparatoG.h>
#include <Xm/ArrowB.h>
#include <Xm/Frame.h>
#include <Xm/Protocols.h>
#include <Xm/DragDrop.h>
#include <Xm/DrawnB.h>
#include <Xm/AtomMgr.h>

#include <lber.h>
#include <ldap.h>

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>
#include <time.h>
#include <malloc.h>
#include <unistd.h>
#include <string.h>

#include "disptmpl.h"
#include "dtypes.h"
#include "util.h"
#include "menu.h"
#include "globals.h"
#include "display.h"
#include "syn.h"
#include "bitmaps.h"
#include "dialog.h"
#include "browse.h"
#include "write.h"
#include "search.h"
#include "edit.h"
#include "find.h"
#include "errtxt.h"
#include "read.h"
#include "help.h"

/* forward references */
#ifndef NEEDPROTOS
void		displayNewDisplayAs();
void		freeDisplayData();
void		freeEntryData();
void		freeDisplayAttrDataList();
void		updateDisplayDlgSaveButton();
static void	displaySaveCB();
static void	displayClose();
static Boolean	displayConvertProc();
static void	handleLabelDrop();
static void	transferDrop();
static void	doAddDN2();
static void	checkAddDN();
#else /* !NEEDPROTOS */
void		displayNewDisplayAs();
void		freeDisplayData(displayData *dd);
void		freeEntryData(entryData *en);
void		freeDisplayAttrDataList(displayAttrData *dlist);
void		updateDisplayDlgSaveButton(Widget w, int val);
static void	displaySaveCB(Widget w, XtPointer client_data,
			XtPointer call_data);
static void	displayClose(Widget w, displayData *d,
			XmPushButtonCallbackStruct *cbs);
static Boolean	displayConvertProc(Widget widget, Atom *selection,
			Atom *target, Atom *type_return,
			XtPointer *value_return, unsigned long *length_return,
			int *format_return);
static void	handleLabelDrop(Widget w, XtPointer client_data,
			XtPointer call_data);
static void	transferDrop(Widget widget, XtPointer client_data,
			Atom *seltype, Atom *type, XtPointer value,
			unsigned long length, int format);
static void	doAddDN2(displayAttrData *dad);
static void	checkAddDN(int rc, int ldaprc, actionData *actd);
#endif /* !NEEDPROTOS */

/* Globals */
displayData	*displayDlgData;	/* linked list of on-screen entries */
displayData	*latestDisplayDlg;	/* most-recently opened entry */

static ActionAreaItem   displayActions[] = {
    { "displayDismissButton",  NULL, NULL, NULL },
    { "displaySaveButton",  NULL, NULL, NULL },
    { "displayHelpButton",  NULL, NULL, NULL },
};



/*
 * Fire up the WWW viewer.  Actually, this code is not very generic, as
 * it knows only how to start up NCSA Mosaic and cause it to open a
 * specific URL.
 */
void
openURL(w, lurl)
Widget	w;
char	*lurl;
{
    char		pidfile[64];
    char		*l2, *url, *label;
    FILE		*fp;
    int			rc;
    struct forkedApps	*fa;
    displayData		*di;

    if (lurl == NULL) {
	return;
    }

    l2 = strdup(lurl);	/* Copy to avoid changing original */
    split_labeled_url(l2, &label, &url);

startover:
    if (!WWWViewerActive) {
	/* Fork the WWW browser */
	switch (WWWpid = fork()) {
	case 0:
	    rc = execlp(appConfig.wwwBrowser, appConfig.wwwBrowser,
		    "-home", url, 0);	
	    if (rc < 0) {
		perror("execlp failed");
		exit(-1);
	    }
	case -1:
	    fprintf(stderr, "fork failed\n");
	    free(l2);
	    return;
	default:
	    WWWViewerActive = 1;
	    di = getDisplayDataByWidget(w);
	    if (di) {
		timedStatusUpdate(di->di_status, "Invoking Web Browser...", 3);
	    }
	    /* 
	     * Add info about this forked process to the list
	     * of forked apps, so we can clean up the files
	     * later.
	     */
	    fa = (struct forkedApps *)
		    malloc(sizeof(struct forkedApps));
	    if (fa == NULL) {
		    fprintf(stderr, "no memory for forkedApps\n");
		    exit(1);
	    }
	    fa->fa_next = forkedAppsList;
	    forkedAppsList = fa;
	    sprintf(pidfile, "/tmp/Mosaic.%d", WWWpid);
	    fa->fa_pid = WWWpid;
	    fa->fa_files = (char **) malloc(2 * sizeof(char *));
	    fa->fa_files[0] = strdup(pidfile);
	    fa->fa_files[1] = NULL;
	    free(l2);
	    return;
	}
    }

    /*
     * Mosaic was active, so set up a file with a command in it, and
     * send Mosaic a USR1 signal.  It will read the file and execute
     * the commands in it.
     */
    sprintf(pidfile, "/tmp/Mosaic.%d", WWWpid);
    if ((fp = fopen(pidfile, "w")) == NULL) {
	errorDlg(w, "Cannot write WWW pid file");
	free(l2);
	return;
    } else {
	fprintf(fp, "goto\n%s\n", url);
	fclose(fp);
	/*
	 * XXX - need to *not* send a signal if we just started Mosaic.
	 * Since its signal handlers aren't set up, we'll end up killing it.
	 */
	if (kill(WWWpid, SIGUSR1) < 0) {
	    if (errno == ESRCH) {
		/* Process went away */
		WWWViewerActive = 0;
		unlink(pidfile);
		goto startover;
	    } else {
		errorDlg(w, "Cannot send signal to WWW viewer");
	    }
	}
    }
    free(l2);
}


void
invokeWWWViewer(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmPushButtonCallbackStruct	*cbs;
{
    openURL(w, (char *) client_data);
}



/*
 * Play audio data.
 */
void
invokeAudioViewer(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmPushButtonCallbackStruct	*cbs;
{
    char		template[32];
    int			i, fd, rc, argc;
    pid_t		pid;
    char		*argv[24];	/* XXX max 24 values */
    struct berval	**vals = (struct berval **) client_data;
    struct forkedApps	*fa;

    /* XXX add support for deferred reading of values */

    if (vals == NULL) {
	return;
    }
    for (i = 0; vals[i] != NULL; i++) {
	strcpy(template, TMPFILE_TEMPLATE);
	fd = mkstemp(template);
	write(fd, vals[i]->bv_val, vals[i]->bv_len);
	close(fd);
	argv[i + 1] = strdup(template);
    }
    argc = i + 1;
    argv[argc] = NULL;

    switch (pid = fork()) {
    case 0:
	argv[0] = appConfig.audioViewer;
	rc = execvp(appConfig.audioViewer, argv);
	if (rc < 0) {
	    perror("execv failed");
	    exit(-1);
	}
    case -1:
	fprintf(stderr, "fork failed\n");
	return;
    default:
	/* 
	 * Add info about this forked process to the list
	 * of forked apps, so we can clean up the files
	 * later.
	 */
	fa = (struct forkedApps *)
		malloc(sizeof(struct forkedApps));
	if (fa == NULL) {
	    fprintf(stderr, "no memory for forkedApps\n");
	    exit(1);
	}
	fa->fa_next = forkedAppsList;
	forkedAppsList = fa;
	fa->fa_pid = pid;
	fa->fa_files = (char **) malloc(argc * sizeof(char *));
	for (i = 0; i < argc; i++) {
	    fa->fa_files[i] = argv[i + 1];
	}
	return;
    }	
}




/*
 * Invoke a program which knows how to display JPEG images.
 * XXX makes temp files and invokes viewer on all files at once.
 * Should do something more general.
 */
void
invokeJpegViewer(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmPushButtonCallbackStruct	*cbs;
{
    char		template[32];
    int			i, fd, rc, argc;
    pid_t		pid;
    char		*argv[24];	/* XXX max 24 values */
    struct berval	**vals = (struct berval **) client_data;
    struct forkedApps	*fa;

    if (vals == NULL) {
	return;
    }
    for (i = 0; vals[i] != NULL; i++) {
	strcpy(template, TMPFILE_TEMPLATE);
	fd = mkstemp(template);
	write(fd, vals[i]->bv_val, vals[i]->bv_len);
	close(fd);
	argv[i + 1] = strdup(template);
    }
    argc = i + 1;
    argv[argc] = NULL;

    switch (pid = fork()) {
    case 0:
	argv[0] = appConfig.jpegViewer;
	rc = execvp(appConfig.jpegViewer, argv);
	if (rc < 0) {
	    perror("execv failed");
	    exit(-1);
	}
    case -1:
	fprintf(stderr, "fork failed\n");
	return;
    default:
	fa = (struct forkedApps *) malloc(sizeof(struct forkedApps));
	if (fa == NULL) {
	    fprintf(stderr, "no memory for forkedApps\n");
	    exit(1);
	}
	/* 
	 * Add info about this forked process to the list
	 * of forked apps, so we can clean up the files
	 * later.
	 */
	fa->fa_next = forkedAppsList;
	forkedAppsList = fa;
	fa->fa_pid = pid;
	fa->fa_files = (char **) malloc(argc * sizeof(char *));
	for (i = 0; i < argc; i++) {
	    fa->fa_files[i] = argv[i + 1];
	}
	return;
    }	
}
	



/*
 * Called when user clicks on a DN button - causes a new entry window
 * to be opened for the given DN.
 */
void
openDN(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmPushButtonCallbackStruct	*cbs;
{
    char	*dn = (char *) client_data;

    do_read(w, ldp, dn, NULL, NULL);
}


/*
 * Handle a click on an searchAction button, e.g. a "Subscribers" button
 *
 */
void
doSearchAction(w, dad, cbs)
Widget				w;
displayAttrData			*dad;
XmPushButtonCallbackStruct	*cbs;
{
    actionData		*actd;
    searchactData	*actargs;
    int			i;
    char 		*dattrs[2];
    

    /* Sanity-check the arguments from the template structures */
    /* Arguments we expect:
     * 0 - value to search for. Mandatory.
     * 1 - Search filter to use.  Mandatory.
     * 2 - Disambiguation attribute for searching.  Mandatory.
     * 3 - Label for disambiguation attribute.  Mandatory.
     * 4 - Disable attribute.  Optional.
     */ 

    actd = (actionData *) malloc(sizeof(actionData));
    actd->ac_type = ACTION_TYPE_SEARCHACT;
    actd->ac_dd = getDisplayDataByWidget(dad->dad_valueWidget);
    if (actd->ac_dd != NULL) {
	w = actd->ac_dd->di_displayDlg;
    }
    for (i = 0; i <= 2; i++) {
	if (dad->dad_ti->ti_args[i] == NULL) {
	    errorDlgEnum(w, E_BADADDDNARGS);
	    return;
	}
    }
    if (dad->dad_ti->ti_args[SP_ADIS] != NULL) {
	actd->ac_disableattr = strdup(dad->dad_ti->ti_args[AP_ADIS]);
    } else {
	actd->ac_disableattr = NULL;
    }
    actargs = (searchactData *) ch_malloc(sizeof(searchactData));
    actd->ac_args = actargs;

    /* XXX - need to figure out a way to handle -dns escapes */
    actargs->search_value = convertDNEscape(dad->dad_ti->ti_args[SP_AVAL],
	    dad->dad_entryData->en_dn);
    actargs->search_filter =
	    convertMessageDNEscape(dad->dad_ti->ti_args[SP_AFIL],
	    dad->dad_entryData->en_dn);
    actargs->search_disambig = strdup(dad->dad_ti->ti_args[SP_ADAMB]);
    actargs->search_disambiglabel =
	    convertMessageDNEscape(dad->dad_ti->ti_args[SP_ADAMBL],
	    dad->dad_entryData->en_dn);
    /* Perform a search */
    dattrs[0] = actargs->search_disambig;
    dattrs[1] = NULL;
    /* XXX - actargs->search_disambiglabel ia not used */
    do_search(w, actargs->search_value, NULL, dattrs,
	    actargs->search_filter, NULL, NULL, do_read,
	    "selectDlg", SRCH_OPT_LITERALFILT | SRCH_OPT_PERSIST, NULL);
}


/*
 * Handle a click on an action button, e.g. a "Join..." button
 *
 */
void
doAddDN(w, dad, cbs)
Widget				w;
displayAttrData			*dad;
XmPushButtonCallbackStruct	*cbs;
{
    if (bound_as[0] == '\0') {
	/* not bound - don't know who to bind as */
	start_auth(w, doAddDN2, dad, 1);
    } else {
	doAddDN2(dad);
    }
}



static void
doAddDN2(dad)
displayAttrData		*dad;
{
    int			resp;
    int			i;
    Widget		w;
    actionData		*actd;
    adddnactData	*actargs;
    

    /* Sanity-check the arguments from the template structures */
    /* Arguments we expect:
     * 0 - Value to add.  "-dnb" = DN we're bound as, "-dnt" = DN of
     *     entry containing the attribute to be written, "-dns" =
     *     DN found in a search.
     * 1 - DN of destination.  Mandatory.
     * 2 - destination attribute.  Mandatory.
     * 3 - disable attribute - name of boolean attr in this entry - if
     *     false, the action button should be disabled.  Optional.
     * 4 - Error message to display if value exists in destination attribute.
     *     Optional.
     * 5 - Confirmation message to display before doing addition. Optional.
     *
     */ 

    actd = (actionData *) malloc(sizeof(actionData));
    actd->ac_type = ACTION_TYPE_ADDDNACT;
    actd->ac_dd = getDisplayDataByWidget(dad->dad_valueWidget);
    if (actd->ac_dd != NULL) {
	w = actd->ac_dd->di_displayDlg;
    }
    for (i = 0; i <= 5; i++) {
	if (dad->dad_ti->ti_args[i] == NULL) {
	    errorDlgEnum(w, E_BADADDDNARGS);
	    return;
	}
    }
    if (dad->dad_ti->ti_args[AP_ADIS] != NULL) {
	actd->ac_disableattr = strdup(dad->dad_ti->ti_args[AP_ADIS]);
    } else {
	actd->ac_disableattr = NULL;
    }
    actargs = (adddnactData *) ch_malloc(sizeof(adddnactData));
    actd->ac_args = actargs;

    /* XXX - need to figure out a way to handle -dns escapes */
    actargs->adddn_value = convertDNEscape(dad->dad_ti->ti_args[AP_AVAL],
	    dad->dad_entryData->en_dn);
    actargs->adddn_dstdn = convertDNEscape(dad->dad_ti->ti_args[AP_ADDN],
	    dad->dad_entryData->en_dn);
    actargs->adddn_dstattr =
	    convertMessageDNEscape(dad->dad_ti->ti_args[AP_ADAT],
	    dad->dad_entryData->en_dn);
    actargs->adddn_existsmsg =
	    convertMessageDNEscape(dad->dad_ti->ti_args[AP_AEERR],
	    dad->dad_entryData->en_dn);
    actargs->adddn_confmsg =
	    convertMessageDNEscape(dad->dad_ti->ti_args[AP_ACMG],
	    dad->dad_entryData->en_dn);
    if (strlen(actargs->adddn_confmsg) > 0) {
	resp = modalYesNo(w, actargs->adddn_confmsg);
	if (resp != dlgYES) {
	    free_actiondata(actd);
	    return;
    }
    /* Modify the appropriate DN's attribute. */
    addValueToAttribute(w, actargs->adddn_value, actargs->adddn_dstdn,
	    actargs->adddn_dstattr, checkAddDN, actd);
    }
}




/*
 * Check the return code from the AddDN operation.  Generally, error
 * dialogs are created by the code in write.c, so we probably don't need
 * to do anything here.  XXX extend this to handle returns from other
 * action routines.
 */
static void
checkAddDN(rc, ldaprc, actd)
int		rc;
int		ldaprc;
actionData	*actd;
{
    char	errstr[1024];

    switch (rc) {
    case WR_CANCELLED:
	infoDlg(actd->ac_dd->di_displayDlg, "Cancelled.");
	break;
    case WR_ERR_TIMEOUT:
	errorDlg(actd->ac_dd->di_displayDlg, "Operation timed out");
	break;
    case WR_LDAP_ERROR:
	if (actd->ac_type == ACTION_TYPE_ADDDNACT) {
	switch (ldaprc) {
	    case LDAP_TYPE_OR_VALUE_EXISTS: {
		adddnactData *args = actd->ac_args;
		strcpy(errstr, args->adddn_existsmsg);
		break;
	    }
	    default:
		strcat(errstr, ldap_err2string(ldaprc));
	    }
	} else {
	    strcat(errstr, ldap_err2string(ldaprc));
	}
	errorDlg(actd->ac_dd->di_displayDlg, errstr);
	break;
    case WR_OK:
    default:
    }
}





/*
 * Create a form widget holding an arrow button and a label.
 *
 */
Widget
createURLWidget(parent, dad, lurl)
Widget		parent;
displayAttrData	*dad;
char		*lurl;
{
    Widget	URLForm, URLButton, URLLabel;
    char	*url, *label, *l2;
    XmString	tmp;
    valueData	*vd;
    attrData	*ad = dad->dad_attrData;
    displayData	*dd;

    /*
     * Save stuff we'll need for drag operations.  get_valuedata() places
     * vd on a linked list to be freed when the entry is closed.
     */
    dd = getDisplayDataByWidget(parent);
    vd = get_valuedata(dd);
    vd->vd_tag = VD_URL;
    vd->vd_displayAttrData = dad;
    vd->vd_displayData = dd;
    vd->vd_value = strdup(lurl);

    if (dad->dad_numValueWidgets > ad->ad_nvals) {
	/* Have extra cached, re-use */
	URLForm = dad->dad_valueWidgetList[ad->ad_nvals];
	URLButton = getArrowButtonChild(URLForm);
	URLLabel = getLabelGadgetChild(URLForm);
    } else {
	dad->dad_numValueWidgets++;

	/* Create a form widget to hold the button and label */
	URLForm = XtVaCreateManagedWidget("URLForm",
		xmFormWidgetClass, parent,
		XmNverticalSpacing, 0,
		XmNhorizontalSpacing, 0,
		XmNtranslations, parsedNoDragTranslations,
		NULL);

	/* Create the button widget */
	URLButton = XtVaCreateManagedWidget("URLButton",
		xmArrowButtonWidgetClass, URLForm,
		XmNleftAttachment, XmATTACH_FORM,
		XmNtranslations, parsedDragTranslations,
		NULL);

	/* Create the label */
	URLLabel = XtVaCreateManagedWidget("URLLabel",
		xmLabelGadgetClass, URLForm,
		XmNalignment, XmALIGNMENT_BEGINNING,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, URLButton,
		NULL);
    }

    /* Set the label's value */
    l2 = strdup(lurl);
    split_labeled_url(l2, &label, &url);
    tmp = XmStringCreateLocalized(label);
    XtVaSetValues(URLLabel, XmNlabelString, tmp, NULL);
    free(l2);
    XmStringFree(tmp);

    /*
     * We store a pointer to the attrData struct in the XmNuserData
     * resource of the form, and pointer to the URL in the userData
     * resource of the label.
     */
    XtAddCallback(URLButton, XmNactivateCallback, 
	(XtCallbackProc) invokeWWWViewer, vd->vd_value);
    XtVaSetValues(URLForm, XmNuserData, dad, NULL);
    XtVaSetValues(URLLabel, XmNuserData, vd->vd_value, NULL);
    XtVaSetValues(URLButton, XmNuserData, vd, NULL);

    XtManageChild(URLForm);

    return(URLForm);
}




/*
 * Create a form widget holding an arrow button and a label.
 *
 */
Widget
createDNWidget(parent, dad, dn)
Widget		parent;
displayAttrData	*dad;
char		*dn;
{
    Widget	DNForm, DNButton, DNLabel;
    char	**xdn;
    XmString	tmp;
    valueData	*vd;
    attrData	*ad = dad->dad_attrData;
    displayData	*dd;

    /*
     * Save stuff we'll need for drag operations.
     */
    dd = getDisplayDataByWidget(parent);
    vd = get_valuedata(dd);
    vd->vd_tag = VD_DN;
    vd->vd_displayAttrData = dad;
    vd->vd_displayData = getDisplayDataByWidget(parent);
    vd->vd_value = strdup(dn);


    if (dad->dad_numValueWidgets > ad->ad_nvals) {
	/* Have extra cached, re-use */
	DNForm = dad->dad_valueWidgetList[ad->ad_nvals];
	DNButton = getArrowButtonChild(DNForm);
	DNLabel = getLabelGadgetChild(DNForm);
    } else {
	dad->dad_numValueWidgets++;

	/* Create a form widget to hold the button and label */
	DNForm = XtVaCreateWidget("DNForm",
		xmFormWidgetClass, parent,
		XmNverticalSpacing, 0,
		XmNhorizontalSpacing, 0,
		XmNtranslations, parsedNoDragTranslations,
		NULL);

	/* Create the button widget */
	DNButton = XtVaCreateManagedWidget("DNButton",
		xmArrowButtonWidgetClass, DNForm,
		XmNleftAttachment, XmATTACH_FORM,
		XmNtranslations, parsedDragTranslations,
		NULL);

	/* Create the label */
	DNLabel = XtVaCreateManagedWidget("DNLabel",
		xmLabelGadgetClass, DNForm,
		XmNalignment, XmALIGNMENT_BEGINNING,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, DNButton,
		NULL);
    }

    /* Set the label's value */
    xdn = ldap_explode_dn(dn, 1);
    tmp = XmStringCreateLocalized(xdn[0]);
    XtVaSetValues(DNLabel, XmNlabelString, tmp, NULL);
    XmStringFree(tmp);
    ldap_value_free(xdn);

    /*
     * We store a pointer to the attrData struct in the XmNuserData
     * resource of the form, and pointer to the DN in the userData
     * resource of the label.
     */
    XtAddCallback(DNButton, XmNactivateCallback, openDN, vd->vd_value);
    XtVaSetValues(DNForm, XmNuserData, dad, NULL);
    XtVaSetValues(DNLabel, XmNuserData, vd->vd_value, NULL);
    XtVaSetValues(DNButton, XmNuserData, vd, NULL);

    XtManageChild(DNForm);

    return(DNForm);
}





/*
 * Find the "Save" button associated with the entry window containing
 * widget "w" and enable or disable it.
 *
 * If val == -1, then the attribute data for the window is scanned
 * to see if any values have been modified, and if so, the Save button
 * is enabled.  This is the normal way to call this function.
 *
 * If val == 0, then the Save button is unconditionally disabled.
 * If val is any other value, the Save button is unconditionally enabled.
 *
 * Also sets the di_numMods * field in the displayInfo struct to 
 * reflect the number of modified attributes.
 *
 */
void
updateDisplayDlgSaveButton(w, val)
Widget	w;
int	val;
{
    displayData		*di;
    displayAttrData	*dad;
    attrData		*ad;
    Boolean		enabled;

    di = getDisplayDataByWidget(w);

    if (di == NULL) {
	errorDlg(toplevel, "Can't find entry window for that item");
	return;
    }

    if (val != -1) {
	enabled = val;
    } else if (di->di_entryData->en_isnew) {
	    enabled = True;
    } else {
	di->di_numMods = 0;
	enabled = False;
	for (dad = di->di_displayAttrData; dad != NULL; dad = dad->dad_next) {
	    ad = dad->dad_attrData;
	    if (dad->dad_attrData->ad_flags & AD_DIRTY) {
#ifdef notdef
		if ((dad->dad_attrData->dad_flags & AD_STARTED_EMPTY) &&
			((ad->ad_vals == NULL) ||
			(*(ad->ad_vals) == NULL))) {
		    dad->dad_attrData->dad_flags &= ~AD_STARTED_EMPTY;
		    continue;
		}
#endif
		enabled = True;
		di->di_numMods++;
	    }
	}
    }
    di->di_entryData->en_modified = enabled;
    XtSetSensitive(di->di_displaySaveButton, enabled);
}




/*
 * Make cursor visible or invisible, as appropriate.
 * Since we're called when the user either clicks in a different
 * text field, or moves the cursor out of the entry window, we
 * take this opportunity to see if the user changed any values in
 * the entry window, and update the "dirty" flags and "Save" button.
 */
void
displayTextFocus(w, dad, cbs)
Widget			w;
displayAttrData		*dad;
XmAnyCallbackStruct	*cbs;
{
    char		*text;
    struct berval	**bvs;
    int			rc;
    displayData		*dd;
    attrData		*ad;

    dd = getDisplayDataByWidget(w);

    if (cbs->reason == XmCR_FOCUS) {
	XtVaSetValues(w, XmNcursorPositionVisible, True, NULL);
	dd->di_currAttr = dad;
	if (dd->di_currAttr != dd->di_prevAttr) {
	    if ((dd->di_prevWidget == NULL) || (!XmIsText(dd->di_prevWidget))) {
		return;
	    }
	    /* XXX should be verifying the osval or the berval??? */
	    text = XmTextGetString(dd->di_prevWidget);
	    ad = dd->di_prevAttr->dad_attrData;
	    if ((ad->ad_syndata->sd_vrfyvalue != NULL) &&
		    ((rc = ad->ad_syndata->sd_vrfyvalue(text)) != 0)) {
		    errorDlgEnum(GetTopShell(w), rc);
	    }
	    XtFree(text);
	    dd->di_prevAttr = dd->di_currAttr;
	}
	    
    } else if (cbs->reason == XmCR_LOSING_FOCUS) {
	dd->di_prevAttr = dad;
	dd->di_prevWidget = w;
	XtVaSetValues(w, XmNcursorPositionVisible, False, NULL);
	/* If not touched, we're done */
	if (!(dad->dad_flags & DAD_MAYBE_CHANGED)) {
	    return;
	}
	text = XmTextGetString(w);
	bvs = dad->dad_attrData->ad_syndata->sd_fromnative(text);
	/* Clear "maybe touched" flag */
	dad->dad_flags &= ~DAD_MAYBE_CHANGED;
	if (bervec_compare(bvs, dad->dad_attrData->ad_vals)) {
	    /* Value unchanged, clear dirty bit */
	    dad->dad_attrData->ad_flags &= ~AD_DIRTY;
	    XtFree(text);
	} else {
	    /* Value changed, set dirty bit */
	    dad->dad_attrData->ad_flags |= AD_DIRTY;
	    XtFree(dad->dad_osval);
	    dad->dad_osval = text;
	}
	if (bvs != NULL) {
	    ber_bvecfree(bvs);
	}
	updateDisplayDlgSaveButton(w, -1);
    } else if (cbs->reason == XmCR_VALUE_CHANGED) {
	updateDisplayDlgSaveButton(w, True);
	dad->dad_flags |= DAD_MAYBE_CHANGED;
    } else {
printf("unknown reason %d, FOCUS = %d, LOSING = %d, VALUE = %d\n",
cbs->reason, XmCR_FOCUS, XmCR_LOSING_FOCUS, XmCR_VALUE_CHANGED);
    }
}


/*
 * Handle a click on a check box in a display window.
 */
void
displayButtonToggle(w, dad, cbs)
Widget				w;
displayAttrData			*dad;
XmToggleButtonCallbackStruct	*cbs;
{
    struct berval	**bvs;
    attrData		*ad = dad->dad_attrData;

    bvs = ad->ad_syndata->sd_fromnative(cbs->set);
    if (bervec_compare(bvs, ad->ad_vals)) {
	dad->dad_attrData->ad_flags &= ~AD_DIRTY;
    } else {
	dad->dad_attrData->ad_flags |= AD_DIRTY;
    }
    dad->dad_osval = (void *) cbs->set;
    updateDisplayDlgSaveButton(w, -1);
}




void
doDropSiteHighlight(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmDragProcCallbackStruct	*cbs;
{
    /*
     * Only animate if this is a valid drop site for the
     * data type being dragged.
     * XXX just drawing a one-pixel border around the labelGadget
     * is kind of boring, but we need to get version 1.0 out the
     * door...
     */
    if (cbs->dropSiteStatus == XmINVALID_DROP_SITE) {
	return;
    }

    switch (cbs->reason) {
    case XmCR_DROP_SITE_ENTER_MESSAGE:
	XtVaSetValues(w, XmNhighlightThickness, 1, NULL);

	break;
    case XmCR_DROP_SITE_LEAVE_MESSAGE:
	XtVaSetValues(w, XmNhighlightThickness, 0, NULL);
	break;
    }
}


/*
 * If appropriate, set up this label widget as a drop site for
 * drag and drop operations.
 */
static void
setUpLabelDropSite(w, dad)
Widget		w;
displayAttrData	*dad;
{
    Display	*dpy;
    Atom	importList[5], TEXT, COMPOUND_TEXT, DN, LABELED_URL;
    Arg		args[8];
    int		ac, ic;
    attrData	*ad;

    ac = ic = 0;
    dpy = XtDisplay(toplevel);
    TEXT = XmInternAtom(dpy, "TEXT", False);
    COMPOUND_TEXT = XmInternAtom(dpy, "COMPOUND_TEXT", False);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);

    if (dad != NULL) {
	ad = dad->dad_attrData;
	switch (dad->dad_ti->ti_syntaxid) {
	case LDAP_SYN_LABELEDURL:
	    /*
	     * Labeled URL labels will take text, compound text, or real
	     * labeled URLs.
	     */
	    importList[ic] = LABELED_URL; ic++;
	    importList[ic] = TEXT; ic++;
	    importList[ic] = COMPOUND_TEXT; ic++;
	    XtSetArg(args[ac], XmNdropProc, handleLabelDrop); ac++;
	    break;
	case LDAP_SYN_DN:
	    /*
	     * DN labels will take text, compound text, or real DNs
	     */
	    importList[ic] = DN; ic++;
	    importList[ic] = TEXT; ic++;
	    importList[ic] = COMPOUND_TEXT; ic++;
	    XtSetArg(args[ac], XmNdropProc, handleLabelDrop); ac++;
	    break;
	default:
	    /*
	     * No other types support drag and drop at this time.
	     */
	    return;
	}
    }

    XtSetArg(args[ac], XmNimportTargets, importList); ac++;
    XtSetArg(args[ac], XmNnumImportTargets, ic); ac++;
    XtSetArg(args[ac], XmNdropSiteOperations, XmDROP_COPY); ac++;
    XtSetArg(args[ac], XmNdragProc, doDropSiteHighlight); ac++;
    XtSetArg(args[ac], XmNanimationStyle, XmDRAG_UNDER_HIGHLIGHT); ac++;
    XmDropSiteRegister(w, args, ac);
}






static void
displayAddDNGlue(w, ld, dn, ocl, status, cdata)
Widget		w;
LDAP		*ld;
char		*dn;
char		**ocl;
Widget		status;
XtPointer	cdata;
{
    if (syn_AddDN((displayAttrData *) cdata, dn) == AD_DUPVALUE) {
	char **xdn;
	char errstr[256];
	displayAttrData	*dad = (displayAttrData *) cdata;

	xdn = ldap_explode_dn(dn, 1);
	sprintf(errstr,
	    "The value \"%s\" already appears in the \"%s\" attribute", 
	    xdn[0], dad->dad_label);
	errorDlg(GetTopShell(w), errstr);
	ldap_value_free(xdn);
    }
}


static void
displayAddDN(d, dad)
displayData	*d;
displayAttrData	*dad;
{
    (void) do_simple_find(d->di_displayDlg, "addDNDlg",
	    XmDIALOG_FULL_APPLICATION_MODAL, False, displayAddDNGlue, dad);
}


static void
displayAddJPEG(d, dad)
displayData	*d;
displayAttrData	*dad;
{
    errorDlg(d->di_displayDlg, "Adding JPEG photos is not yet supported");
}



static void
displayAddFax(d, dad)
displayData	*d;
displayAttrData	*dad;
{
    errorDlg(d->di_displayDlg, "Adding fax images is not yet supported");
}



static void
displayAddAudio(d, dad)
displayData	*d;
displayAttrData	*dad;
{
    errorDlg(d->di_displayDlg, "Adding Audio data is not yet supported");
}



static void
displayAddLabeledURL(d, dad)
displayData	*d;
displayAttrData	*dad;
{
    char	*label, *url, *value;
    int		ret;

    createURLEditDialog(d->di_displayDlg, dad->dad_attrData, &label, &url);
    if ((url == NULL) && (label == NULL)) {
	return;
    }
    if (url && strlen(url) < 1) {
	errorDlgEnum(d->di_displayDlg, E_BLANK_URL_NOT_ALLOWED);
	return;
    }
    if (label == NULL) {
	label = "";
    }
    value = (char *) malloc(strlen(url) + strlen(label) + 2);
    sprintf(value, "%s %s", url, label);
    ret = syn_AddURL(dad, value);
    switch (ret) {
    case AD_DUPVALUE:
	errorDlg(d->di_displayDlg,
		"That label/URL combination already is present in this entry");
	break;
    case AD_NOERR:
	break;
    default:
	errorDlg(d->di_displayDlg, "Unknown error adding labeled URL");
    }
    free(value);
    XtFree(url);
    XtFree(label);
}




static void
displayAddValue(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    struct ldap_tmplitem	*ti = (struct ldap_tmplitem *) client_data;
    displayAttrData		*dad;
    displayData			*d;

    if ((d = getDisplayDataByWidget(w)) == NULL) {
	errorDlg(w, "Can't find display data for that window");
	return;
    }

    if ((dad = getDisplayAttrDataByTemplate(d, ti)) == NULL) {
	errorDlg(d->di_displayDlg,
	"Can't find attribute data for that attribute");
	return;
    }

    switch(dad->dad_ti->ti_syntaxid) {
    case LDAP_SYN_DN:
	displayAddDN(d, dad);
	break;
    case LDAP_SYN_JPEGBUTTON:
	displayAddJPEG(d, dad);
	break;
    case LDAP_SYN_FAXBUTTON:
	displayAddFax(d, dad);
	break;
    case LDAP_SYN_AUDIOBUTTON:
	displayAddAudio(d, dad);
	break;
    case LDAP_SYN_LABELEDURL:
	displayAddLabeledURL(d, dad);
	break;
    default:
	errorDlg(GetTopShell(w), "Cannot add that attribute type");
	break;
    }
}



static void
deleteEntryPrompt(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    int	response;
    static char *prompt =
    "Deleting entry will permanently remove it\nfrom the directory.  Proceed?";

    response = modalYesNo(GetTopShell(w), prompt);
    switch (response) {
    case dlgYES:
	deleteEntry(getDisplayDataByWidget(w));
	break;
    default:
	;
    }
}



/*
 * Handle a drop operation on the trash can icon.  First, determine the data
 * type dropped, and if valid, delete the value dropped.
 */
static void
handleTrashDrop(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    Display			*dpy;
    XmDropProcCallback		DropData;
    Widget			dc;
    Boolean			isDN = False, isLABELED_URL = False;
    XmDropTransferEntryRec	transferEntries[1];
    XmDropTransferEntry		transferList;
    Arg				args[10];
    int				n, i;
    Cardinal			numExportTargets;
    Atom			*exportTargets, DN, LABELED_URL;
    valueData			*vd;

    dpy = XtDisplay(w);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);
    DropData = (XmDropProcCallback) call_data;
    dc = DropData->dragContext;

    n = 0;
    XtSetArg(args[n], XmNexportTargets, &exportTargets); n++;
    XtSetArg(args[n], XmNnumExportTargets, &numExportTargets); n++;
    XtSetArg(args[n], XmNuserData, &vd); n++;
    XtGetValues(dc, args, n);
    for (i = 0; i < numExportTargets; i++) {
	if (exportTargets[i] == DN) {
	    isDN = True;
	} else if (exportTargets[i] == LABELED_URL) {
	    isLABELED_URL = True;
	}
    }

    n = 0;

    if (!(isDN || isLABELED_URL) || (DropData->dropAction != XmDROP) ||
	    (DropData->operation != XmDROP_COPY)) {
	XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE); n++;
	XtSetArg(args[n], XmNnumDropTransfers, 0); n++;
    } else {
	transferEntries[0].target = isDN ? DN : LABELED_URL;
	transferEntries[0].client_data = (XtPointer) w;
	transferList = transferEntries;
	XtSetArg(args[n], XmNdropTransfers, transferEntries); n++;
	XtSetArg(args[n], XmNnumDropTransfers,
		XtNumber(transferEntries)); n++;
	XtSetArg(args[n], XmNtransferProc, transferDrop); n++;
    }
    XmDropTransferStart(dc, args, n);
}





/*
 * Create a pixmap button with a picture of a trash can, and arrange
 * for it to (1) delete the entry when clicked on, and (2) serve as a
 * drop site for labeled URLs and DNs, in which case the value will be
 * deleted.
 */
static Widget
createTrashButton(parent)
Widget	parent;
{
    Atom	importList[7], DN, LABELED_URL;
    int		ic, ac;
    Arg		args[8];
    Widget	trashButton;
    Display	*dpy;
    Pixmap	open_trash;

    dpy = XtDisplay(parent);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);

    /* Create the trash button */
    trashButton = create_pixmap_button(parent, "trash_image",
	    NULL, "trashButton");
    XtManageChild(trashButton);

    /* Set up the activate callback */
    XtAddCallback(trashButton, XmNactivateCallback, deleteEntryPrompt, NULL);

    /* Set up as a drop site */
    ac = ic = 0;
    importList[ic] = DN; ic++;
    importList[ic] = LABELED_URL; ic++;
    XtSetArg(args[ac], XmNimportTargets, importList); ac++;
    XtSetArg(args[ac], XmNnumImportTargets, ic); ac++;
    XtSetArg(args[ac], XmNdropSiteOperations, XmDROP_COPY); ac++;
    XtSetArg(args[ac], XmNdropProc, handleTrashDrop); ac++;
    XtSetArg(args[ac], XmNanimationStyle, XmDRAG_UNDER_PIXMAP); ac++;
    open_trash = XmGetPixmapByDepth(XtScreen(parent), "trash_open_image",
	    1, 0, 1);
    XtSetArg(args[ac], XmNanimationPixmap, open_trash); ac++;
    XmDropSiteRegister(trashButton, args, ac);
    return(trashButton);
}


/*
 * Create a frame into which we'll place the entry's attributes.
 */
static Widget
create_display_frame(parent, retDisplayData, tmpl)
Widget			parent;
displayData		**retDisplayData;
struct ldap_disptmpl	*tmpl;
{
    Widget		displayDlg, displayPane, displaySWin,
			displayAction, displayDNForm, displayStatus,
			displayAsOption, displaySaveButton, displayHelpButton,
			browseButton, displayMenuBar, addMenu,
			findButton, dismissButton, trashButton;
    XmString		s;
    int			i, sel_item;
    Arg			args[10];
    MenuItem		*menuItems;
    displayData		*d;
    struct ldap_tmplitem *ti, *ti2;
    Atom		wm_delete_window;

    if ((d = getDisplayData()) == NULL) {
	fprintf(stderr, "out of memory, malloc displayData\n");
	exit(1);
    }
    /* put this at the head of the linked list of entries */
    d->di_next = displayDlgData;

    displayDlgData = d;
    *retDisplayData = d;

    displayDlg = XtVaAppCreateShell("displayDlg", "Xax500",
	    applicationShellWidgetClass,
	    XtDisplay(toplevel),
	    NULL);

    wm_delete_window = XmInternAtom(XtDisplay(displayDlg),
	    "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(displayDlg, wm_delete_window,
	    (XtCallbackProc) displayClose, d);

    /*
     * We need to manage the whole thing, and set sizes.  So,
     * arrange for the dialog to remain unmapped when it is
     * managed.  Otherwise, we (might?) get flicker.
     */
    XtSetMappedWhenManaged(displayDlg, False);
    XtRealizeWidget(displayDlg);

    displayPane = XtVaCreateWidget("displayPane",
	    xmPanedWindowWidgetClass,
	    displayDlg,
	    XmNsashWidth, 1,
	    XmNsashHeight, 1,
	    NULL);

    /*
     * Add a little form widget to hold the DN and the shortcut buttons.
     */
    displayDNForm = XtVaCreateWidget("displayDNForm",
	    xmFormWidgetClass, displayPane,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    NULL);

    /* Create the trash button */
    trashButton = createTrashButton(displayDNForm);
    XtVaSetValues(trashButton,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    NULL);
    XtManageChild(trashButton);

    /* Create a browse button - shortcut to get to the browser */
    browseButton = create_pixmap_button(displayDNForm, "world_image",
	    NULL, "browseButton");
    XtVaSetValues(browseButton,
	    XmNrightAttachment, XmATTACH_WIDGET,
	    XmNrightWidget, trashButton,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    NULL);
    XtAddCallback(browseButton, XmNactivateCallback,
	    (XtCallbackProc) do_browse, NULL);
    XtManageChild(browseButton);

    /* Create a find button - shortcut to raise find window */
    findButton = create_find_button(displayDNForm);

    XtVaSetValues(findButton,
	    XmNrightAttachment, XmATTACH_WIDGET,
	    XmNrightWidget, browseButton,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    NULL);

    XtAddCallback(findButton, XmNactivateCallback,
	    (XtCallbackProc) raiseFindWindow, NULL);
    XtManageChild(findButton);

    /*
     * Create a label widget with the RDN of this entry and the type of
     * entry.  In display_entry, it will be set up as a drag source.
     */
    i = 0;
    XtSetArg(args[i], XmNalignment, XmALIGNMENT_BEGINNING); i++;
    d->di_displayDN = XtVaCreateManagedWidget("displayDN",
	    xmLabelWidgetClass, displayDNForm, 
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_WIDGET,
	    XmNrightWidget, findButton,
	    NULL);
    XtManageChild(d->di_displayDN);

    /*
     * Create a menu bar 
     */
    displayMenuBar = XmCreateMenuBar(displayPane, "displayMenuBar", NULL, 0);
    XtVaSetValues(displayMenuBar,
	    XmNleftAttachment, XmATTACH_FORM,
	    NULL);
    /*
     * Add an "Add" option menu with a list of all the attributes one
     * can add a value to (e.g. all non-read-only).
     */
    /* count the number of writable attributes */
    i = 0;
    for (ti = ldap_first_tmplrow(tmpl); ti != NULLTMPLITEM;
	    ti = ldap_next_tmplrow(tmpl, ti)) {
	for (ti2 = ldap_first_tmplcol(tmpl, ti); ti2 != NULLTMPLITEM;
		ti2 = ldap_next_tmplcol(tmpl, ti, ti2)) {
	    if (LDAP_IS_TMPLITEM_OPTION_SET(ti2, LDAP_DITEM_OPT_READONLY)) {
		continue;
	    } else {
		i++;
	    }
	}
    }

    /*
     * Create a pulldown menu with the attribute labels as the menu
     * items.  The callback invoked is displayAddValue(), and the call_data
     * is a pointer to the template info for that attribute.  May
     * need a pointer to the display information so we can update
     * the attributes and window when finished.
     */

    menuItems = (MenuItem *) malloc((i + 1) * sizeof(MenuItem));
    sel_item = 0;
    for (i = 0, ti = ldap_first_tmplrow(tmpl); ti != NULLTMPLITEM;
	    ti = ldap_next_tmplrow(tmpl, ti)) {
	for (ti2 = ldap_first_tmplcol(tmpl, ti); ti2 != NULLTMPLITEM;
		ti2 = ldap_next_tmplcol(tmpl, ti, ti2)) {
	    /*
	     * Skip any readonly, boolean, text, time, or date
	     * atributes.
	     */
	    if (LDAP_IS_TMPLITEM_OPTION_SET(ti2, LDAP_DITEM_OPT_READONLY)) {
		continue;
	    }
	    if ((ti2->ti_syntaxid == LDAP_SYN_DN) ||
		    (ti2->ti_syntaxid == LDAP_SYN_JPEGIMAGE) ||
		    (ti2->ti_syntaxid == LDAP_SYN_JPEGBUTTON) ||
		    (ti2->ti_syntaxid == LDAP_SYN_FAXIMAGE) ||
		    (ti2->ti_syntaxid == LDAP_SYN_FAXBUTTON) ||
		    (ti2->ti_syntaxid == LDAP_SYN_AUDIOBUTTON) ||
		    (ti2->ti_syntaxid == LDAP_SYN_LABELEDURL)) {

		menuItems[i].label = ti2->ti_label;
		menuItems[i].widgetname = ti2->ti_attrname;
		menuItems[i].class = &xmPushButtonGadgetClass;
		menuItems[i].mnemonic = '\0';
		menuItems[i].accelerator = NULL;
		menuItems[i].accel_text = NULL;
		menuItems[i].callback = displayAddValue;
		menuItems[i].callback_data = (XtPointer) ti2;
		menuItems[i].subitems = NULL;
		menuItems[i].ret_widget = NULL;
		i++;
	    }
	}
    }

    menuItems[i].widgetname = NULL;
    /* XXX not hardcoded string, calc default selected menu item */
    addMenu = BuildMenu(displayMenuBar, XmMENU_PULLDOWN,
	    "addMenu", 'A', menuItems, 0);
    free(menuItems);
		
    XtManageChild(addMenu);
    XtManageChild(displayMenuBar);


#ifdef notyet
    /*
     * Add a "Display as: " option menu, and a small icon which will
     * bring up a browser.
     */

    /* First count the number of templates */
    /* XXX - optimize this out */
    for (i = 0, dt = ldap_first_disptmpl(templates); dt != NULL;
	    dt = ldap_next_disptmpl(templates, dt), i++);
    optionItems = (MenuItem *) malloc((i + 1) * sizeof(MenuItem));
    sel_item = 0;	/* XXX should be index of "generic" */
    for (i = 0, dt = ldap_first_disptmpl(templates); dt != NULL;
	    dt = ldap_next_disptmpl(templates, dt), i++) {
	optionItems[i].label = dt->dt_name;
	optionItems[i].widgetname = dt->dt_name;
	optionItems[i].class = &xmPushButtonGadgetClass;
	optionItems[i].mnemonic = '\0';
	optionItems[i].accelerator = NULL;
	optionItems[i].accel_text = NULL;
	optionItems[i].callback = displayNewDisplayAs;
	optionItems[i].callback_data = (XtPointer) d;
	optionItems[i].subitems = NULL;
	optionItems[i].ret_widget = NULL;
	if (dt->dt_name == tmpl->dt_name) {
	    sel_item = i;
	}
    }
    optionItems[i].widgetname = NULL;
    /* XXX not hardcoded string, calc default selected menu item */
    displayAsOption = BuildMenu(displayMenuBar, XmMENU_OPTION,
	    "displayAsOption", 'F', optionItems,
	    sel_item);
    free(optionItems);
		
    XtManageChild(displayAsOption);
#else
    displayAsOption = NULL;
#endif /* notyet - display as option menu */

    XtManageChild(displayDNForm);

    /*
     * This ScrolledWindow widget is where all the entry's
     * attributes will be put.
     */
    i = 0;
    XtSetArg(args[i], XmNscrollingPolicy, XmAUTOMATIC); i++;
    displaySWin = XmCreateScrolledWindow(displayPane, "displaySWin",
	    args, i);

    /* Status area */
    s = XmStringCreateLtoR(" ", XmSTRING_DEFAULT_CHARSET);
    i = 0;
    XtSetArg(args[i], XmNlabelString, s); i++;
    displayStatus = XtCreateManagedWidget("displayStatus",
	    xmLabelWidgetClass,
	    displayPane, args, i);
    XmStringFree(s);

    /* Create an action area */
    displayActions[0].data = (caddr_t) d;
    displayActions[0].widget = &dismissButton;
    displayActions[0].callback = displayClose;
    displayActions[1].data = (caddr_t) d;
    displayActions[1].widget = &displaySaveButton;
    displayActions[1].callback = displaySaveCB;
    displayActions[2].data = (XtPointer) HLP_TOPIC_VIEWING;
    displayActions[2].widget = &displayHelpButton;
    displayActions[2].callback = doHelpDlg;
    displayAction = CreateActionArea(displayPane, displayActions,
	    XtNumber(displayActions));
    XtSetSensitive(displaySaveButton, False);
    XtManageChild(displayAction);


    /* Save away the data we'll need in callbacks */
    d->di_displayDlg = displayDlg;
    d->di_status = displayStatus;
    d->di_displaySWin = displaySWin;
    d->di_displaySaveButton = displaySaveButton;
    d->di_displayDismissButton = dismissButton;
    d->di_displayAsOption = displayAsOption;
    d->di_numMods = 0;
	
    XtManageChild(displayPane);
    return(displayDlg);
}



/*
 * Return true if the attribute data struct points to a DN or URL
 * attribute.
 */
static Boolean
isDNorURL(dad)
displayAttrData	*dad;
{
    return ((dad->dad_ti->ti_syntaxid == LDAP_SYN_DN) ||
           (dad->dad_ti->ti_syntaxid == LDAP_SYN_LABELEDURL));
}


/*
 * Event handler to ensure that the topmost display dialog is the one
 * used when deciding who to authenticate as.  Changes global
 * "latestDisplayDlg" to point to this dialog.
 */
static void
updateLatest(w, unused, event, continue_to_dispatch)
Widget			w;
XtPointer		unused;
XVisibilityEvent	*event;
Boolean			*continue_to_dispatch;
{
    displayData	*d;

    if (event->state == VisibilityUnobscured) {
	for (d = displayDlgData; d != NULL; d = d->di_next) {
	    if (d->di_displayDlg == w) {
		latestDisplayDlg = d;
	    }
	}
    }
}


/*
 * Display an entry.  Calls create_display_frame() to construct a frame
 * to draw into, then puts entries into the frame based on the template
 * specification pointed to by dt.
 */
void
display_entry(en, dt)
entryData		*en;
struct ldap_disptmpl	*dt;
{
    Widget			displayFrame, displayForm;
    char			*dn, **xdn;
    displayAttrData		**dad, *dad2;
    attrData			*ad;
    displayData			*dd;
    int				row, col, nrows, ncols, bi;
    Dimension			labelWidth;
    Dimension			*labelWidths;
    Widget			upperRightValue, leftLabel;
    struct ldap_tmplitem	*ti, *ti2;
    struct	syntaxdata	*sy;
    XmString			s;
    valueData			*vd;
    char			buf[128];
    char			*label;
    Boolean			disableWidget;

    /*
     * If app config says not to allow multiple windows, then
     * find the currently open window and close it, invoking
     * callbacks so the entry will get saved if it's been modified.
     */
    if (!appConfig.multipleEntryWindows && entryDlgActive) {
	entryDlgActive = False;
	/* destroy existing dialog */
	XtCallCallbacks(displayDlgData->di_displayDismissButton,
		XmNactivateCallback, displayDlgData);
    }
    entryDlgActive = True;
    
    /*
     * Create the new display frame and start to fill in the
     * displayData struct
     */
    displayFrame = create_display_frame(toplevel, &dd, dt);
    dd->di_entryData = en;
    dd->di_tmpl = dt;
    dd->di_valueDataList = NULL;

    /* set the dialog title */
    xdn = ldap_explode_dn(dd->di_entryData->en_dn, 1);
    XtVaSetValues(displayFrame, XmNtitle, xdn[0], XmNiconName, xdn[0], NULL);

    /*
     * Arrange for a callback to check for modifications to values
     * and update data structures as needed.
     */
    XtAddEventHandler(displayFrame, VisibilityChangeMask, True,
	    updateLatest, dd);

    /* Have to know how many cols before we start.  Sigh. */
    for (row = 0, ti = ldap_first_tmplrow(dt); ti != NULLTMPLITEM;
	    row++, ti = ldap_next_tmplrow(dt, ti)) {
	for (col = 0, ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULLTMPLITEM;
		col++, ti2 = ldap_next_tmplcol(dt, ti, ti2)) {
	}
    }
    nrows = row + 1;
    ncols = col + 1;

    /* 
     * Create a Form Widget to hold the labels and values.  We set the
     * translations on the form widget so that the label Gadgets contained
     * in the form widget aren't drag sources.
     */
    displayForm = XtVaCreateWidget("displayForm", xmFormWidgetClass,
	    dd->di_displaySWin,
	    XmNfractionBase, ncols,
	    XmNtranslations, parsedNoDragTranslations,
	    NULL);

    labelWidths = (Dimension *) malloc(ncols * sizeof(Dimension));
    memset(labelWidths, 0, ncols * sizeof(Dimension));

    /* set the DN area of the display widget XXX - bug here */
    vd = get_valuedata(dd);
    vd->vd_tag = VD_DN;
    vd->vd_displayAttrData = NULL;
    vd->vd_displayData = dd;
    vd->vd_value = strdup(dd->di_entryData->en_dn);
    sprintf(buf, "%s (%s)", xdn[0], dd->di_tmpl->dt_name);
    s = XmStringCreateLocalized(buf);
    XtVaSetValues(dd->di_displayDN,
	    XmNlabelString, s,
	    XmNtranslations, parsedDragTranslations,
	    XmNuserData, vd,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);
    XmStringFree(s);
    ldap_value_free(xdn);


    /*
     * Create the linked list data structure and squirrel away
     * the data we'll need to calculate the layout geometry,
     * as well as stuff we'll need for modifies.  Also, when
     * creating labels and values, keep track of the largest
     * in each column.
     */

    dad = &dd->di_displayAttrData;
    /* Create the widgets to display the entry's attributes */
    for (row = 0, ti = ldap_first_tmplrow(dt); ti != NULLTMPLITEM;
	     row++, ti = ldap_next_tmplrow(dt, ti)) {
	for (col = 0, ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULLTMPLITEM;
		     col++, ti2 = ldap_next_tmplcol(dt, ti, ti2)) {

	    disableWidget = False;
	    *dad = (displayAttrData *) malloc(sizeof (displayAttrData)); 
	    (*dad)->dad_label = ti2->ti_label;
	    (*dad)->dad_labelWidget = NULL;
	    (*dad)->dad_valueWidget = NULL;
	    (*dad)->dad_valueWidgetList = NULL;
	    (*dad)->dad_numValueWidgets = 0;
	    (*dad)->dad_editinfo = NULL;
	    (*dad)->dad_entryData = en;
	    (*dad)->dad_osval = NULL;
	    (*dad)->dad_buttonValIndex = 0;
	    (*dad)->dad_col = col;
	    (*dad)->dad_row = row;
	    (*dad)->dad_flags = 0;
	    sy = LDAP_GET_TMPLITEM_APPDATA(ti2, syntaxdata *);
	    (*dad)->dad_ti = ti2;

	    if (LDAP_GET_SYN_TYPE(ti2->ti_syntaxid) == LDAP_SYN_TYPE_ACTION) {

		/*
		 * Handle action items
		 */

		/* XXX - skip over types not yet implemented */
		switch (ti2->ti_syntaxid) {
		case LDAP_SYN_ADDDNACTION:
		    break;
		case LDAP_SYN_SEARCHACTION:
		    break;
		default:
		    continue;
		}

		
		ad = (*dad)->dad_attrData = 
			(attrData *) malloc(sizeof(attrData));
		ad->ad_attrname = NULL;
		ad->ad_vals = NULL;
		ad->ad_nvals = 0;
		ad->ad_syndata = sy;
		ad->ad_flags = 0;
		/* XYZZY!!! Arrgh! We've already allocated an ad for this
		somewhere, but how do we find it?  My brain is mush now. */
		/* Ok, modify ldapmessage2entry to not alloc one?? */

		switch (ti2->ti_syntaxid) {
		case LDAP_SYN_ADDDNACTION:
		    /* Disable this item if disable attr == FALSE (or empty) */
		    if ((ti2->ti_args != NULL) &&
			    (ti2->ti_args[AP_ADIS] != NULL)) {
			attrData *tad;
			tad = getAttrDataByName(en->en_attrData,
				ti2->ti_args[AP_ADIS]);
			if (tad != NULL) {
			    if (tad->ad_vals != NULL) {
				if (strcasecmp(tad->ad_vals[0]->bv_val,
					"TRUE")) {
				    disableWidget = True;
				}
				    
			    }
			}
		    }
		    (*dad)->dad_osval = NULL;
		    ad->ad_nvals = 0;
		    break;
		case LDAP_SYN_SEARCHACTION:
		    (*dad)->dad_osval = NULL;
		    ad->ad_nvals = 0;
		    /* Disable this item if disable attr == FALSE (or empty) */
		    if ((ti2->ti_args != NULL) &&
			    (ti2->ti_args[SP_ADIS] != NULL)) {
			attrData *tad;
			tad = getAttrDataByName(en->en_attrData,
				ti2->ti_args[SP_ADIS]);
			if (tad != NULL) {
			    if (tad->ad_vals != NULL) {
				if (strcasecmp(tad->ad_vals[0]->bv_val,
					"TRUE")) {
				    disableWidget = True;
				}
				    
			    }
			}
		    }
		    break;
		default:
		}
		label = "";	/* No label for action items */

	    } else {

		/*
		 * Handle normal attributes
		 */

		ad = (*dad)->dad_attrData = getAttrDataByName(en->en_attrData,
			ti2->ti_attrname);
		ad->ad_nvals = 0;

		/* count the values we got */
		if ((*dad)->dad_attrData->ad_vals == NULL) {
		    ad->ad_nvals = 0;
		    (*dad)->dad_attrData->ad_flags |= AD_STARTED_EMPTY;
		} else {
		    for (bi = 0; (*dad)->dad_attrData->ad_vals[bi] != NULL;
			    bi++);
		    ad->ad_nvals = bi;
		}

		/* Convert to on-screen format */
		(*dad)->dad_osval =
			sy->sd_tonative((*dad)->dad_attrData->ad_vals);

		/* Use the label text for the label widget */
		label = (*dad)->dad_label;
	    }


	    /* Create the label */
	    s = XmStringCreateSimple(label);
	    (*dad)->dad_labelWidget = XtVaCreateManagedWidget(
		    (*dad)->dad_label,
		    xmLabelGadgetClass,
		    displayForm,
		    XmNalignment, XmALIGNMENT_END,
		    XmNuserData, (*dad)->dad_label,
		    XmNlabelString, s,
		    XmNrecomputeSize, False,
		    NULL);
	    XmStringFree(s);
	    setUpLabelDropSite((*dad)->dad_labelWidget, *dad);

	    /* Create the widget for the value */
	    (*dad)->dad_valueWidget = sy->sd_display(displayForm, *dad);
	    XtManageChild((*dad)->dad_valueWidget);

	    /* Set attachments for the value widget */
	    XtVaSetValues((*dad)->dad_valueWidget,
		    XmNleftAttachment,
		    XmATTACH_WIDGET,
		    XmNleftWidget,
		    (*dad)->dad_labelWidget,
		    NULL);
	    if (row == 0) {
		XtVaSetValues((*dad)->dad_valueWidget,
			XmNtopAttachment,
			XmATTACH_FORM,
			NULL);
	    } else {
		XtVaSetValues((*dad)->dad_valueWidget,
			XmNtopAttachment,
			XmATTACH_OPPOSITE_WIDGET,
			XmNtopWidget,
			(*dad)->dad_labelWidget,
			NULL);
	    }
	    /* Disable if flagged above */
	    if (disableWidget) {
		XtSetSensitive((*dad)->dad_valueWidget, False);
	    }

	    /* Set attachment resources for the label widget */
	    if (col == 0) {
		if (isDNorURL(*dad)) {
		    XtVaSetValues((*dad)->dad_labelWidget,
			    XmNrightAttachment,
			    XmATTACH_WIDGET,
			    XmNrightWidget,
			    upperRightValue,
			    NULL);
		} else {
		    XtVaSetValues((*dad)->dad_labelWidget,
			    XmNleftAttachment,
			    XmATTACH_POSITION,
			    XmNleftPosition,
			    col,
			    NULL);
		}
		if (row == 0) {
		    XtVaSetValues((*dad)->dad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_FORM,
			    NULL);
		} else {
		    XtVaSetValues((*dad)->dad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_WIDGET,
			    XmNtopWidget,
			    upperRightValue,
			    NULL);
		}
	    } else {
		XtVaSetValues((*dad)->dad_labelWidget,
			XmNleftAttachment,
			XmATTACH_POSITION,
			XmNleftPosition,
			col,
			NULL);
		if (row == 0) {
		    XtVaSetValues((*dad)->dad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_FORM,
			    NULL);
		} else {
		    XtVaSetValues((*dad)->dad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_OPPOSITE_WIDGET,
			    XmNtopWidget,
			    upperRightValue,
			    NULL);
		}
	    }
	    leftLabel = (*dad)->dad_labelWidget;
	    upperRightValue = (*dad)->dad_valueWidget;

	    /* Is this the widest label? */
	    XtVaGetValues((*dad)->dad_labelWidget, XmNwidth, &labelWidth,
		    NULL);
	    if (labelWidths[col] < labelWidth) {
		labelWidths[col] = labelWidth;
	    }
	    dad = &((*dad)->dad_next);
	}
    }
    *dad = NULL;


    /*
     * Now we've created all the widgets and set the left and top
     * attachments, set the label sizes.
     */
	
    for (dad2 = dd->di_displayAttrData; dad2 != NULL; dad2 = dad2->dad_next) {
	if (!isDNorURL(dad2)) {
	    /* Set the label width */
	    XtVaSetValues(dad2->dad_labelWidget, XmNwidth,
		    labelWidths[dad2->dad_col],
		    NULL);
	}
    }
    dd->di_currAttr = dd->di_prevAttr = dd->di_displayAttrData;
    dd->di_prevWidget = NULL;

    free(labelWidths);
    XtManageChild(displayForm);
    XtManageChild(dd->di_displaySWin);
    XtRealizeWidget(displayFrame);
    latestDisplayDlg = dd;
    XtMapWidget(displayFrame);
    updateDisplayDlgSaveButton(displayFrame, -1);
}



#ifdef notdef
/*
 * Display an entry.  Calls create_display_frame() to construct a frame
 * to draw into, then puts entries into the frame based on the template
 * specification pointed to by dt.
 */
void
old_display_entry(ld, e, dt)
LDAP			*ld;
LDAPMessage		*e;
struct ldap_disptmpl	*dt;
{
    Widget			w, displayForm;
    char			*dn, **xdn;
    attrData			*ad, **aad;
    displayData			*dd;
    entryData			*entry;
    int				row, col, nrows, ncols, bi;
    Dimension			labelWidth;
    Dimension			*labelWidths;
    Widget			upperRightValue, leftLabel;
    struct ldap_tmplitem	*ti, *ti2;
    struct	syntaxdata	*sy;
    XmString			s;
    valueData			*vd;
    char			buf[128];
    char			*label;

    if (!appConfig.multipleEntryWindows && entryDlgActive) {
	entryDlgActive = False;
	/* destroy existing dialog */
	XtCallCallbacks(displayDlgData->di_displayDismissButton,
		XmNactivateCallback, displayDlgData);
    }
    entryDlgActive = True;
    w = create_display_frame(toplevel, &dd, dt);
    XtAddEventHandler(w, VisibilityChangeMask, True,
	    updateLatest, dd);

    /* Have to know how many cols before we start.  Sigh. */
    for (row = 0, ti = ldap_first_tmplrow(dt); ti != NULLTMPLITEM;
	    row++, ti = ldap_next_tmplrow(dt, ti)) {
	for (col = 0, ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULLTMPLITEM;
		col++, ti2 = ldap_next_tmplcol(dt, ti, ti2)) {
	}
    }
    nrows = row + 1;
    ncols = col + 1;

    /* 
     * Create a Form Widget to hold the labels and values.  We set the
     * translations on the form widget so that the label Gadgets contained
     * in the form widget aren't drag sources.
     */
    displayForm = XtVaCreateWidget("displayForm", xmFormWidgetClass,
	    dd->di_displaySWin,
	    XmNfractionBase, ncols,
	    XmNtranslations, parsedNoDragTranslations,
	    NULL);


    /* set the dialog title */
    dn = ldap_get_dn(ld, e);
    xdn = ldap_explode_dn(dn, 1);
    dd->di_dn = dn;
    XtVaSetValues(w, XmNtitle, xdn[0], XmNiconName, xdn[0], NULL);

    labelWidths = (Dimension *) malloc(ncols * sizeof(Dimension));
    memset(labelWidths, 0, ncols * sizeof(Dimension));

    /*
     * Set up an entry structure for the X.500 entry.
     */
    if ((entry = (entryData *) malloc(sizeof(entryData))) == NULL) {
	fprintf(stderr, "out of memory, malloc entryData\n");
	exit(1);
    }
    entry->en_oclist = ldap_get_values(ld, e, "objectClass");
    entry->en_krbNames = ldap_get_values(ld, e, "krbName");
    entry->en_dn = strdup(dn);
    entry->en_isnew = False;
    entry->en_modified = False;
    entry->en_tmpl = dt;
    entry->en_valueDataList = NULL;
    dd->di_entryData = entry;

    /* set the DN area of the display widget */
    vd = get_valuedata(entry);
    vd->vd_tag = VD_DN;
    vd->vd_attrData = NULL;
    vd->vd_displayData = dd;
    vd->vd_value = strdup(dd->di_dn);
    sprintf(buf, "%s (%s)", xdn[0], dd->di_entryData->en_tmpl->dt_name);
    s = XmStringCreateLocalized(buf);
    XtVaSetValues(dd->di_displayDN,
	    XmNlabelString, s,
	    XmNtranslations, parsedDragTranslations,
	    XmNuserData, vd,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);
    XmStringFree(s);
    ldap_value_free(xdn);


    /*
     * Create the linked list data structure and squirrel away
     * the data we'll need to calculate the layout geometry,
     * as well as stuff we'll need for modifies.  Also, when
     * creating labels and values, keep track of the largest
     * in each column.
     */

    aad = &entry->en_attrData;
    /* Create the widgets to display the entry's attributes */
    for (row = 0, ti = ldap_first_tmplrow(dt); ti != NULLTMPLITEM;
	     row++, ti = ldap_next_tmplrow(dt, ti)) {
	for (col = 0, ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULLTMPLITEM;
		     col++, ti2 = ldap_next_tmplcol(dt, ti, ti2)) {

	    if (LDAP_GET_SYN_TYPE(ti2->ti_syntaxid) == LDAP_SYN_TYPE_ACTION) {
		switch (ti2->ti_syntaxid) {
		case LDAP_SYN_ADDDNACTION:
		    break;
		default:
		    continue;
		}
	    }

	    *aad = (attrData *) malloc(sizeof (attrData)); 
	    (*aad)->ad_label = ti2->ti_label;
	    (*aad)->ad_col = col;
	    (*aad)->ad_row = row;
	    (*aad)->ad_editinfo = NULL;
	    (*aad)->ad_flags = 0;
	    (*aad)->ad_buttonValIndex = 0;
	    (*aad)->ad_attrname = ti2->ti_attrname;
	    (*aad)->ad_valueWidgetList = NULL;
	    (*aad)->ad_numValueWidgets = 0;
	    (*aad)->ad_entryData = entry;
	    sy = LDAP_GET_TMPLITEM_APPDATA(ti2, syntaxdata *);

	    (*aad)->ad_ti = ti2;
	    /* Don't mess with vals if this is an action item */
	    if (LDAP_GET_SYN_TYPE(ti2->ti_syntaxid) != LDAP_SYN_TYPE_ACTION) {
		(*aad)->ad_vals = ldap_get_values_len(ld, e, ti2->ti_attrname);

		/* count the values we got */
		if ((*aad)->ad_vals == 0) {
		    (*aad)->ad_nvals = 0;
		    (*aad)->ad_flags |= AD_STARTED_EMPTY;
		} else {
		    for (bi = 0; (*aad)->ad_vals[bi] != NULL; bi++);
		    (*aad)->ad_nvals = bi;
		}
		/* Convert to on-screen format */
		(*aad)->ad_osval = sy->sd_tonative((*aad)->ad_vals);
		/* Use the label text for the label widget */
		label = (*aad)->ad_label;
	    } else {
		/* Handle action types */
		switch (ti2->ti_syntaxid) {
		case LDAP_SYN_ADDDNACTION:
		    (*aad)->ad_osval = NULL;
		    (*aad)->ad_nvals = 0;
		    (*aad)->ad_vals = NULL;
		    /* Disable button if disable attr = TRUE */
		    break;
		default:
		}
		label = "";	/* No label for action items */
	    }

	    (*aad)->ad_syndata = sy;
			
	    /* Create the label */
	    s = XmStringCreateSimple(label);
	    (*aad)->ad_labelWidget = XtVaCreateManagedWidget(
		    (*aad)->ad_label,
		    xmLabelGadgetClass,
		    displayForm,
		    XmNalignment, XmALIGNMENT_END,
		    XmNuserData, (*aad)->ad_label,
		    XmNlabelString, s,
		    XmNrecomputeSize, False,
		    NULL);
	    XmStringFree(s);
	    setUpLabelDropSite((*aad)->ad_labelWidget, *aad);

	    /* Create the widget for the value */
	    (*aad)->ad_valueWidget = sy->sd_display(displayForm, *aad);
	    XtManageChild((*aad)->ad_valueWidget);

	    /* Set attachments for the value widget */
	    XtVaSetValues((*aad)->ad_valueWidget,
		    XmNleftAttachment,
		    XmATTACH_WIDGET,
		    XmNleftWidget,
		    (*aad)->ad_labelWidget,
		    NULL);
	    if (row == 0) {
		XtVaSetValues((*aad)->ad_valueWidget,
			XmNtopAttachment,
			XmATTACH_FORM,
			NULL);
	    } else {
		XtVaSetValues((*aad)->ad_valueWidget,
			XmNtopAttachment,
			XmATTACH_OPPOSITE_WIDGET,
			XmNtopWidget,
			(*aad)->ad_labelWidget,
			NULL);
	    }

	    /* Set attachment resources for the label widget */
	    if (col == 0) {
		if (isDNorURL(*aad)) {
		    XtVaSetValues((*aad)->ad_labelWidget,
			    XmNrightAttachment,
			    XmATTACH_WIDGET,
			    XmNrightWidget,
			    upperRightValue,
			    NULL);
		} else {
		    XtVaSetValues((*aad)->ad_labelWidget,
			    XmNleftAttachment,
			    XmATTACH_POSITION,
			    XmNleftPosition,
			    col,
			    NULL);
		}
		if (row == 0) {
		    XtVaSetValues((*aad)->ad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_FORM,
			    NULL);
		} else {
		    XtVaSetValues((*aad)->ad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_WIDGET,
			    XmNtopWidget,
			    upperRightValue,
			    NULL);
		}
	    } else {
		XtVaSetValues((*aad)->ad_labelWidget,
			XmNleftAttachment,
			XmATTACH_POSITION,
			XmNleftPosition,
			col,
			NULL);
		if (row == 0) {
		    XtVaSetValues((*aad)->ad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_FORM,
			    NULL);
		} else {
		    XtVaSetValues((*aad)->ad_labelWidget,
			    XmNtopAttachment,
			    XmATTACH_OPPOSITE_WIDGET,
			    XmNtopWidget,
			    upperRightValue,
			    NULL);
		}
	    }
	    leftLabel = (*aad)->ad_labelWidget;
	    upperRightValue = (*aad)->ad_valueWidget;

	    /* Is this the widest label? */
	    XtVaGetValues((*aad)->ad_labelWidget, XmNwidth, &labelWidth,
		    NULL);
	    if (labelWidths[col] < labelWidth) {
		labelWidths[col] = labelWidth;
	    }
	    aad = &((*aad)->ad_next);
	}
    }
    *aad = NULL;


    /*
     * Now we've created all the widgets and set the left and top
     * attachments, set the label sizes.
     */
	
    for (ad = entry->en_attrData; ad != NULL; ad = ad->ad_next) {
	if (!isDNorURL(ad)) {
	    /* Set the label width */
	    XtVaSetValues(ad->ad_labelWidget, XmNwidth, labelWidths[ad->ad_col],
		    NULL);
	}
    }
    dd->di_currAttr = dd->di_prevAttr = entry->en_attrData;
    dd->di_prevWidget = NULL;

    free(labelWidths);
    XtManageChild(displayForm);
    XtManageChild(dd->di_displaySWin);
    XtRealizeWidget(w);
    latestDisplayDlg = dd;
    XtMapWidget(w);
}
#endif /* notdef */



void
displayNewDisplayAs()
{
    fprintf(stderr, "Display-As function not yet implemented\n");
}





/*
 * Entry point used when we want to save directly from another function.
 */
static void
displayDoSave(d)
displayData	*d;
{
    latestDisplayDlg = d;
    writeEntry(d);
}


/*
 * Entry point used when we want to save via a callback, e.g. when user
 * clicks on the "Save" button.
 */
static void
displaySaveCB(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    displayDoSave((displayData *) client_data);
}



/*
 * Close this entry window.
 */
static void
displayClose(w, d, cbs)
Widget				w;
displayData			*d;
XmPushButtonCallbackStruct	*cbs;
{
    int		resp;

    if (!d) {
	fprintf(stderr, "Illegal displayData pointer, displayClose\n");
	exit(1);
    }
    if (d->di_entryData->en_modified) {
	/* Modal "save entry?" dialog XXX - get string from rsrc file */
	resp = modalYesNoCancel(GetTopShell(w), "Save entry?");
	switch (resp) {
	case dlgYES:
	    d->di_closeOnWrite = True;
	    displayDoSave(d);
	    return;
	case dlgNO:
	    break;
	case dlgCANCEL:
	    return;
	default:
	    return;
	}
    }
    freeDisplayData(d);
}


/*
 * Free all the data structures associated with this display window.
 */
void
freeDisplayData(dd)
displayData	*dd;
{
    displayData		*d, *d2;
    attrData		*aad, *ad;
    displayAttrData	*dad, *dad2;
    entryData		*en;
    valueData		*vd, *vd2;

    /*
     * In multiple entry windows allowed case, this isn't correct,
     * but we don't care then anyway.
     */
    entryDlgActive = False;

    /*
     * Find this entry in the linked list of on-screen entries and
     * unlink it.
     */
    if (dd == NULL) {
	return;
    }
    if (dd == displayDlgData) {
	/* Delete list head */
	displayDlgData = dd->di_next;
	d = dd;
    } else {
	d = displayDlgData;
	while ((d != NULL) && (d != dd)) {
	    d2 = d;
	    d = d->di_next;
	}
	if (d == NULL) {
	    errorDlg(toplevel, "No entryData found!");
	    return;
	}
	d2->di_next = d->di_next;
    }
    latestDisplayDlg = displayDlgData;
    en = d->di_entryData;
    dad = d->di_displayAttrData;
    vd = d->di_valueDataList;
    XtDestroyWidget(d->di_displayDlg);
    free(d);

    freeDisplayAttrDataList(dad);
    if (--(en->en_refcnt) == 0) {
	freeEntryData(en);
    }
    /* Clean up the valueData and associated data. */
    while (vd != NULL) {
	vd2 = vd;
	vd = vd->vd_next;
	free_valuedata(vd2);
    }
}




/*
 * Free a linked list of displayAttrData structs
 */
void
freeDisplayAttrDataList(dlist)
displayAttrData	*dlist;
{
    displayAttrData	*dad, *dad2;
    attrData		*ad;

    dad = dlist;
    while (dad != NULL) {
	dad2 = dad->dad_next;
	ad = dad->dad_attrData;
	/*
	 * This is a pretty gross hack, but...
	 * ad->ad_vals and dad->dad_osval might point to the
	 * same thing.  If so, we don't want to free anything
	 * here.
	 */
	if (!(LDAP_GET_SYN_TYPE(dad->dad_ti->ti_syntaxid) ==
			LDAP_SYN_TYPE_BOOLEAN) &&
		(dad->dad_osval != NULL) &&
		(ad->ad_vals != dad->dad_osval)) {
	    /*
	     * Only free dad_osval if it's not a boolean, and if
	     * it exists and isn't the same as ad->ad_vals
	     */
	    free(dad->dad_osval);
	}
	if (dad->dad_valueWidgetList) {
	    free(dad->dad_valueWidgetList);
	}
	free(dad);
	dad = dad2;
    }
}



void
freeAttrDataList(adlist)
attrData *adlist;
{
    attrData	*ad, *ad2;

    ad = adlist;
    while (ad != NULL) {
	ad2 = ad->ad_next;
	if (ad->ad_vals != NULL) {
	    ber_bvecfree(ad->ad_vals);
	}
	/*
	 * Note: don't free ad_attrname - it points to data in the
	 * templates area.
	 */
	free(ad);
	ad = ad2;
    }
}



void
freeEntryData(en)
entryData	*en;
{
    /*
     * Free the entryData struct
     */

    if (en->en_refcnt != 0) {
	fprintf(stderr, "Error: attempt to free entryData with refcnt > 0\n");
	return;
    }
    if (en->en_dn != NULL) {
	free(en->en_dn);
    }
    if (en->en_rdn != NULL) {
	free(en->en_rdn);
    }
    if (en->en_oclist != NULL) {
	ldap_value_free(en->en_oclist);
    }
    freeAttrDataList(en->en_attrData);
#ifdef KERBEROS
    if (en->en_krbNames) {
	ldap_value_free(en->en_krbNames);
    }
#endif /* KERBEROS */
    free(en);
}




/*
 * All drag/drop operations in a display window are vectored through
 * this routine.  It takes care of destroying the drag icon, and, if
 * the user dragged the icon out of the window, it takes care of
 * deleting that value from the display.
 */
static void
displayDragDropFinish(widget, client_data, cbs)
Widget				widget;
XtPointer			client_data;
XmDropFinishCallbackStruct	*cbs;
{
    Widget	source_icon = NULL;

    XtVaGetValues(widget, XmNsourceCursorIcon, &source_icon, NULL);

    if (source_icon != NULL) {
	XtDestroyWidget(source_icon);
    }
}



/*
 * Start a drag from a DN Button or a labeled URL button.  Creates the
 * drag icon, starts the drag operation, and arranges for
 * displayDragDropFinish() to be called upon completion of the drop.
 */
void
displayStartDrag(widget, event, params, num_params)
Widget		widget;
XEvent		*event;
String		*params;
Cardinal	*num_params;
{
    Arg		args[10];
    int		n, et;
    Widget	drag_source_icon, dc;
    Pixel	fg, bg;
    Pixmap	dsiconP, dsiconmskP;
    valueData	*vd;
    Atom	exportTargets[2], DN, TEXT, LABELED_URL;
    Display	*dpy;

    dpy = XtDisplay(toplevel);
    TEXT = XmInternAtom(dpy, "TEXT", False);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);

    XtVaGetValues(widget,
	    XmNbackground, &bg,
	    XmNforeground, &fg,
	    XmNuserData, &vd,
	    NULL);

    if (vd == NULL) {
	errorDlg(GetTopShell(widget), "Illegal drag source");
	return;
    }

    et = 0;
    if (vd->vd_tag == VD_DN) {
	/* Set up the drag icon to represent a DN */
	dsiconP = XmGetPixmapByDepth(XtScreen(widget),
		"dndragicon_image", 1, 0, 1);
	dsiconmskP = XmGetPixmapByDepth(XtScreen(widget),
		"dndragiconmsk_image", 1, 0, 1);
	if (dsiconP == XmUNSPECIFIED_PIXMAP ||
		dsiconmskP == XmUNSPECIFIED_PIXMAP) {
	    puts("bad pixmaps");
	    exit(1);	/* XXX fall back to default */
	}
	exportTargets[et] = DN; et++;
	/* exportTargets[et] = TEXT; et++; */
    } else if (vd->vd_tag == VD_URL) {
	/* Set up the drag icon to represent a URL */
	dsiconP = XmGetPixmapByDepth(XtScreen(widget),
		"urldragicon_image", 1, 0, 1);
	dsiconmskP = XmGetPixmapByDepth(XtScreen(widget),
		"urldragiconmsk_image", 1, 0, 1);
	if (dsiconP == XmUNSPECIFIED_PIXMAP ||
		dsiconmskP == XmUNSPECIFIED_PIXMAP) {
	    puts("bad pixmaps");
	    exit(1);	/* XXX fall back to default */
	}
	exportTargets[et] = LABELED_URL; et++;
	/* exportTargets[et] = TEXT; et++; */
    } else {
	errorDlg(widget,"Can't determine type of drag source");
    }
	

    n = 0;
    XtSetArg(args[n], XmNpixmap, dsiconP); n++;
    XtSetArg(args[n], XmNmask, dsiconmskP); n++;
    drag_source_icon = XmCreateDragIcon(widget, "drag_icon", args, n);
    n = 0;
    n = 0;
    XtSetArg(args[n], XmNblendModel, XmBLEND_ALL); n++;
    XtSetArg(args[n], XmNcursorBackground, bg); n++;
    XtSetArg(args[n], XmNcursorForeground, fg); n++;
    XtSetArg(args[n], XmNsourceCursorIcon, drag_source_icon); n++;
    XtSetArg(args[n], XmNdragOperations, XmDROP_COPY); n++;
    XtSetArg(args[n], XmNexportTargets, exportTargets); n++;
    XtSetArg(args[n], XmNnumExportTargets, et); n++;
    XtSetArg(args[n], XmNconvertProc, displayConvertProc); n++;
    XtSetArg(args[n], XmNclientData, vd); n++;

    dc = XmDragStart(widget, event, args, n);
    XtAddCallback(dc, XmNdragDropFinishCallback, displayDragDropFinish, NULL);
}






static Boolean
displayConvertProc(widget, selection, target, type_return, value_return,
	length_return, format_return)
Widget		widget;
Atom		*selection;
Atom		*target;
Atom		*type_return;
XtPointer	*value_return;
unsigned long	*length_return;
int		*format_return;
{
    Display	*dpy;
    valueData	*vd;
    Atom	TEXT, DN, LABELED_URL, MOTIF_DROP;

    dpy = XtDisplay(widget);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);
    TEXT = XmInternAtom(dpy, "TEXT", False);
    MOTIF_DROP = XmInternAtom(dpy, "_MOTIF_DROP", False);

    if (*selection != MOTIF_DROP) {
	return False;
    }

    /* get the drag source widget */
    XtVaGetValues(widget, XmNclientData, &vd, NULL);

    if (*target == DN) {
	*type_return = DN;
	*value_return = (XtPointer) vd;
	*length_return = sizeof(valueData *);
    } else if (*target == LABELED_URL) {
	*type_return = LABELED_URL;
	*value_return = (XtPointer) vd;
	*length_return = sizeof(valueData *);
    } else if (*target == TEXT) {
	*type_return = TEXT;
	*value_return = (XtPointer) vd->vd_value;
	*length_return = strlen((char *) vd->vd_value);
    } else {
	return False;
    }
    *format_return = 8;
    return True;
}





/*
 * This function is called for all drop destinations in entry windows.
 * It looks at the type of attribute being dropped on, decides if the
 * type of data being dropped is appropriate, and does appropriate
 * conversions, warnings, etc.
 */
static void
transferDrop(widget, client_data, seltype, type, value, length, format)
Widget		widget;
XtPointer	client_data;
Atom		*seltype;
Atom		*type;
XtPointer	value;
unsigned long	length;
int		format;
{
    Widget		dest, ddlg;
    displayData		*dd;
    displayAttrData	*dad;
    valueData		*vd;
    int			resp;
    Display		*dpy;
    Atom		TEXT, COMPOUND_TEXT, DN, LABELED_URL;
    static char		errstr[256];

    dest = (Widget) client_data;
    dpy = XtDisplay(GetTopShell(dest));
    TEXT = XmInternAtom(dpy, "TEXT", False);
    COMPOUND_TEXT = XmInternAtom(dpy, "COMPOUND_TEXT", False);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);

    /*
     * See if the drop site is a drawnButton.  If so, then we're handling
     * a drop on the trash icon.  Delete the value from the attribute
     * specified in the valueData structure (passed in as "value").
     *
     * Assuming all drops onto drawnButton widgets is probably a 
     * bad idea.
     */
    if (XmIsDrawnButton(dest)) {

	/* Drop on the trash icon */
	if (*type  == DN) {
	    if (value != NULL) {
		vd = (valueData *) value;
		syn_DeleteDN(vd->vd_displayAttrData, vd->vd_value);
	    }
	} else if (*type  == LABELED_URL) {
	    if (value != NULL) {
		vd = (valueData *) value;
		syn_DeleteURL(vd->vd_displayAttrData, vd->vd_value);
	    }
	}
	return;
    }

    /*
     * Handle a drop on something other than a drawnButton
     */
    
    /* Get the display dialog associated with the drop site */
    ddlg = GetTopShell(dest);

    for (dd = displayDlgData; dd != NULL; dd = dd->di_next) {
	if (ddlg == dd->di_displayDlg) {
	    break;
	}
    }
    if (dd == NULL) {
	fprintf(stderr, "error: no display dlg found for drop site\n");
	return;
    }

    /* Find the appropriate label */
    for (dad = dd->di_displayAttrData; dad != NULL; dad = dad->dad_next) {
	if (dest == dad->dad_labelWidget) {
	    break;
	}
    }
    if (dad == NULL) {
	fprintf(stderr, "error: no attr data found for drop site\n");
	return;
    }

    /* 
     * Based on the type of attribute being dropped on, perform the
     * appropriate action.
     */
    switch (dad->dad_ti->ti_syntaxid) {
    case LDAP_SYN_DN:
	/* Handle a drop on a DN label */
	if (*type == TEXT || *type == COMPOUND_TEXT) {
	    /* text was dropped on the label */
	    if ((strchr(value, '=') == NULL) || (strchr(value, ',') == NULL)) {
		resp = modalYesNo(GetTopShell(dest),
			"This doesn't appear to be a DN - continue?");
		if (resp == dlgNO) {
		    return;
		}
	    }
	    /*
	     * Disallow adding a group's own DN to its
	     * X.500 members attribute.
	     */
	    if (!okToAddValuetoAttribute(dd, dad, value)) {
		sprintf(errstr, "You may not add a group's own distinguished \nname to its %s attribute", dad->dad_ti->ti_label);
		errorDlg(ddlg, errstr);
		return;
	    }
	    if (syn_AddDN(dad, value) == AD_DUPVALUE) {
		char **xdn;
		xdn = ldap_explode_dn(value, 1);
		sprintf(errstr,
		    "Value \"%s\" already appears in the \"%s\" attribute", 
		    xdn[0], dad->dad_label);
		errorDlg(ddlg, errstr);
		ldap_value_free(xdn);
	    }
	} else if (*type == DN) {
	    /* a DN value was dropped on the label */
	    if (value != NULL) {
		vd = (valueData *) value;
		if (!okToAddValuetoAttribute(dd, dad, vd->vd_value)) {
		    sprintf(errstr, "You may not add a group's own distinguished \nname to its %s attribute", dad->dad_ti->ti_label);
		    errorDlg(ddlg, errstr);
		    return;
		}
		if (syn_AddDN(dad, vd->vd_value) == AD_DUPVALUE) {
		    char **xdn;
		    xdn = ldap_explode_dn(vd->vd_value, 1);
		    sprintf(errstr,
			"Value \"%s\" already appears in the \"%s\" attribute", 
			xdn[0], dad->dad_label);
		    errorDlg(ddlg, errstr);
		    ldap_value_free(xdn);
		}
	    }
	}
	break;
    case LDAP_SYN_LABELEDURL:
	/* Handle a drop on a labeled URL label */
	if (*type == TEXT || *type == COMPOUND_TEXT) {
	    /* text was dropped on the label */
	    if (strstr(value, "://") == NULL) {
		/* Doesn't look like a labeled URL */
		resp = modalYesNo(GetTopShell(dest),
		    "This doesn't appear to be a valid labeled URL - continue?");
		if (resp == dlgNO) {
		    return;
		}
	    }
	    if (syn_AddURL(dad, value) == AD_DUPVALUE) {
		sprintf(errstr,
		    "Value \"%s\" already appears in the \"%s\" attribute", 
		    (char *) value, dad->dad_label);
		errorDlg(ddlg, errstr);
	    }
	} else if (*type == LABELED_URL) {
	    /* a URL value was dropped on the label */
	    if (value != NULL) {
		vd = (valueData *) value;
		if (syn_AddURL(dad, vd->vd_value) == AD_DUPVALUE) {
		    sprintf(errstr,
			"Value \"%s\" already appears in the \"%s\" attribute", 
			(char *) vd->vd_value, dad->dad_label);
		    errorDlg(ddlg, errstr);
		}
	    }
	}
	break;
    default:
	errorDlg(GetTopShell(dest), "Unknown Drag and Drop destination type");
	break;
    }
    /* Check if we dirtied anything */
    updateDisplayDlgSaveButton(dest, -1);
}




static void
handleLabelDrop(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    Display			*dpy;
    XmDropProcCallback		DropData;
    Widget			dc;
    Boolean			isCompoundText, isText, isDN, isURL;
    XmDropTransferEntryRec	transferEntries[1];
    XmDropTransferEntry		transferList;
    Arg				args[10];
    int				n, i;
    Cardinal			numExportTargets;
    Atom			*exportTargets, TEXT, COMPOUND_TEXT,
				DN, LABELED_URL;
    displayAttrData	*dad;

/* XXX this is ugly - rewrite */
    dpy = XtDisplay(GetTopShell(w));
    DropData = (XmDropProcCallback) call_data;
    dc = DropData->dragContext;

    TEXT = XmInternAtom(dpy, "TEXT", False);
    COMPOUND_TEXT = XmInternAtom(dpy, "COMPOUND_TEXT", False);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_URL", False);

    n = 0;
    XtSetArg(args[n], XmNexportTargets, &exportTargets); n++;
    XtSetArg(args[n], XmNnumExportTargets, &numExportTargets); n++;
    XtGetValues(dc, args, n);

    isCompoundText = isText = isDN = isURL = False;
    for (i = 0; i < numExportTargets; i++) {
	if (exportTargets[i] == COMPOUND_TEXT) {
	    isCompoundText = True;
	} else if (exportTargets[i] == TEXT) {
	    isText = True;
	} else if (exportTargets[i] == DN) {
	    isDN = True;
	} else if (exportTargets[i] == LABELED_URL) {
	    isURL = True;
	}
    }

    dad = getDisplayAttrDataByLabelWidget(w);
    n = 0;
    /*
     * Check to see if the targets make sense, and if the
     * operation is something we can handle.
     */
    if ((!isCompoundText && !isText && !isDN && !isURL) ||
	    (DropData->dropAction != XmDROP) ||
	    (DropData->operation != XmDROP_COPY)) {
	/* no can do */
	XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE); n++;
	XtSetArg(args[n], XmNnumDropTransfers, 0); n++;
    /*
     * Make sure that the type of label dropped on makes sense
     * for the type of data being dropped.  I realize this is
     * duplicated code, but without it, the "if" above would
     * be hideous.
     */
    } else if ((isDN) && ((dad == NULL) || !isDNAttrData(dad))) {
	/* no can do */
	XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE); n++;
	XtSetArg(args[n], XmNnumDropTransfers, 0); n++;
    } else if ((isURL) && ((dad == NULL) || !isURLAttrData(dad))) {
	/* no can do */
	XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE); n++;
	XtSetArg(args[n], XmNnumDropTransfers, 0); n++;
    } else {
	if (isDN) {
	    transferEntries[0].target = DN;
	} else if (isURL) {
	    transferEntries[0].target = LABELED_URL;
	} else if (isCompoundText) {
	    transferEntries[0].target = COMPOUND_TEXT;
	} else if (isText) {
	    transferEntries[0].target = TEXT;
	}
	transferEntries[0].client_data = (XtPointer) w;
	transferList = transferEntries;
	XtSetArg(args[n], XmNdropTransfers, transferEntries); n++;
	XtSetArg(args[n], XmNnumDropTransfers,
		XtNumber(transferEntries)); n++;
	XtSetArg(args[n], XmNtransferProc, transferDrop); n++;
    }
    XmDropTransferStart(dc, args, n);
}
