/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * Portions Copyright 1991 O'Reilly & Associates, Inc.
 *
 */

/*
 * Structure passed to BuildMenu()
 */
typedef struct _menu_item {
    char		*widgetname;
    char		*label;
    WidgetClass		*class;
    char		mnemonic;
    char		*accelerator;
    char		*accel_text;
    void		(*callback)();
    caddr_t		callback_data;
    struct _menu_item	*subitems;
    Widget		*ret_widget;
} MenuItem;


/*
 * Main applications configuration options
 */
typedef struct {
    char *ldapServer;		/* hostname of our LDAP server */
    int  ldapPort; 		/* port for our LDAP server */
    char *ldapBindDN;		/* default DN to bind as upon startup */
    char *searchBase;		/* default search base */
    char *friendlyMap;		/* friendly name-mapping file */
    char *ldapFilterFile;	/* default LDAP filter file */
    int  ldapCacheSize;		/* size of LDAP cache in kb, 0 disables */
    int  ldapCacheTimeout;	/* LDAP cache timeout in minutes */
    int  ldapSearchTimeout;	/* timeout for LDAP searches, in seconds */
    int  ldapReadTimeout;	/* timeout for LDAP reads, in seconds */
    int  ldapWriteTimeout;	/* timeout for LDAP writes, in seconds */
    char *ldapTemplatesFile;	/* LDAP template configurations */
    char *ldapSearchprefsFile;	/* LDAP search preferences config */
    char *configFile;		/* user config file for LDAP attrs, etc. */
    char *wwwBrowser;		/* Program to fork for WWW Browsing */
    char *jpegViewer;		/* Program to view JPEG images */
    char *audioViewer;		/* Program to play audio data */
    int  maxSelLen;		/* Max length of strings in selecion dlg */
    Boolean	multipleBrowseWindows;
    Boolean	multipleEntryWindows;
    Boolean	moreChoices;	/* Start up w/more choices enabled */
    char *fewerButtonLabel;	/* "Fewer choices" button label */
    char *moreButtonLabel;	/* "More choices button label */
    char *lookWithinString;	/* "Look Within" preface for Searchbase */
    char *theWorldString;	/*  Used when searchbase is top of DIT */
    char *boundAsString;	/* "Bound As:" preface */
    char *authFilter;		/* filter to use for authentication searches */
    char *helpFile;
    char *helpFont;
    char *helpFontBold;
    char *helpFontItalic;
    char *helpFontBoldItalic;
    char *xax500NewsURL;
    char *msgSortByFullName;
    char *msgSortBySurname;
    /* error texts */
    char *emHelpTextNotFound;
    char *emHelpTextNotFoundDlgLabel;
    char *emBlanksNotAllowed;
    char *emSearchTimedOut;
    char *emBlankURLNotAllowed;
    char *emBadDNArgs;
    char *emSizeLimit;
    char *emTimeLimit;
} AppData;

typedef struct dlgData {
    Widget	status;		/* for updating parent dialog's status widget */
} dlgData;

typedef struct selectList {
    char	*dn;
    char	*surname;
    char	**objclasses;
    XmString	xmstr;
} selectList;

typedef struct selectData {
    Widget	status;		/* for updating parent dialog's status widget */
    Widget	okButton;	/* select dlg's ok button */
    Widget	sortButton;	/* select dlg's "sort by" button */ 
    Widget	selectionDlg;	/* Widget ID of dlg itself */
    Widget	selectionList;	/* Widget ID of scrolling list */
    unsigned long	options; /* select dlg options, e.g. PERSIST */
    char	**dn_list;	/* null-terminated list of DNs */
    char	***objcllist;	/* null-terminated list of objectClass lists */
    void	(*disp_fn)();	/* fn to call when user selects an item */
    void	*cdata;		/* client data, passed to disp_fn() */
    selectList	*selectionItems;	/* Array of info for list */
    int		count;		/* number of items in list */
    int		sort_mode;	/* current sort mode */
} selectData;

/*
 * Known object classes
 */
typedef struct objclassData {
    char	*name;
    char	icon;
} objclassData;


typedef struct findDataStruct {
    Widget		status;		/* for updating dialog status widget */
    char		*searchbase;	/* Current searchbase for this window */
    int			findObject;	/* index into ldap_searchobj list */
    struct ldap_searchobj *findObjectP;	/* ptr into ldap_searchobj list */
    int	findMatch;	/* index into list of match types */
    struct ldap_searchmatch *findMatchP;/* ptr into list of match types */
    int	findAttr;	/* index into list of search attrs */
    struct ldap_searchattr *findAttrP;	/* ptr into list of search attrs */
    int	setup;		/* bitfield of options */
#define	MORE_CHOICES	0x01
} findDataStruct;


/*
 * Available syntaxes
 */
typedef struct syntaxdata {
    char		*sd_name;
    struct berval	**(*sd_fromnative)();
    void		*(*sd_tonative)();
    void		*(*sd_display)();
    int			(*sd_addvalue)();
    int			(*sd_delvalue)();
    int			(*sd_vrfyvalue)();
} syntaxdata;


/*
 * Data structure associated with search operations, selection lists.
 */
typedef struct searchData {
    Widget		status;		/* for updating parent status widget */
    Widget		parent;		/* parent for any new widgets created */
    Widget		workingDlg;	/* id of working dialog if present */
    Widget		browseDlg;	/* browse Dialog */
    Widget		listWidget;	/* pointer to list widget */
    Widget		browseOption;	/* pointer to popup w/searchbases */
    int			msgid;		/* currently outstanding search id */
    time_t		start_time;	/* time search was started */
    int			ldaprc;		/* return code from last op */
    struct ldap_disptmpl *etype;	/* pointer to entry type */
    char		*objtype;	/* type of object, e.g. people */
    short		scope;		/* scope of search */
    XtIntervalId	timer;		/* Xt timer */
    char		*filterspec;	/* FIlter or filter tag to use */
    char		*filterprefix;	/* Filter prefix to use */
    LDAPFiltDesc	*ldfp;		/* array of search filters to use */
    LDAPFiltInfo	*currfilter;	/* array of search filters to use */
    char		*search_string;	/* string we're searching for */
    char		*searchbase;	/* searchbase for this find dialog */
    char		*newsb;		/* headed for this searchbase */
    char		**dn_list;	/* null-terminated list of DNs */
    char		**sattrs;	/* attributes searched for */
    char		**dattrs;	/* disambiguation attributes */
    char		**allattrs;	/* all attributes */
    char		*selectWidgetName;	/* X rsrc name of sel dlg */
    unsigned long	options;
#define	SRCH_OPT_PERSIST	1	/* select dlg should persist */
#define	SRCH_OPT_PEOPLEONLY	2	/* only search for people */
#define SRCH_OPT_LITERALFILT	4	/* filterspec gives literal filter */
    char		***objcllist;	/* null-term list of objectClasses */
    dlgData		*bd;		/* pointer to dlgData info */
    int			wantLeaves;	/* nonzero if user wants leaf nodes */
    void		(*disp_fn)();	/* fn to call when search done */	
    void		*cdata;		/* client data, passed to disp_fn() */
} searchData;

#define CHECK_SEARCH_INTERVAL 500 /* 1/2 second between checks */

/*
 * Linked list of apps we've forked.  Files we created can be cleaned up
 * when the app terminates.
 */
typedef struct forkedApps {
    pid_t		fa_pid;
    char		**fa_files;
    struct forkedApps	*fa_next;
} forkedApps;

typedef struct editInfo {
    Widget			parent;
    char			*data;
    struct displayAttrData	*dad;
    int				dirty;
} editInfo;


/*
 * Data structure for attribute information.  The contents
 * of this data structure are independent of any on-screen
 * representation.
 *
 * ad_attrname - X.500 name of the attribute
 * ad_vals     - values
 * ad_nvals    - number of values
 * ad_syndata  - pointer to syntax handler info
 * ad_next     - pointer to next attribute in list
 *
 */
typedef struct attrData {
    char		*ad_attrname;
    struct berval	**ad_vals;
    int			ad_nvals;
    syntaxdata		*ad_syndata;
    int			ad_flags;
#define	AD_DIRTY		1
#define	AD_STARTED_EMPTY	2
    struct attrData	*ad_next;
} attrData;



/*
 * Data structure which encapsulates all information about an attribute's
 * on-screen representation.  Does *not* contain the actual values, attribute
 * names, or syntax data - those are contained in the attrData struct, which
 * we keep a pointer to.  This allows a level of indirection - multiple
 * views of the same entry may be on-screen at the same time.
 */
typedef struct displayAttrData {
    char		*dad_label;
    attrData		*dad_attrData;
    Widget		dad_labelWidget;
    Widget		dad_valueWidget;	/* "top" value widget */
    WidgetList		dad_valueWidgetList;	/* individual values */
    int			dad_numValueWidgets;	/* num individual values */
    editInfo		*dad_editinfo;	/* ptr to editing window info */
    struct entryData	*dad_entryData;
    void		*dad_osval;		/* on-screen value */
    int			dad_buttonValIndex;	/* which val we're on */
    int			dad_col;
    int			dad_row;
    struct ldap_tmplitem *dad_ti;
    int			dad_flags;
#define	DAD_MAYBE_CHANGED	1
#define	DAD_NOSTRETCH		2
#define DAD_DEFER		4
    struct displayAttrData *dad_next;
} displayAttrData;
    


/*
 * Data structure for an X.500 entry.  When an entry is retrieved, an
 * entryData struct is allocated for it.  If another "view" of the entry
 * is opened, any additional entries are retrieved over ldap and
 * added to this struct.  *Note:* everything contained here is *independent*
 * of any on-screen representation of the entry.  When an entry is to
 * be written, or when one needs to know if an entry has been modified,
 * flags in this struct should be checked.
 *
 * en_dn       - Distinguished name of this entry
 * en_oclist   - Object class list for this entry
 * en_isnew    - True if entry in new and has not been saved
 * en_modified - True if entry has been modified but not saved
 * en_attrData - Linked list of attribute info
 * en_krbNames - Kerberos names for this entry (used in binding/auth)
 * en_refcnt   - reference count - number of windows open for this entry
 * 
 */
typedef struct entryData {
    char		*en_dn;
    char		*en_rdn;
    char		**en_oclist;
    Boolean		en_isnew;
    Boolean		en_modified;
    attrData		*en_attrData;
#ifdef KERBEROS
    char		**en_krbNames;
#endif /* KERBEROS */
    int			en_refcnt;
} entryData;


/*
 * Data structure associated with a display window:
 *
 * di_displayDlg           - Pointer to top-level dialog shell
 * di_status               - Label widget used for status updates, feedback
 *			     to user
 * di_displaySWin          - Scrolled window containing attributes and labels
 * di_displaySaveButton    - Save button for the dialog
 * di_displayDismissButton - Dismiss button for the dialog
 * di_displayDN            - Widget in upper-left containing the RDN of
 *                           the entry
 * di_currAttr             - Attribute currently being edited
 * di_prevAttr             - Attribute previously being edited
 * di_prevWidget           - Value widget of previously edited attribute
 * di_displayAttrData      - Linked list of displayAttriData (see below)
 * di_entryData            - entry Data for this entry
 * di_valueDataList        - linked list of valueData structs associated
 *                           with this view on the entry
 * di_closeOnWrite         - If True, close the window when the write operation
 *                           completes (e.g. the write operation was initiated
 *                           by the user closing an unsaved entry and responding
 *                           "Yes" to the "Save this entry?" dialog)
 * di_tmpl                 - pointer to disptmpl data for this window
 * di_nextDlgForEntry      - pointer to next displayData struct, if more than
 *                           one "view" of the entry is on-screen
 * di_next                 - pointer to next displayData struct (all
 *                           displayData structs are kept on a linked list)
 */
typedef struct displayData {
    Widget			di_displayDlg;
    Widget			di_status;
    Widget			di_displaySWin;
    Widget			di_displaySaveButton;
    Widget			di_displayDismissButton;
    Widget			di_displayAsOption;
    Widget			di_displayDN;
    displayAttrData		*di_currAttr;
    displayAttrData		*di_prevAttr;
    Widget			di_prevWidget;
    displayAttrData		*di_displayAttrData;
    entryData			*di_entryData;
    struct valueData		*di_valueDataList;
    Boolean			di_closeOnWrite;
    struct ldap_disptmpl	*di_tmpl;
    struct displayData		*di_nextDlgForEntry;
    struct displayData		*di_next;

    int				di_numMods;	/* XXX going away? */
} displayData;


typedef struct authTypeStruct {
    int		method;	/* auth method, e.g. LDAP_AUTH_KRBV41 */
    char	*data;	/* e.g. kerberos principal@realm */
} authTypeStruct;


typedef struct authStruct {
    int			ldaprc;		/* ret code from ldap operation */
    int			msgid;		/* msg id of async ldap op */
    int			timer;		/* time ldap op started */
    Widget		workingDlg;	/* widget ID of working dialog */
    Widget		authTextWidget;	/* widget ID of XXX ??? whaeezees? */
    Widget		parent;	
    char		*ident;
    char		*dn;		/* distinguished name to bind as */
    char		**dn_list;	/* null-term array of possible DNs */
    int			numDNs;		/* number of DNs found */
    void		(*callback)();	/* fn to call when we're all done */
    XtPointer		call_data;	/* call data to pass on */
    authTypeStruct	*authTypes;	/* array of authentication type info */
    authTypeStruct	*activeAuth;	/* Which auth type we're using */
    Boolean		sticky;		/* Don't override bound_as[] */
} authStruct;

#define AD_NOERR        0
#define AD_DUPVALUE     1
#define AD_NOTFOUND     2


/*
 * The valueData is used to squirrel away information about on on-screen
 * representation of "special" objects like DNs and URLs - mostly
 * for handling drag and drop operations.
 * vd_tag                - Type of value (DN, URL currently suported)
 * vd_displayAttrData    - Pointer to the displayAttrData
 * vd_displayData        - Pointer to display data
 * vd_value              - The value (typ. RDN, or label of URL)
 * vd_next               - All valueData structs are kept on a list, so they can
 *                         be freed when the window is closed.
 */
typedef struct valueData {
    unsigned char	vd_tag;
#define VD_DN		1	/* valueData is for a DN */
#define VD_URL		2	/* valueData is for a labeledURL */
#define VD_NOTATTR	99	/* valueData is not associated with an attr */
    displayAttrData	*vd_displayAttrData;
    displayData		*vd_displayData;
    char		*vd_value;
    struct valueData	*vd_next;
} valueData;

typedef struct actionData {
    int                 ac_type;
    displayData         *ac_dd;
    char                *ac_disableattr;
    void                *ac_args;
} actionData;

