/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * Portions Copyright 1991 O'Reilly & Associates, Inc.
 *
 */

/*
 * Code which implements edit dialogs for DNs, URLs, and maybe other
 * things eventually.
 */


#include <Xm/Text.h>
#include <Xm/LabelG.h>
#include <Xm/RowColumn.h>
#include <Xm/SelectioB.h>

#include <stdio.h>
#include <malloc.h>

#include <lber.h>
#include <ldap.h>

#include "dtypes.h"
#include "globals.h"
#include "util.h"
#include "dialog.h"

typedef struct URLeditInfo {
    Widget	labelWidget;
    Widget	URLWidget;
    char	*label;
    char	*url;
} URLeditInfo;

/*
 * Free memory allocated for an editInfo struct.
 */
freeEditInfo(ei)
editInfo	*ei;
{
    if (ei == NULL) {
	return;
    }

    free(ei);
}



/*
 * Collect the text from all the text widgets, weed out any duplicates,
 * and update the entry's data structure.
 */
void
URLeditDone(w, ei, cbs)
Widget			w;
editInfo		*ei;
XmAnyCallbackStruct	*cbs;
{
printf("URLeditDone\n");
    freeEditInfo(ei);
}
	



void
editCancel(w, ei, cbs)
Widget			w;
editInfo		*ei;
XmAnyCallbackStruct	*cbs;
{
    ei->dad->dad_editinfo = NULL;
    XtDestroyWidget(GetTopShell(w));
    freeEditInfo(ei);
}




void
URLeditResponse(w, client_data, call_data)
Widget		w;
XtPointer	client_data;
XtPointer	call_data;
{
    int					*answer = (int *) client_data;
    XmSelectionBoxCallbackStruct	*cbs;
    URLeditInfo				*uei;

    cbs = (XmSelectionBoxCallbackStruct *) call_data;

    switch (cbs->reason) {
    case XmCR_OK:
	XtVaGetValues(w, XmNuserData, &uei, NULL);
	uei->label = XmTextGetString(uei->labelWidget);
	uei->url = XmTextGetString(uei->URLWidget);
	*answer = dlgOK;
	break;
    case XmCR_CANCEL:
	*answer = dlgCANCEL;
	break;
    default:
	;
    }
}


void
createURLEditDialog(w, ad, label, url)
Widget		w;
attrData	*ad;
char		**label;
char		**url;
{
    Widget	URLeditDlg, URLeditRC, URLeditLabelString, URLeditLabel;
    Arg		args[3];
    int		n, answer;
    URLeditInfo	*uei;

    uei = (URLeditInfo *) malloc(sizeof(URLeditInfo));
    n = 0;
    XtSetArg(args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
    XtSetArg(args[n], XmNuserData, uei); n++;
    URLeditDlg = XmCreatePromptDialog(GetTopShell(w), "URLeditDlg", args, n);

    URLeditRC = XtVaCreateWidget("URLeditRC", xmRowColumnWidgetClass,
	    URLeditDlg, NULL);
    URLeditLabelString = XtVaCreateWidget("URLeditLabelString",
	    xmLabelGadgetClass, URLeditRC, NULL);
    URLeditLabel = XtVaCreateWidget("URLeditLabel",
	    xmTextWidgetClass, URLeditRC, NULL);
    uei->labelWidget = URLeditLabel;
    uei->URLWidget = XmSelectionBoxGetChild(URLeditDlg, XmDIALOG_TEXT);

    XtManageChild(URLeditLabelString);
    XtManageChild(URLeditLabel);
    XtManageChild(URLeditRC);
    XtManageChild(URLeditDlg);
    XtPopup(XtParent(URLeditDlg), XtGrabNone);

    XtAddCallback(URLeditDlg, XmNokCallback, URLeditResponse, &answer);
    XtAddCallback(URLeditDlg, XmNcancelCallback, URLeditResponse, &answer);
    answer = 0;
    SetFocus(URLeditRC, URLeditLabel);

    while (answer == 0) {
	XtAppProcessEvent(app, XtIMAll);
    }

    if (answer == dlgOK) {
	*label = uei->label;
	*url = uei->url;
    } else {
	*label = *url = NULL;
    }
    XtDestroyWidget(URLeditDlg);
    free(uei);
}
