/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Support for find dialog, LDAP searching
 */

#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/SeparatoG.h>
#include <Xm/Label.h>
#include <Xm/SelectioB.h>

#include <lber.h>
#include <ldap.h>

#include <stdio.h>
#include <sys/time.h>

#include "dtypes.h"
#include "util.h"
#include "srchpref.h"
#include "menu.h"
#include "globals.h"
#include "find.h"
#include "browse.h"
#include "bitmaps.h"
#include "dialog.h"
#include "read.h"
#include "display.h"

/* forward references */
#ifndef NEEDPROTOS
void setFindObject();
Widget createSearchAttrPopup();
Widget createSearchMatchPopup();
Widget createFindObjectPopup();
void toggleFindSetup();
void updateSearchBaseText();
void updateBoundAsText();
void startSearch();
void crText();
void updateFB();
#else /* !NEEDPROTOS */
void setFindObject(Widget w, int item_no, XmAnyCallbackStruct *cbs);
Widget createSearchAttrPopup(Widget parent, MenuItem *items, char *label);
Widget createSearchMatchPopup(Widget parent, MenuItem *items, char *label);
Widget createFindObjectPopup(Widget parent, XtPointer callback,
	    struct ldap_searchobj *sel);
void toggleFindSetup(Widget w, XtPointer client_data,
	    XmPushButtonCallbackStruct *cbs);
void updateSearchBaseText(char *sb);
void updateBoundAsText(char *ba);
void startSearch(Widget w, XtPointer client_data, XmAnyCallbackStruct *cbs);
void crText(Widget w, XtPointer client_data, XmAnyCallbackStruct *cbs);
void updateFB(Widget w, XtPointer client_data, XmAnyCallbackStruct *cbs);
#endif /* !NEEDPROTOS */

static Widget	mainMenubar, searchObjectPopup, searchAttrPopup,
		searchMatchPopup, mainSB, mainSep1, mainSep2,
		mainAction, mainSearchLabel, mainSearchText,
		mainBoundAs = NULL, mainFindButton, mainMCButton, mainForm;

static MenuItem	findObjectMenuItems[MAX_MENU_ITEMS];
static MenuItem	findMatchMenuItems[MAX_MENU_ITEMS];
static MenuItem	findAttrMenuItems[MAX_MENU_ITEMS];

static ActionAreaItem   mainActions[] = {
	{NULL, NULL, NULL, NULL},
	{NULL, NULL, NULL, NULL},
};

static char findErrstr[1024], *sf_prefix = NULL, *sf_dattr = NULL,
	    *sf_tag = NULL, *sf_objtype = NULL;
static struct ldap_searchobj *sf_prevsetting = NULL;



/*
 * Create the app's main find window.
 */
findDataStruct *
createFindWindow(parent)
Widget parent;
{
    findDataStruct		*f;
    Widget			browseButton;

    if ((f = (findDataStruct *) malloc(sizeof(findDataStruct))) == NULL) {
	fprintf(stderr, "out of memory allocating findDataStruct\n");
	exit(1);
    }
    memset(f, 0, sizeof(findDataStruct));
    f->searchbase = strdup(search_base);
    f->findObjectP = searchprefs;
    f->findMatchP = searchprefs->so_smlist;
    f->findMatch = 0;
    f->findAttrP = searchprefs->so_salist;
    f->findAttr = 0;

    mainForm = XtVaCreateWidget("findWindow", xmFormWidgetClass,
	    mainPane,
	    NULL);
    create_menus(&mainForm, &mainMenubar);

    XtVaSetValues(mainMenubar,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    NULL);
	
    /* Create the browseButton */
    browseButton = create_pixmap_button(mainForm, "world_image",
	    NULL, "browseButton");
    XtVaSetValues(browseButton,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainMenubar,
	    NULL);
    XtAddCallback(browseButton, XmNactivateCallback,
	    (XtCallbackProc) do_browse, NULL);
    XtManageChild(browseButton);

    searchObjectPopup = createFindObjectPopup(mainForm, setFindObject, NULL);
#ifdef notyet
    /* Create the FindObject popup */
    for (so = searchprefs, i = 0; so != NULL; so=so->so_next, i++) {
	findObjectMenuItems[i].widgetname = 
	findObjectMenuItems[i].label = so->so_objtypeprompt;
	findObjectMenuItems[i].class = &xmPushButtonGadgetClass;
	findObjectMenuItems[i].mnemonic = (char) NULL;
	findObjectMenuItems[i].accelerator = NULL;
	findObjectMenuItems[i].accel_text = NULL;
	findObjectMenuItems[i].callback = setFindObject;
	findObjectMenuItems[i].callback_data = (XtPointer) i;
	findObjectMenuItems[i].subitems = NULL;
	findObjectMenuItems[i].ret_widget = NULL;
    }
    findObjectMenuItems[i].widgetname = NULL;
    findObjectMenuItems[i].subitems = NULL;
    searchObjectPopup = BuildMenu(mainForm, XmMENU_OPTION,
	    "searchObjectOption",
	    NULL, findObjectMenuItems, 0);
    XtVaSetValues(searchObjectPopup,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainMenubar,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNleftOffset, 5,
	    XmNnavigationType, XmTAB_GROUP,
	    NULL);
#endif /* notyet */
    searchAttrPopup = createSearchAttrPopup(mainForm, findAttrMenuItems,
	    "searchAttrPopup");
    XtVaSetValues(searchAttrPopup,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainMenubar,
	    XmNleftAttachment, XmATTACH_WIDGET,
	    XmNleftWidget, searchObjectPopup,
	    XmNleftOffset, 5,
	    XmNnavigationType, XmTAB_GROUP,
	    NULL);
    searchMatchPopup = createSearchMatchPopup(mainForm, findMatchMenuItems, "");
    XtVaSetValues(searchMatchPopup,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNnavigationType, XmTAB_GROUP,
	    NULL);


    /*
     * For "fewer choices," we unmanage the searchMatchPopup menu
     * and manage this simple label.
     */
		
    mainSearchLabel = XtVaCreateWidget("mainSearchLabel",
	    xmLabelGadgetClass, mainForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNleftOffset, 5,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);

    /* Create editable text area for user to type in */
    mainSearchText = XtVaCreateWidget("mainSearchText",
	    xmTextFieldWidgetClass, mainForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNrightOffset, 5,
	    XmNnavigationType, XmTAB_GROUP,
	    NULL);
    XtAddCallback(mainSearchText, XmNactivateCallback,
	    (XtCallbackProc) crText, NULL);
    XtAddCallback(mainSearchText, XmNvalueChangedCallback,
	    (XtCallbackProc) updateFB, NULL);

    /* separator */
    mainSep1 = XtVaCreateWidget("mainSep1", xmSeparatorGadgetClass,
	    mainForm,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainSearchText,
	    XmNtopOffset, 10,
	    XmNbottomOffset, 10,
	    NULL);
					   
    /* Search base area */
    mainSB = XtVaCreateWidget("mainSB", xmLabelWidgetClass,
	    mainForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainSep1,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    XmNrightOffset, 5,
	    XmNleftOffset, 5,
	    XmNtopOffset, 5,
	    XmNbottomOffset, 5,
	    NULL);
    updateSearchBaseText(f->searchbase);

    /* separator */
    mainSep2 = XtVaCreateWidget("mainSep2", xmSeparatorGadgetClass,
	    mainForm,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainSB,
	    XmNtopOffset, 10,
	    XmNbottomOffset, 10,
	    NULL);
					   
    /* "Bound as:" area */
    mainBoundAs = XtVaCreateWidget("mainBoundAs", xmLabelWidgetClass,
	    mainForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainSep2,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    XmNrightOffset, 5,
	    XmNleftOffset, 5,
	    XmNtopOffset, 5,
	    XmNbottomOffset, 5,
	    NULL);
    updateBoundAsText(bound_as);


    init_pixmaps();
    XtVaGetValues(mainForm,
	    XmNforeground, &fg,
	    XmNbackground, &bg,
	    NULL);

    /* Create an action area */
    mainActions[0].label = "findButtonLabel";
    mainActions[0].callback = startSearch;
    mainActions[1].callback = toggleFindSetup;
	
    /* More choices or not? */
    if (appConfig.moreChoices) {
	f->setup |= MORE_CHOICES;
    }
    if (f->setup & MORE_CHOICES) {
	mainActions[1].label = appConfig.fewerButtonLabel;
    } else {
	mainActions[1].label = appConfig.moreButtonLabel;
    }
    mainActions[0].widget = &mainFindButton;
    mainActions[1].widget = &mainMCButton;
    mainAction = CreateActionArea(parent, mainActions, XtNumber(mainActions));

    XtVaSetValues(mainFindButton,
	    XmNshowAsDefault, True,
	    NULL);
    XtSetSensitive(mainFindButton, False);

    XtManageChild(mainMenubar);
    XtManageChild(searchObjectPopup);
    if (f->setup & MORE_CHOICES) {
	/* Set alignment resources */
	XtVaSetValues(mainSearchText,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, searchMatchPopup,
		NULL);
	XtManageChild(searchAttrPopup);
	XtManageChild(searchMatchPopup);
    } else {
	XtVaSetValues(mainSearchText,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, mainSearchLabel,
		NULL);
	XtManageChild(mainSearchLabel);
    }
    XtVaSetValues(mainForm, XmNinitialFocus, mainSearchText, NULL);
    XtVaSetValues(mainSearchText,
	    XmNresizeWidth, True,
	    XmNuserData, mainAction,
	    NULL);
    XtManageChild(mainSearchText);
    XtManageChild(mainSep1);
    XtManageChild(mainSB);
    XtManageChild(mainSep2);
    XtManageChild(mainBoundAs);
    XtManageChild(mainAction);
    XtManageChild(mainForm);

    return(f);
}




Widget
createFindObjectPopup(parent, callback, sel)
Widget			parent;
XtPointer		callback;
struct ldap_searchobj	*sel;
{
    int i, sel_item;
    struct ldap_searchobj *so;

    /* Create the FindObject popup */
    for (so = searchprefs, i = 0, sel_item = 0; so != NULL;
	    so=so->so_next, i++) {
	if (so == sel) {
	    sel_item = i;
	}
	findObjectMenuItems[i].widgetname = 
	findObjectMenuItems[i].label = so->so_objtypeprompt;
	findObjectMenuItems[i].class = &xmPushButtonGadgetClass;
	findObjectMenuItems[i].mnemonic = (char) NULL;
	findObjectMenuItems[i].accelerator = NULL;
	findObjectMenuItems[i].accel_text = NULL;
	findObjectMenuItems[i].callback = (void *) callback;
	findObjectMenuItems[i].callback_data = (XtPointer) i;
	findObjectMenuItems[i].subitems = NULL;
	findObjectMenuItems[i].ret_widget = NULL;
    }
    findObjectMenuItems[i].widgetname = NULL;
    findObjectMenuItems[i].subitems = NULL;
    searchObjectPopup = BuildMenu(parent, XmMENU_OPTION,
	    "searchObjectOption",
	    NULL, findObjectMenuItems, sel_item);
    XtVaSetValues(searchObjectPopup,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainMenubar,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNleftOffset, 5,
	    XmNnavigationType, XmTAB_GROUP,
	    NULL);
    return(searchObjectPopup);
}




Widget 
createSearchAttrPopup(parent, items, label)
Widget		parent;
MenuItem	*items;
char		*label;
{
    struct ldap_searchattr	*sa;
    struct ldap_searchobj	*so;
    int				i;

    if (searchprefs == NULL) {
	errorDlg(parent, "Bad search preferences (check config file).");
	return(NULL);
    }
    for (i = 0, so = searchprefs; so && (i < findData->findObject);
	     i++, so = so->so_next);
    if (!so) {
	sprintf(findErrstr, "No attr list for that search object!");
	errorDlg(mainForm, findErrstr);
	items[0].label = NULL;
	items[0].widgetname = NULL;
	items[0].subitems = NULL;
	return(BuildMenu(parent, XmMENU_OPTION, label, NULL, items, 0));
    }

    for (sa = so->so_salist, i = 0; sa != NULL; sa = sa->sa_next, i++) {
	items[i].widgetname = items[i].label = sa->sa_attrlabel;
	items[i].class = &xmPushButtonGadgetClass;
	items[i].mnemonic = (char) NULL;
	items[i].accelerator = NULL;
	items[i].accel_text = NULL;
	items[i].subitems = NULL;
	items[i].ret_widget = NULL;
	items[i].callback = setFindAttr;
	items[i].callback_data = (XtPointer) i;
    }
    items[i].widgetname = NULL;
    items[i].subitems = NULL;
    return(BuildMenu(parent, XmMENU_OPTION, label, NULL, items, 0));
}



Widget
createSearchMatchPopup(parent, items, label)
Widget		parent;
MenuItem	*items;
char		*label;
{
    struct ldap_searchmatch	*sm;
    struct ldap_searchobj	*so;
    struct ldap_searchattr	*sa;
    int				i, j;

    if (searchprefs == NULL) {
	errorDlg(parent, "Bad search preferences (check config file).");
	return(NULL);
    }

    for (i = 0, so = searchprefs; so && (i < findData->findObject);
	     i++, so = so->so_next);
    if (!so) {
	sprintf(findErrstr, "Bad search object index");
	errorDlg(mainForm, findErrstr);
	goto bad;
    }

    for (sa = so->so_salist, i = 0; sa && (i < findData->findAttr);
	     i++, sa = sa->sa_next);
    if (!sa) {
	sprintf(findErrstr, "Bad attribute list index");
	errorDlg(mainForm, findErrstr);
	goto bad;
    }
    
    for (sm = so->so_smlist, i = 0, j = 0; sm != NULL; sm = sm->sm_next, i++) {
	if (sa->sa_matchtypebitmap & (0x1 << i)) {
	    items[j].widgetname = items[j].label = sm->sm_matchprompt;
	    items[j].class = &xmPushButtonGadgetClass;
	    items[j].mnemonic = (char) NULL;
	    items[j].accelerator = NULL;
	    items[j].accel_text = NULL;
	    items[j].subitems = NULL;
	    items[j].ret_widget = NULL;
	    items[j].callback = setFindMatch;
	    items[j].callback_data = (XtPointer) j;
	    j++;
	}
    }

    items[j].widgetname = NULL;
    items[j].subitems = NULL;
    return(BuildMenu(parent, XmMENU_OPTION, label, NULL, items, 0));

bad: items[0].label = NULL;
    items[0].subitems = NULL;
    return(BuildMenu(parent, XmMENU_OPTION, label, NULL, items, 0));
}


void
updateSearchAttrPopup()
{
    Widget	parent, new;
    char	*buf;

    /* XXX - updating the menu in place would be far better */
    parent = XtParent(searchAttrPopup);
    new = createSearchAttrPopup(parent, findAttrMenuItems, "searchAttrOption");
    XtVaSetValues(new,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, mainMenubar,
	    XmNleftAttachment, XmATTACH_WIDGET,
	    XmNleftWidget, searchObjectPopup,
	    NULL);
    XtDestroyWidget(searchAttrPopup);
    searchAttrPopup = new;
    XtManageChild(searchAttrPopup);

    /* Update the match type popup menu */
    /* XXX do in place if possible */
    parent = XtParent(searchMatchPopup);
    new = createSearchMatchPopup(parent, findMatchMenuItems,
	    "searchMatchOption");
    XtVaSetValues(new,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNnavigationType, XmTAB_GROUP,
	    NULL);
    XtDestroyWidget(searchMatchPopup);
    searchMatchPopup = new;
    XtManageChild(searchMatchPopup);

    buf = XmTextGetString(mainSearchText);
    XtDestroyWidget(mainSearchText);
    /* Re-create editable text area for user to type in */
    mainSearchText = XtVaCreateWidget("mainSearchText",
	    xmTextFieldWidgetClass, parent,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNrightOffset, 5,
	    XmNuserData, mainAction,
	    NULL);
    XtAddCallback(mainSearchText, XmNactivateCallback,
	    (XtCallbackProc) crText, NULL);
    XtAddCallback(mainSearchText, XmNvalueChangedCallback,
	    (XtCallbackProc) updateFB, NULL);
    XmTextSetString(mainSearchText, buf);
    XtFree(buf);
    if (findData->setup & MORE_CHOICES) {
	/* Set alignment resources */
	XtVaSetValues(mainSearchText,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, searchMatchPopup,
		NULL);
    } else {
	XtVaSetValues(mainSearchText,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, mainSearchLabel,
		NULL);
    }
    XtManageChild(mainSearchText);
    XmProcessTraversal(mainSearchText, XmTRAVERSE_CURRENT);
}
		
		

		
		


/*
 * Called when a selection is made from the "Search For:" option popup
 * menu in the main find dialog.  This selects the type of object (people,
 * groups, etc.) we're searching for.
 */
void
setFindObject(w, item_no, cbs)
Widget			w;
int			item_no;
XmAnyCallbackStruct	*cbs;
{
    struct ldap_searchobj	*p;
    int				i;

    /*
     * Don't do anything if the user simply re-selected the
     * same menu item.
     */
    if (findData->findObject == item_no) {
	return;
    }

    for (p = searchprefs, i = 0; p  && (i < item_no); p = p->so_next, i++);
    if (p) {
	findData->findObject = item_no;
	findData->findObjectP = p;
    }

    /*
     * If we're in "fewer choices," we're done.  Otherwise,
     * set the attribute and match type popups to appropriate
     * values.
     */
    findData->findAttr = findData->findMatch = 0; /* XXX remember prev? */
    findData->findAttrP = searchprefs->so_salist;
    findData->findMatchP = searchprefs->so_smlist;
    if (findData->setup & MORE_CHOICES) {
	updateSearchAttrPopup();
    }
    XmProcessTraversal(mainSearchText, XmTRAVERSE_CURRENT);
}
	

/*
 * Called when a selection is made from the attribute option popup
 * menu in the main find dialog.  This selects the attribute by which
 * we'll do the search.
 */
void
setFindAttr(w, item_no, cbs)
Widget			w;
int			item_no;
XmAnyCallbackStruct	*cbs;
{
    Widget			parent, newMenu;
    struct ldap_searchattr	*p;
    int				i;

    /*
     * Don't do anything if the user simply re-selected the
     * same menu item.
     */
    if (findData->findAttr == item_no) {
	return;
    }

    for (p = findData->findObjectP->so_salist, i = 0; p && (i < item_no);
	     p = p->sa_next, i++);
    if (p) {
	findData->findAttr = item_no;
	findData->findAttrP = p;
	findData->findMatchP = findData->findObjectP->so_smlist;
	findData->findMatch = 0;
    }
    /* Update the match type popup menu */
    /* XXX do in place if possible */
    parent = XtParent(searchMatchPopup);
    newMenu = createSearchMatchPopup(parent, findMatchMenuItems,
	    "searchMatchOption");
    XtVaSetValues(newMenu,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNleftAttachment, XmATTACH_FORM,
	    NULL);
    XtDestroyWidget(searchMatchPopup);
    searchMatchPopup = newMenu;
    XtManageChild(searchMatchPopup);
    XmProcessTraversal(mainSearchText, XmTRAVERSE_CURRENT);
}



/*
 * Called when a selection is made from the match type popup
 * menu in the main find dialog.  This selects the type of match
 * (exact, approximate, ends with) we'll allow when searching
 */
void
setFindMatch(w, item_no, cbs)
Widget			w;
int			item_no;
XmAnyCallbackStruct	*cbs;
{
    struct ldap_searchmatch	*p;
    int				i;

    p = findData->findObjectP->so_smlist;
    i = 0;
    while (p) {
	if (!(findData->findAttrP->sa_matchtypebitmap & (0x1 << i))) {
	    item_no++;
	}
	if (i >= item_no) {
	    break;
	}
	i++;
	p = p->sm_next;
    }
    if (p) {
	findData->findMatch = item_no;
	findData->findMatchP = p;
    }
    XmProcessTraversal(mainSearchText, XmTRAVERSE_CURRENT);
}



void
toggleFindSetup(w, client_data, cbs)
Widget				w;
XtPointer			client_data;
XmPushButtonCallbackStruct	*cbs;
{
    XmString	s;
    char	*buf;
    Widget	parent = XtParent(mainSearchText);
    caddr_t	dat;

	
    findData->setup ^= MORE_CHOICES;
    /* preserve text and userData resource */
    buf = XmTextGetString(mainSearchText);
    XtVaGetValues(mainSearchText, XmNuserData, &dat, NULL);
    XtDestroyWidget(mainSearchText);
    mainSearchText = XtVaCreateWidget("mainSearchText",
	    xmTextFieldWidgetClass, parent,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, searchObjectPopup,
	    XmNnavigationType, XmTAB_GROUP,
	    XmNtraversalOn, True,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNrightOffset, 5,
	    XmNuserData, dat,
	    NULL);
    XmTextSetString(mainSearchText, buf);
    XtFree(buf);
    XtAddCallback(mainSearchText, XmNactivateCallback,
	    (XtCallbackProc)crText, NULL);
    XtAddCallback(mainSearchText, XmNvalueChangedCallback,
	    (XtCallbackProc) updateFB, NULL);

    if (findData->setup & MORE_CHOICES) {
	/* Enable more choices setup */
	XtUnmanageChild(mainSearchLabel);
	XtManageChild(searchAttrPopup);
	XtVaSetValues(searchAttrPopup,
		XmNtraversalOn, True,
		XmNnavigationType, XmTAB_GROUP,
		NULL);
	XtManageChild(searchMatchPopup);
	XtVaSetValues(searchMatchPopup,
		XmNtraversalOn, True,
		XmNnavigationType, XmTAB_GROUP,
		NULL);
	XtVaSetValues(mainSearchText,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, searchMatchPopup,
		NULL);
	XtManageChild(mainSearchText);
	s = XmStringCreateSimple(appConfig.fewerButtonLabel);
	XtVaSetValues(w,
		XmNlabelString, s,
		NULL);
	XmStringFree(s);
#ifdef notdef
	findData->findObjectP = searchprefs;
	findData->findAttrP = searchprefs->so_salist;
	findData->findAttr = 0;
	findData->findMatchP = searchprefs->so_smlist;
	findData->findMatch = 0;
#endif
	updateSearchAttrPopup();
    } else {
	XtUnmanageChild(searchAttrPopup);
	XtUnmanageChild(searchMatchPopup);
	XtManageChild(mainSearchLabel);
	XtVaSetValues(mainSearchText,
		XmNleftAttachment, XmATTACH_WIDGET,
		XmNleftWidget, mainSearchLabel,
		NULL);
	XtManageChild(mainSearchText);
	s = XmStringCreateSimple(appConfig.moreButtonLabel);
	XtVaSetValues(w,
		XmNlabelString, s,
		NULL);
	XmStringFree(s);
#ifdef notdef
	findData->findObjectP = searchprefs;
#endif
    }
    SetFocus(mainForm, mainSearchText);
}



void
startSearch(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    char	*t;
    char	*dattrs[2];
    char	*sattrs[3];
    char	*filtertag, *filterprefix;

    t = XmTextGetString(mainSearchText);
    if (*t) {
	dattrs[0] = ((findData->setup & MORE_CHOICES) &&
		*(findData->findAttrP->sa_selectattr)) ?
		findData->findAttrP->sa_selectattr :
		findData->findObjectP->so_defaultselectattr;
	dattrs[1] = NULL;
	sattrs[0] = "objectClass";
	sattrs[1] = "sn";
	sattrs[2] = NULL;
	if (findData->setup & MORE_CHOICES) {
	    filtertag = filterprefix = NULL;
	} else {
	    filtertag = findData->findObjectP->so_filtertag;
	    filterprefix = findData->findObjectP->so_filterprefix;
	}
	do_search(w, t, sattrs, dattrs, filtertag, filterprefix,
		findData->findObjectP->so_objtypeprompt,
		do_read, "selectDlg", 0 /* SRCH_OPT_PERSIST */, NULL);
	XtFree(t);
    }
}
	


void
crText(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    XtCallActionProc(mainFindButton, "ArmAndActivate", cbs->event, NULL, 0);
}



/*
 * If there are now characters in the editable text field, enable the
 * find button.  If not, disable it.
 */
void
updateFB(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    char	*s;

    if (*(s = XmTextGetString(mainSearchText))) {
	XtSetSensitive(mainFindButton, True);
	XtVaSetValues(mainFindButton,
		XmNshowAsDefault, True,
		NULL);
	XtVaSetValues(mainAction,
		XmNdefaultButton, mainFindButton,
		NULL);
    } else {
	XtSetSensitive(mainFindButton, False);
	XtVaSetValues(mainFindButton,
		XmNshowAsDefault, False,
		NULL);
	XtVaSetValues(mainAction,
		XmNdefaultButton, mainMCButton,
		NULL);
    }
    XtFree(s);
}




void
updateSearchBaseText(sb)
char	*sb;
{
    XmString	s;
    static char	buf[1024];
    char	*b;

    b = (sb ? sb : search_base);
    if (findData->searchbase != NULL) {
	free(findData->searchbase);
    }
    findData->searchbase = strdup(b);
    sprintf(buf, "%s: %s", appConfig.lookWithinString, *b ? b :
	    appConfig.theWorldString);
    s = XmStringCreateSimple(buf);
    XtVaSetValues(mainSB, XmNlabelString, s, NULL);
    XmStringFree(s);
}


void
updateBoundAsText(ba)
char	*ba;
{
    XmString	s;
    static char	buf[1024];
    char	*b;

    if (!mainBoundAs) {
	return;
    }
    b = (ba ? ba : bound_as);
    sprintf(buf, "%s: %s", appConfig.boundAsString, *b ? b : "(not bound)");
    s = XmStringCreateSimple(buf);
    XtVaSetValues(mainBoundAs, XmNlabelString, s, NULL);
    XmStringFree(s);
}



void
simpleSetFindObject(w, item_no, cbs)
Widget			w;
int			item_no;
XmAnyCallbackStruct	*cbs;
{
    struct ldap_searchobj	*p;
    int				i;

    for (p = searchprefs, i = 0; p  && (i < item_no); p = p->so_next, i++);
    if (p) {
	sf_prefix = p->so_filterprefix;
	sf_dattr = p->so_defaultselectattr;
	sf_tag = p->so_filtertag;
	sf_objtype = p->so_objtypeprompt;
	sf_prevsetting = p;
    }
}






/*
 * Callback routine to get text typed by user in a simple find dialog.
 */
void
sf_getresponse(w, client_data, call_data)
Widget          w;
XtPointer       client_data;
XtPointer       call_data;
{
    char				**value = (char **) client_data;
    XmSelectionBoxCallbackStruct	*cbs;

    cbs = (XmSelectionBoxCallbackStruct *) call_data;

    if (cbs->reason == XmCR_CANCEL) {
	*value = "";
    } else {
	XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, value);
    }
}



/*
 * "Simple" find dialog - intended to be used for adding DNs to 
 * entries and for (in the future) authenticating.
 *
 * Returns: DN found, or NULL if none found or cancelled.
 */
char *
do_simple_find(parent, dlgName, modality, people_only, fn, cdata)
Widget		parent;
char		*dlgName;
unsigned char	modality;
Boolean		people_only;	/* limit search to people, e.g. when binding */
XtPointer	fn;		/* function to call when done */
void		*cdata;		/* argument to be passed to fn() */
{
    Widget		simpleFindDlg, findObjectPopup;
    char		*value;
    char		swname[32];
    unsigned long	options = 0L;
    char		*dattrs[2];
    char		*sattrs[3];

    simpleFindDlg = XmCreatePromptDialog(parent, dlgName, NULL, 0);
    XtVaSetValues(simpleFindDlg,
	    XmNdialogStyle, modality,
	    NULL);
    XtAddCallback(simpleFindDlg, XmNokCallback, sf_getresponse, &value);
    XtAddCallback(simpleFindDlg, XmNcancelCallback, sf_getresponse, &value);
    value = NULL;

    sprintf(swname, "%s%s", dlgName, "Status");
#ifdef notdef
    statusArea = XtVaCreateManagedWidget(swname, xmLabelWidgetClass,
	    simpleFindDlg, NULL);
#endif

    /*
     * If the "people_only" flag was set, e.g. when authenticating
     * just pre-set the search configuration.
     */
    if (people_only) {
	options |= SRCH_OPT_PEOPLEONLY;
	sf_prefix = "(&objectClass=Person)";
    } else {
	/*
	 * Remember the previous setting of the simple-find search prefs,
	 * if there was any.
	 */
	if (sf_prevsetting != NULL) {
	    sf_prefix = sf_prevsetting->so_filterprefix;
	    sf_dattr = sf_prevsetting->so_defaultselectattr;
	    sf_tag = sf_prevsetting->so_filtertag;
	    sf_objtype = sf_prevsetting->so_objtypeprompt;
	    findObjectPopup = createFindObjectPopup(simpleFindDlg,
		    simpleSetFindObject, sf_prevsetting);
	} else {
	    sf_prefix = searchprefs->so_filterprefix;
	    sf_dattr = searchprefs->so_defaultselectattr;
	    sf_tag = searchprefs->so_filtertag;
	    sf_objtype = searchprefs->so_objtypeprompt;
	    findObjectPopup = createFindObjectPopup(simpleFindDlg,
		    simpleSetFindObject, NULL);
	}
	XtManageChild(findObjectPopup);
    }

    XtManageChild(simpleFindDlg);
    XtPopup(XtParent(simpleFindDlg), XtGrabNone);

    while (value == NULL) {
	XtAppProcessEvent(app, XtIMAll);
    }

    if (strlen(value) > 0) {
	if (is_dn(value)) {
	    /* Verify that the DN exists */
	} else if (is_ufn(value)) {
	    /* Do UFN search */
	} else {
	    /* Do a standard search */
	    dattrs[0] = sf_dattr;
	    dattrs[1] = NULL;
	    sattrs[0] = "objectClass";
	    sattrs[1] = "sn";
	    sattrs[2] = NULL;
	    do_search(parent, value, sattrs, dattrs, sf_tag, sf_prefix, 
		    sf_objtype, fn, "simpleFindSelectDlg", options, cdata);
	}
    }
    XtDestroyWidget(simpleFindDlg);
    return(value);
}

