/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */


#include <lber.h>
#include <ldap.h>

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>

#include "dtypes.h"
#include "syn.h"
#include "help.h"
#include "sys/time.h"

/*
 * Globals
 */
char			*appname;		/* name of app, from argv[0] */
Widget			toplevel;
Widget			mainPane;
Widget			workingDlg;
Widget			browseDlg;
Widget			helpDlg;
Widget			aboutDlg = NULL;
Boolean			browserActive = False;
Boolean			entryDlgActive = False;
XtWorkProcId		work_proc_id;
findDataStruct		fdat;
findDataStruct		*findData = &fdat;
struct ldap_disptmpl	*templates;
struct ldap_searchobj	*searchprefs;
Boolean			WWWViewerActive = False;
pid_t			WWWpid = -1;
struct forkedApps	*forkedAppsList = NULL;
FriendlyMap		*fr_map;
Pixel			fg, bg;
LDAP			*ldp;
char 			search_base[1024];
char			bound_as[1024];
Boolean			sticky_bind = False;
XtAppContext		app;
struct timeval		zerotimest = {0L, 0L};
struct timeval		*zerotime = &zerotimest;
AppData			appConfig;
int			linit();
XtTranslations		parsedDragTranslations;
XtTranslations		parsedNoDragTranslations;
helpNode		*helpTree;
int			numAddableTemplates = 0; /* num "addable" templates */

/* 
 * XXX replace this with code to get LDAP host, identitiy,
 * etc, from X resource or a command-line option.
 */
#define get_ldap_host()			appConfig.ldapServer
#define default_ldap_credentials()	NULL
#define default_ldap_auth_method()	LDAP_AUTH_SIMPLE

syntaxdata syntaxes[]   = {
/*
 * "native" means native to the client
 * Name,
 * Function to convert from client-internal format to bervals
 * Function to convert from bervals to client-internal format
 * Function to display values
 * Function to add a value to the displayed values
 * Function to delete a value from the displayed values
 * Function to verify that a given value is valid
 */
/* 0*/  {"CaseIgnoreString", syn_TextToCIS, syn_CISToText, syn_DisplayText,
	NULL, NULL, NULL},
/* 1*/  {"MultiLineString", syn_TextToMLS, syn_MLSToText, syn_DisplayText,
	NULL, NULL, NULL},
/* 2*/  {"DN", syn_FromDN, syn_ToDN, syn_DisplayDN,
	syn_AddDN, syn_DeleteDN, NULL},
/* 3*/  {"audio", syn_FromAudio, syn_ToAudio, syn_DisplayAudio,
	NULL, NULL, NULL},
/* 4*/  {"jpeg", syn_FromJpeg, syn_ToJpeg, syn_DisplayJpeg, 
	NULL, NULL, NULL},
/* 5*/	{"Boolean", syn_FromBoolean, syn_ToBoolean, syn_DisplayBoolean,
	NULL, NULL, NULL},
/* 6*/	{"LabeledURL", syn_FromURL, syn_ToURL, syn_DisplayURL,
	NULL, NULL, NULL},
/* 7*/	{"Time", syn_FromTime, syn_ToTime, syn_DisplayText,
	NULL, NULL, NULL},
/* 8*/	{"Date", syn_FromDate, syn_ToDate, syn_DisplayText,
	NULL, NULL, NULL},
/* 9*/  {"rfc822addr", syn_TextToRFC822, syn_RFC822ToText, syn_DisplayText,
	NULL, NULL, syn_VerifyRFC822},
/*10*/  {"searchAction", NULL, NULL, syn_DisplaySearchAction,
	NULL, NULL, NULL},
/*11*/  {"linkAction", NULL, NULL, syn_DisplayLinkAction,
	NULL, NULL, NULL},
/*12*/  {"addDNAction", NULL, NULL, syn_DisplayAddDNAction,
	NULL, NULL, NULL},
/*13*/  {"verifyDNAction", NULL, NULL, syn_DisplayVerifyDNAction,
	NULL, NULL, NULL},
/*14*/  {"MultiLineString1val", syn_TextToMLS1val, syn_MLSToText,
	syn_DisplayText, NULL, NULL, NULL},
/*15*/  NULL
};


struct objclassData objClasses[] = {
    {"person", 'Z'},
    {"organizationalRole", 'Y'},
    {"organizationalUnit", 'x'},
    {"organization", 'M'},
    {"friendlyCountry", 'g'},
    {"state", 'k'},
    {"locality", 'l'},
    {"aDMD", 'X'},
    {"organizationalPerson", 'Z'},
    {NULL, (char) NULL}
};
