/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <lber.h>
#include <ldap.h>

#include <Xm/Form.h>
#include <Xm/DialogS.h>
#include <Xm/PushBG.h>
#include <Xm/List.h>
#include <Xm/PanedW.h>
#include <Xm/Label.h>
#include <Xm/Protocols.h>
#include <Xm/Text.h>
#include <Xm/ScrolledW.h>
#include <Xm/RowColumn.h>


#include "dtypes.h"
#include "globals.h"
#include "util.h"
#include "help.h"
#include "menu.h"
#include "display.h"
#include "Hyper.h"

Widget			helpDlg = NULL;
static Widget		helpHyper;
static Widget		helpPrevButton;
static Widget		helpTopicLabel;
static helpNode		*helpHistory[32];
static int		maxHelpHistory = 0;

static	helpNode	*helpTree = NULL;

#ifndef NEEDPROTOS
int		insertHelpNode();
void		print_help_nodes();
static void	freeHelpTree();
#else /* !NEEDPROTOS */
int		insertHelpNode(helpNode **top, helpNode *new);
void		print_help_nodes(helpNode *hn, FILE *fp);
static void	freeHelpTree(helpNode *top);
#endif /* !NEEDPROTOS */

/*
 * Read a help file and construct a binary tree of section headings.
 */
int
parse_help_file(file)
char	*file;
{
#define	SECTION_HEADING	"SECTION"
#define	ALIAS_HEADING	"ALIAS"
#define	HTBUFSIZ	1024
#define MAXHELPALIASES	4

    FILE	*fp;
    char	buf[HTBUFSIZ], *p, *aliases[MAXHELPALIASES + 1];
    long	pos, newpos, prevpos;
    helpNode	*hnode, *prevhnode, *aliasnode;
    int		i, naliases;

    if ((fp = fopen(file, "r")) == NULL) {
	return(1);
    }

    memset(aliases, 0, sizeof(char *) * (MAXHELPALIASES + 1));
    prevpos = pos = 0L;
    hnode = NULL;
    while (fgets(buf, HTBUFSIZ, fp) != NULL) {
	if (!strncmp(buf, SECTION_HEADING, strlen(SECTION_HEADING))) {
	    /* Found a section heading */
	    prevhnode = hnode;
	    hnode = (helpNode *) malloc(sizeof(helpNode));
	    hnode->left = hnode->right = NULL;
	    hnode->length = -1;
	    hnode->offset = pos + strlen(buf);
	    hnode->text = NULL;
	    /* Get section heading */
	    p = buf + strlen(SECTION_HEADING) + 1;
	    while (*p && isspace(*p)) {
		p++;
	    }
	    if (!*p) {
		errorDlg(toplevel, "Missing SECTION tag in help file\n");
		return(2);
	    }
	    p[strlen(p) - 1] = '\0';	/* Get rid of \n */
	    hnode->heading = strdup(p);
	    if (insertHelpNode(&helpTree, hnode)) {
		errorDlg(toplevel,
			"Duplicate help section title in help file\n");
		return(3);
	    }
	    newpos = pos;
	    if (prevhnode != NULL) {
		prevhnode->length = newpos - prevpos;
		/* add nodes for any alises */
		for (i = 0; aliases[i] != NULL; i++) {
		    aliasnode = (helpNode *) malloc(sizeof(helpNode));
		    memcpy(aliasnode, prevhnode, sizeof(helpNode));
		    aliasnode->heading = aliases[i];
		    aliasnode->left = aliasnode->right = NULL;
		     if (insertHelpNode(&helpTree, aliasnode)) {
			errorDlg(toplevel,
				"Duplicate help section title in help file\n");
			return(3);
		    }
		}
		memset(aliases, 0, sizeof(char *) * (MAXHELPALIASES + 1));
		naliases = 0;
	    }
	    prevpos = hnode->offset;
	} else if (!strncmp(buf, ALIAS_HEADING, strlen(ALIAS_HEADING)) &&
		(naliases < MAXHELPALIASES)) {
	    /* Found an alias heading */
	    p = buf + strlen(ALIAS_HEADING) + 1;
	    while (*p && isspace(*p)) {
		p++;
	    }
	    if (!*p) {
		errorDlg(toplevel, "Missing LINK tag in help file\n");
		return(2);
	    }
	    p[strlen(p) - 1] = '\0';	/* Get rid of \n */
	    aliases[naliases++] = strdup(p);
	    pos = ftell(fp);
	    hnode->offset = pos;
	    prevpos = pos;
	} else {
	    pos = ftell(fp);
	}
    }
    newpos = pos;
    if (hnode != NULL) {
	hnode->length = newpos - prevpos;
    }
    /* ick, duplicated code */
    for (i = 0; aliases[i] != NULL; i++) {
	aliasnode = (helpNode *) malloc(sizeof(helpNode));
	memcpy(aliasnode, hnode, sizeof(helpNode));
	aliasnode->heading = aliases[i];
	aliasnode->left = aliasnode->right = NULL;
	 if (insertHelpNode(&helpTree, aliasnode)) {
	    errorDlg(toplevel,
		    "Duplicate help section title in help file\n");
	    return(3);
	}
    }
    /* print_help_nodes(helpTree, fp); */
    return(0);
}


int
insertHelpNode(top, new)
helpNode	**top;
helpNode	*new;
{
    int	i;


    if (*top == NULL) {
	*top = new;
	return(0);
    }


    i = strcasecmp(new->heading, (*top)->heading);
    if (i < 0) {
	return(insertHelpNode(&((*top)->left), new));
    } else if (i > 0) {
	return(insertHelpNode(&((*top)->right), new));
    } else {
	return(-1);	/* duplicate section heading */
    }
}


void
print_help_text(hn, fp)
helpNode	*hn;
FILE		*fp;
{
    int		nb, rb;
    char	buf[512];

    rewind(fp);
    fseek(fp, hn->offset, 0);
    nb = (int) hn->length;
    while (nb > 0) {
	fgets(buf, nb, fp);
	printf("<%s>\n", buf);
	if ((rb = strlen(buf)) == 0) {
	    break;
	}
	nb -= rb;
    }
}




void
freeHelp()
{
    freeHelpTree(helpTree);
}




static void
freeHelpTree(top)
helpNode	*top;
{
    if (top == NULL) {
	return;
    }
    freeHelpTree(top->left);
    freeHelpTree(top->right);
    if (top->heading != NULL) {
	free(top->heading);
    }
    if (top->text != NULL) {
	free(top->text);
    }
    free(top);
}




helpNode *
find_help_node(section, top)
char		*section;
helpNode	*top;
{
    int		i;

    if (top == NULL) {
	return(NULL);
    }

    i = strcasecmp(section, top->heading);
    if (i < 0) {
	return(find_help_node(section, top->left));
    } else if (i > 0) {
	return(find_help_node(section, top->right));
    } else {
	return(top);
    }
}



/*
 * Given a section name, search the help tree for a matching node.
 * If ret_hnode is non-null, return a pointer to the help node found,
 * or NULL if not found.  Returns NULL if no matching node found.
 */
char *
get_help_text(section, ret_hnode)
char		*section;
helpNode	**ret_hnode;
{
    helpNode	*hn;
    int		fd, nb;

    if ((hn = find_help_node(section, helpTree)) == NULL) {
	if (ret_hnode != NULL) {
	    *ret_hnode = NULL;
	}
	return(NULL);
    } else {
	if (hn->text == NULL) {
	    if ((fd = open(appConfig.helpFile, O_RDONLY)) < 0) {
		return(NULL);
	    }
	    lseek(fd, hn->offset, SEEK_SET);
	    hn->text = (char *) malloc(hn->length + 1);
	    memset(hn->text, 0, hn->length + 1);
	    hn->text[hn->length - 1] = '\0';
	    nb = read(fd, hn->text, hn->length);
	    if (nb != hn->length) {
		free(hn->text);
		hn->text = NULL;
	    }
	}
	close(fd);
	if (ret_hnode != NULL) {
	    *ret_hnode = hn;
	}
	return(hn->text);
    }
}



/*
 * For debugging - descend the tree inorder and print all the nodes.
 */
void
print_help_nodes(hn, fp)
helpNode	*hn;
FILE		*fp;
{
    if (hn == NULL) {
	return;
    }
    print_help_nodes(hn->left, fp);
    printf("\nNODE: %s (offset %d, %d bytes)\n",
	    hn->heading, hn->offset, hn->length);
    print_help_text(hn, fp);
    print_help_nodes(hn->right, fp);
}




/*
 * Unmanage the help dialog
 */
void
hide_help(w)
Widget	w;
{
    XtUnmanageChild(helpDlg);
}




void helpClose(w)
Widget	w;
{
    XtUnmapWidget(helpDlg);
}



void
addHelpHistory(hn)
helpNode	*hn;
{
    helpHistory[++maxHelpHistory] = hn;
    XtSetSensitive(helpPrevButton, True);
}



void
removeHelpHistory()
{
    maxHelpHistory--;
    if (maxHelpHistory == 0) {
	XtSetSensitive(helpPrevButton, False);
    }
}



void
setHelpTopicLabel(hn)
helpNode	*hn;
{
    XmString	str;

    if (hn == NULL) {
	/* no heading found - totle the dialog appropriately */
	str = XmStringCreateLocalized(appConfig.emHelpTextNotFoundDlgLabel);
    } else {
	str = XmStringCreateLocalized(hn->heading);
    }
    XtVaSetValues(helpTopicLabel, XmNlabelString, str, NULL);
    XmStringFree(str);
}




void
updateHelpDlg(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
hyperCallbackStruct	*cbs;
{
    char	*buf;
    helpNode	*hn;

    buf = get_help_text(cbs->text, &hn);
    if (buf == NULL) {
	/* Can't find appropriate help text */
	buf = appConfig.emHelpTextNotFound;
	maxHelpHistory++; /* XXX this fixes things up so prevTopic works */
    }
    HyperSetText(w, buf);
    if (hn != NULL) {
	addHelpHistory(hn);
	setHelpTopicLabel(hn);
    }
}



void
helpPrev(w, hyper, cbs)
Widget			w;
Widget			hyper;
XmAnyCallbackStruct	*cbs;
{
    if (maxHelpHistory < 1) {
	return;
    }
    removeHelpHistory();
    HyperSetText(hyper, (helpHistory[maxHelpHistory])->text);
    setHelpTopicLabel(helpHistory[maxHelpHistory]);
}





void
doHelpDlg(w, section)
Widget	w;
char	*section;
{
    Widget			helpPane, helpDismissButton,
				helpActionArea, helpSWin;
    static ActionAreaItem	helpActions[3];
    Atom			wm_delete_window;
    int				fd, rc;
    struct stat			stbuf;
    char			*buf;
    Arg				args[10];
    int				n;
    helpNode			*hn;
	
    if (helpDlg != NULL) {
	maxHelpHistory = 0;
	XtSetSensitive(helpPrevButton, False);
	/* Already have a help dialog - just update and map it */
	buf = get_help_text(section, &hn);
	if (buf == NULL) {
	    /* Can't find appropriate help text */
	    buf = appConfig.emHelpTextNotFound;
	}
	if (hn != NULL) {
	    setHelpTopicLabel(hn);
	}
	HyperSetText(helpHyper, buf);
	XtMapWidget(helpDlg);
	return;
    }

    helpDlg = XtVaAppCreateShell("helpDlg", "Xax500",
	    applicationShellWidgetClass,
	    XtDisplay(toplevel),
	    NULL);
    wm_delete_window = XmInternAtom(XtDisplay(helpDlg),
	    "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(helpDlg, wm_delete_window, helpClose, NULL);

    helpPane = XtVaCreateManagedWidget("helpPane",
	    xmPanedWindowWidgetClass,
	    helpDlg,
	    XmNsashWidth, 1,
	    XmNsashHeight, 1,
	    NULL);
    
    helpTopicLabel = XtVaCreateManagedWidget("helpTopicLabel",
	    xmLabelWidgetClass,
	    helpPane,
	    NULL);
    
    if ((rc = stat(appConfig.helpFile, &stbuf)) < 0) {
	buf = "No help file found";
    } else {
	buf = (char *) malloc(stbuf.st_size + 1);
	fd = open(appConfig.helpFile, O_RDONLY);
	read(fd, buf, stbuf.st_size);
	buf[stbuf.st_size] = '\0';
    }

    n = 0;
    XtSetArg(args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
    helpSWin = XmCreateScrolledWindow(helpPane, "helpSWin", args, n);
    XtManageChild(helpSWin);

    helpHyper = CreateHyper(helpSWin, "helpHyper", NULL, 0);
    XtAddCallback(helpHyper, XtNactivateCallback, updateHelpDlg, NULL);
    XtManageChild(helpHyper);

    buf = get_help_text(section, &hn);
    setHelpTopicLabel(hn);

    if (buf == NULL) {
	/* Can't find appropriate help text */
	buf = appConfig.emHelpTextNotFound;
    } else {
	helpHistory[maxHelpHistory] = hn;
    }
    HyperSetText(helpHyper, buf);

    helpActions[0].label = "helpPrevButton";
    helpActions[0].data = (caddr_t) helpHyper;
    helpActions[0].widget = &helpPrevButton;
    helpActions[0].callback = helpPrev;

    helpActions[1].label = "helpDismissButton";
    helpActions[1].data = (caddr_t) helpDlg;
    helpActions[1].widget = &helpDismissButton;
    helpActions[1].callback = helpClose;

    helpActionArea = CreateActionArea(helpPane, helpActions, 2);
    XtSetSensitive(helpPrevButton, False);

    XtRealizeWidget(helpDlg);
    XtMapWidget(helpDlg);
}


void
retrieveNews(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    openURL(w, appConfig.xax500NewsURL);
}
