/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <Xm/PushBG.h>
#include <Xm/LabelG.h>
#include <Xm/SeparatoG.h>
#include <Xm/List.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/DrawnB.h>
#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>
#include <Xm/PanedW.h>

#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#include <stdio.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>

#include "srchpref.h"
#include "dtypes.h"
#include "globals.h"
#include "menu.h"
#include "bitmaps.h"
#include "prefs.h"
#include "find.h"
#include "fallbacks.h"
#include "auth.h"
#include "util.h"
#include "resources.h"
#include "dialog.h"

static Boolean	ldap_init_started = False;

/*
 * Forward references
 */
#ifndef NEEDPROTOS
static void	parseAdditionalArgs();
static int	linit();
static void	quit_application();
#else /* !NEEDPROTOS */
static void	parseAdditionalArgs(int argc, char **argv);
static int	linit(Widget w);
static void	quit_application();
#endif /* !NEEDPROTOS */



main(argc, argv)
int	argc;
char	**argv;
{
    Widget	startupDlg;
    XmString	s;
    char	*init_string = "xax.500: Initializing...";
    Arg		args[10];
    int		i;

    appname = DEFAULT_APP_NAME;

    toplevel = XtAppInitialize(&app, "Xax500", options, XtNumber(options),
	    &argc, argv, fallbacks, 
	    NULL, 0);

    parseAdditionalArgs(argc, argv);

    XtAddEventHandler(toplevel, StructureNotifyMask, False,
	    (XtEventHandler) configureSize, NULL);

    XtVaGetApplicationResources(toplevel, (XtPointer) &appConfig,
	    xresources, XtNumber (xresources), NULL);

    strcpy(search_base, appConfig.searchBase);

    mainPane = XtVaCreateWidget("mainForm", xmPanedWindowWidgetClass,
	    toplevel,
	    XmNsashHeight, 1,
	    XmNsashWidth, 1,
	    NULL);

    init_pixmaps();
    initAtoms();
    initTranslations();
    XtVaGetValues(mainPane,
	    XmNforeground, &fg,
	    XmNbackground, &bg,
	    NULL);

    s = XmStringCreateSimple(init_string);
    i = 0;
    XtSetArg(args[i], XmNmessageString, s); i++;
    XtSetArg(args[i], XmNtitle, appname); i++;
    startupDlg = XmCreateWorkingDialog(mainPane, "startupDlg", args, i);
    XtUnmanageChild(XmMessageBoxGetChild(startupDlg, XmDIALOG_OK_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(startupDlg, XmDIALOG_CANCEL_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(startupDlg, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(startupDlg, XmDIALOG_SEPARATOR));
    XmStringFree(s);

    XtAppAddWorkProc(app, (XtWorkProc) linit, startupDlg);

    SetIcon(toplevel, "app_image");

    /* arrange to reap forked children */
    signal(SIGCHLD, reap_child);

    XtAppMainLoop(app);
}



static int
linit(w)
Widget	w;
{
#define	STARTUP_DIALOG_LATENCY 3 
    int			rc;
    time_t		start, elapsed;
    char		errtxt[128];
    char		*errs[10];
    XmString		s, stmp, scopy, sep;
    static int		num_errs = 0;
    static int		is_fatal = False;
    int			i;
    Widget		eDlg;
    LDAPFiltDesc	*ldfp;

    /* First time through, pop up the working dialog */
    if (!ldap_init_started) {
	ldap_init_started = 1;
	XtManageChild(w);
	XtPopup(XtParent(w), XtGrabExclusive);
	ForceDialog(w);
	return(False);	/* False -> we'll be called again */
    }
    while (XtAppPending(app)) {
	XtAppProcessEvent(app, XtIMAll);
    }

    /* Misc initialization */
    latestDisplayDlg = NULL;

    /* LDAP stuff */
    start = time(NULL);
    /* check ldap filter file */
    if ((ldfp = ldap_init_getfilter(appConfig.ldapFilterFile)) == NULL) {
	sprintf(errtxt, "Cannot open ldap filter config file \"%s\"",
		appConfig.ldapFilterFile);
	errs[num_errs++] = strdup(errtxt);
    } else {
#ifndef LDAP_GETFILTER_FREE_BROKEN
	ldap_getfilter_free(ldfp);
#endif /* LDAP_GETFILTER_FREE_BROKEN */
    }
    strcpy(search_base, appConfig.searchBase);

    if ((rc = init_ldap(&ldp)) != LDAP_SUCCESS) {
	sprintf(errtxt, "Error contacting LDAP server %s: (%s)",
		appConfig.ldapServer, ldap_err2string(rc));
	errs[num_errs++] = strdup(errtxt);
	is_fatal = True;
    } else {
	/* ldp->ld_deref = LDAP_DEREF_SEARCHING; */
	ldp->ld_deref = LDAP_DEREF_FINDING;
    }

    if (ldap_friendly_name(appConfig.friendlyMap, " ", NULL) == NULL) {
	sprintf(errtxt, "Cannot open friendly map file \"%s\"",
		appConfig.friendlyMap);
	errs[num_errs++] = strdup(errtxt);
    }

    /* Open and parse the help file */
    if ((rc = parse_help_file(appConfig.helpFile)) != 0) {
	sprintf(errtxt, "Cannot open help file \"%s\" (error %d)",
		appConfig.helpFile, rc);
	errs[num_errs++] = strdup(errtxt);
    }

    /* Get the template configuration file */
    if (ldap_init_templates(appConfig.ldapTemplatesFile, &templates)) {
	sprintf(errtxt, "Cannot open template file \"%s\"",
		appConfig.ldapTemplatesFile);
	errs[num_errs++] = strdup(errtxt);
	is_fatal = True;
    } else {
	if (fixup_templates(templates) < 0) {
	    sprintf(errtxt, "Bad template file \"%s\"",
		    appConfig.ldapTemplatesFile);
	    errs[num_errs++] = strdup(errtxt);
	    is_fatal = True;
	}
    }

    /* Get the search preferences */
    if (ldap_init_searchprefs(appConfig.ldapSearchprefsFile, &searchprefs)) {
	sprintf(errtxt, "Cannot open search preferences file \"%s\"",
		appConfig.ldapSearchprefsFile);
	errs[num_errs++] = strdup(errtxt);
	is_fatal = True;
    }

    while ((elapsed = (time(NULL) - start)) < STARTUP_DIALOG_LATENCY) {
	XmUpdateDisplay(w);
    }
    XtDestroyWidget(w);

    if (!is_fatal) {
	/* If everything's ok, bring up the find window and start */
	findData = createFindWindow(mainPane);
	XtManageChild(mainPane);
	XtRealizeWidget(toplevel);
    }

    if (num_errs) {
	sep = XmStringSeparatorCreate();
	s = XmStringSegmentCreate(
		"The following errors occured during initialization:",
		XmSTRING_DEFAULT_CHARSET,
		XmSTRING_DIRECTION_L_TO_R,
		True);
	scopy = XmStringCopy(s);
	s = XmStringConcat(scopy, sep);
	XmStringFree(scopy);

	for (i = 0; i < num_errs; i++) {
	    stmp = XmStringSegmentCreate(errs[i],
		    XmSTRING_DEFAULT_CHARSET,
		    XmSTRING_DIRECTION_L_TO_R,
		    True);
	    scopy = XmStringCopy(s);
	    s = XmStringConcat(scopy, stmp);
	    XmStringFree(stmp);
	    XmStringFree(scopy);
	}

	scopy = XmStringCopy(s);
	s = XmStringConcat(scopy, sep);
	XmStringFree(scopy);

	stmp = XmStringSegmentCreate(
		"Check your app-defaults file and X Resources",
		XmSTRING_DEFAULT_CHARSET,
		XmSTRING_DIRECTION_L_TO_R,
		True);
	scopy = XmStringCopy(s);
	s = XmStringConcat(scopy, stmp);
	XmStringFree(stmp);
	XmStringFree(scopy);
	XmStringFree(sep);

	eDlg = createDlg(toplevel, XmDIALOG_ERROR, "errorDlg",
		"",
		is_fatal ? NULL : "Continue", NULL,
		"Quit", quit_application,
		NULL, NULL,
		XmDIALOG_FULL_APPLICATION_MODAL);
	XtVaSetValues(eDlg, XmNmessageString, s, NULL);
	XmStringFree(s);
	for (i = 0; i < num_errs; i++) {
	    free(errs[i]);
	}
	XtManageChild(eDlg);
	XtPopup(XtParent(eDlg), XtGrabNone);
    }

    return(True);
}




static void
quit_application()
{
    cleanup_and_exit();
}



void
usage()
{
static char *usagetext[] = {
"",
"-ldapserver ldap-server",
"\twhere \"ldap-server\" is the name of the LDAP server to use",
"",
"-port port-number",
"\twhere \"port-number\" is the TCP port to use when",
"\tcontacting the LDAP server",
"",
"-binddn initial-bind-DN",
"\twhere \"initial-bind-DN\" is the DN to bind as upon startup",
"",
"-sb initial-searchbase",
"\twhere \"initial-searchbase\" is the searchbase to use upon startup",
"",
"-fm friendly-map-file",
"\twhere \"friendly-map-file\" is the name of the LDAP friendly map file",
"",
"-filt ldap-filter-config-file",
"\twhere \"ldap-filter-config-file\" is the name of the LDAP filter",
"\tconfiguration file",
"",
"-template ldap-templates-config-file",
"\twhere \"ldap-filter-config-file\" is the name of the LDAP display",
"\ttemplate configuration file",
"",
"-searchprefs ldap-searchprefs-config-file",
"\twhere \"ldap-filter-config-file\" is the name of the LDAP search",
"\tpreferences configuration file",
"",
"-helpfile help-file-name",
"\twhere \"help-file-name\" is the name of the xax500 help file",
"",
"-cachesize ldap-cachesize",
"\twhere \"ldap-cachesize\" is the size of the LDAP cache (in kbytes)",
"",
"-cachetime cache-timeout",
"\twhere \"cache-timeout\" is the cache timeout (in minutes)",
"",
"-mc",
"\tstart up in \"More Choices\" mode",
"",
"-sto search-timeout",
"\twhere \"search-timeout\" is the timeout value for search operations",
"\t(in seconds)",
"",
"-rto read-timeout",
"\twhere \"read-timeout\" is the timeout value for read operations",
"\t(in seconds)",
"",
"-wto write-timeout",
"\twhere \"write-timeout\" is the timeout value for write operations",
"\t(in seconds)",
NULL,
};
    int	i;

    printf("usage: %s [options] where \"options\" is one or more of:\n",
	    appname);
    for (i = 0; usagetext[i] != NULL; i++) {
	puts(usagetext[i]);
    }
}

static void
parseAdditionalArgs(argc, argv)
int	argc;
char	**argv;
{
    int	i;
    int	errflg = 0;
    int	helpflg = 0;

    for (i = 1; i < argc; i++) {
	if (!strcasecmp(argv[i], "-help")) {
	    helpflg++;
	} else {
	    fprintf(stderr, "Unknown option: %s\n", argv[i]);
	    errflg++;
	}
    }

    if (errflg || helpflg) {
	usage();
	exit(errflg > 0);
    }
}
