/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <Xm/Xm.h>
#include <Xm/PanedW.h>
#include <Xm/Form.h>
#include <Xm/DialogS.h>
#include <Xm/List.h>
#include <Xm/LabelG.h>
#include <Xm/Text.h>

#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#include "dtypes.h"
#include "globals.h"
#include "auth.h"
#include "util.h"
#include "newentry.h"

/* Forward declarations */
void	doNewEntry2();
Widget	createNewEntryDialog();
void	destroyNewEntryDialog();
void	doSelectLocation();
void	createNewEntry();
void	doNewEntryCancel();
void	changeEntryType();
void	updateNEOK();

/*
 * Callback for "New Entry" menu item.
 */
void
doNewEntry(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    /*
     * If we're not authenticated, authenticate now.
     */
    if (bound_as[0] == '\0') {
	start_auth(w, doNewEntry2, GetTopShell(w), 1);
    } else {
	doNewEntry2(GetTopShell(w));
    }
}



/*
 * Continuation of "New Entry" code.
 */
void
doNewEntry2(w)
Widget	w;
{
    Widget	newEntryDialog;

    /*
     * Create the add entry dialog
     */
    newEntryDialog = createNewEntryDialog(w);
}




/*
 * Create the initial "add a new entry" dialog.
 */
Widget
createNewEntryDialog(parent)
Widget	parent;
{
    struct ldap_disptmpl	*dt;
    Widget			newEntryDialog, newEntryPane, newEntryForm,
				newEntryActionArea, newEntryTypeList,
				newEntryNameLabel, newEntryNameTextWidget,
				newEntryLocationLabel, newEntryTypeLabel,
				newEntryLocation, newEntryTopForm,
				newEntryOKButton, newEntryCancelButton;
    static ActionAreaItem	newEntryActions[3];
    XmStringTable		list_items;
    int				i, n;
    Arg				args[15];
    newEntryDialogInfo		*nedi;

    nedi = (newEntryDialogInfo *) malloc(sizeof(newEntryDialogInfo));
    if (nedi == NULL) {
	fprintf(stderr, "Out of memory, createNewEntryDialog()\n");
	exit(1);
    }

    /* Create a popup shell... */
    newEntryDialog = XtVaCreatePopupShell("newEntryDialog",
	    xmDialogShellWidgetClass, GetTopShell(parent),
	    XmNdeleteResponse, XmDESTROY,
	    NULL);
    nedi->newEntryDialog = newEntryDialog;

    /* ...with a Form as its only child... */
    newEntryTopForm = XtVaCreateWidget("newEntryTopForm",
	    xmFormWidgetClass, newEntryDialog,
	    XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
	    NULL);

    /* The Form manages a PanedWindow... */
    newEntryPane = XtVaCreateWidget("newEntryPane",
	    xmPanedWindowWidgetClass, newEntryTopForm,
	    XmNsashWidth, 1, /* XXX better way? */
	    XmNsashHeight, 1,
	    NULL);

    /* ...and the PanedWindow manages another Form, which contains... */
    newEntryForm = XtVaCreateManagedWidget("newEntryTopForm",
	    xmFormWidgetClass, newEntryPane, NULL);
    
    /* a label which nominally reads "Select the type of entry to add:"... */
    newEntryTypeLabel = XtVaCreateManagedWidget("newEntryTypeLabel",
	    xmLabelGadgetClass, newEntryForm,
	    XmNtopAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);
    

    /* ...a scrolling list of types of entries we can add... */
    /* Allocate the XmStrings for our list */
    list_items = (XmStringTable)
	    XtMalloc((numAddableTemplates) * sizeof(XmString *));
    /* Allocate an array to hold the disptmpl pointers for addable types */
    nedi->newEntryTemplates = (struct ldap_disptmpl **)
	    malloc((numAddableTemplates) * sizeof(struct ldap_disptmpl *));
    /*
     * For each template we know about, see if it's addable, and if so,
     * add it to a scrolling list of entry types we can add.
     */
    for (dt = ldap_first_disptmpl(templates), i = 0; dt != NULL;
	    dt = ldap_next_disptmpl(templates, dt)) {
	if (LDAP_IS_DISPTMPL_OPTION_SET(dt, LDAP_DTMPL_OPT_ADDABLE)) {
	    /* Can add this type, so add to scrolling list */
	    list_items[i] = XmStringCreateSimple(dt->dt_name);
	    nedi->newEntryTemplates[i] = dt;
	    i++;
	}
    }
    n = 0;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget, newEntryTypeLabel); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
    XtSetArg(args[n], XmNvisibleItemCount,
	    (numAddableTemplates >= 10) ? 10 : numAddableTemplates); n++;
    XtSetArg(args[n], XmNitemCount, numAddableTemplates); n++;
    XtSetArg(args[n], XmNitems, list_items); n++;
    XtSetArg(args[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
    newEntryTypeList = XmCreateScrolledList(newEntryForm, "newEntryTypeList",
	    args, n);
    for (i = 0; i < numAddableTemplates; i++) {
	XmStringFree(list_items[i]);
    }
    XtFree((XtPointer) list_items);
    XtManageChild(newEntryTypeList);
    nedi->newEntryTypeList = newEntryTypeList;

    /* ...a label which reads "Entry Name:"... */
    newEntryNameLabel = XtVaCreateManagedWidget("newEntryNameLabel",
	    xmLabelGadgetClass, newEntryForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, newEntryTypeList,
	    XmNleftAttachment, XmATTACH_FORM,
	    NULL);
    
    /* ...a text field where the entry name can be typed... */
    newEntryNameTextWidget = XtVaCreateManagedWidget("newEntryNameTextWidget",
	    xmTextWidgetClass, newEntryForm,
	    XmNleftAttachment, XmATTACH_WIDGET,
	    XmNleftWidget, newEntryNameLabel,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
	    XmNtopWidget, newEntryNameLabel,
	    NULL);
    nedi->newEntryNameTextWidget = newEntryNameTextWidget;

    /* ...a label which reads "Location:"... */
    newEntryLocationLabel = XtVaCreateManagedWidget("newEntryLocationLabel",
	    xmLabelGadgetClass, newEntryForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, newEntryNameTextWidget,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);

    /*
     * ... and a label which displays the location where new entries
     * will be added.
     */
    newEntryLocation = XtVaCreateManagedWidget("newEntryLocation",
	    xmTextWidgetClass, newEntryForm,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, newEntryLocationLabel,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);
    nedi->newEntryLocation = newEntryLocation;

    /*
     * When the user selects and entry type, we need to update the
     * location label.  Arrange for a callback to handle this.
     */
    XtAddCallback(newEntryTypeList, XmNsingleSelectionCallback,
	    changeEntryType, newEntryLocation);

    /*
     * Select the first entry type in the list.  Causes the above
     * callback to be invoked.
     */
    XmListSelectPos(newEntryTypeList, 1, True);

    /*
     * Create an action area.
     */
    newEntryActions[0].label = "newEntryChooseLocationButton";
    newEntryActions[0].data = (caddr_t) nedi;
    newEntryActions[0].widget = NULL;
    newEntryActions[0].callback = doSelectLocation;
    
    newEntryActions[1].label = "newEntryOKButton";
    newEntryActions[1].data = (caddr_t) nedi;
    newEntryActions[1].widget = &newEntryOKButton;
    newEntryActions[1].callback = createNewEntry;

    newEntryActions[2].label = "newEntryCancelButton";
    newEntryActions[2].data = (caddr_t) nedi;
    newEntryActions[2].widget = &newEntryCancelButton;
    newEntryActions[2].callback = destroyNewEntryDialog;

    newEntryActionArea = CreateActionArea(newEntryPane, newEntryActions, 3);
    XtVaSetValues(newEntryActionArea,
	    XmNtopAttachment, XmATTACH_WIDGET,
	    XmNtopWidget, newEntryLocation,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment, XmATTACH_FORM,
	    XmNrightAttachment, XmATTACH_FORM,
	    NULL);

    nedi->newEntryActionArea = newEntryActionArea;
    nedi->newEntryOKButton = newEntryOKButton;
    nedi->newEntryCancelButton = newEntryCancelButton;

    XtVaSetValues(nedi->newEntryOKButton,
	    XmNshowAsDefault, True,
	    NULL);
    XtVaSetValues(nedi->newEntryActionArea,
	    XmNdefaultButton, nedi->newEntryOKButton,
	    NULL);

    /*
     * The "OK" button should only be active if there is text in the
     * Entry Name: field.
     */
    XtSetSensitive(newEntryOKButton, False);
    XtAddCallback(newEntryNameTextWidget, XmNvalueChangedCallback,
	    updateNEOK, nedi);

    XtManageChild(newEntryPane);
    XtManageChild(newEntryTopForm);
    XtRealizeWidget(newEntryDialog);
    XtMapWidget(newEntryDialog);

    /* Arrange for the new entry dialog to be non-resizable */
    XtAddEventHandler(newEntryDialog, StructureNotifyMask, False,
	    (XtEventHandler) configureSize, NULL);

    return(newEntryDialog);
}



void
doSelectLocation(w, nedi, cbs)
Widget			w;
newEntryDialogInfo	*nedi;
XmAnyCallbackStruct	*cbs;
{
printf("doSelectLocation\n");
}



void
createNewEntry(w, nedi, cbs)
Widget			w;
newEntryDialogInfo	*nedi;
XmAnyCallbackStruct	*cbs;
{
    char			*entryName, *entryLocation;
    int				*posl, posc;
    struct ldap_disptmpl	*dt;
    entryData			*entry;
    static char			buf[2048];

    entryName = XmTextGetString(nedi->newEntryNameTextWidget);
    entryLocation = XmTextGetString(nedi->newEntryLocation);
    if (XmListGetSelectedPos(nedi->newEntryTypeList, &posl, &posc) == True) {
	dt = nedi->newEntryTemplates[posl[0] - 1];
    } else {
	fprintf(stderr, "Error: createNewEntry: no items selected\n");
	return;
    }
    
    entry = ldapTemplate2entryData(dt, bound_as);
    buf[0] = '\0';
    sprintf(buf, "%s=%s, %s", dt->dt_defrdnattrname,
	    entryName, entryLocation);
    entry->en_dn = strdup(buf);
    sprintf(buf, "%s", entryName);
    entry->en_rdn = strdup(buf);

    XtCallCallbacks(nedi->newEntryCancelButton, XmNactivateCallback, nedi);
    display_entry(entry, dt);

    XtFree(entryName);
}



void
doNewEntryCancel(w, nedi, cbs)
Widget			w;
newEntryDialogInfo	*nedi;
XmAnyCallbackStruct	*cbs;
{
printf("doNewEntryCancel\n");
}



void
changeEntryType(w, location, cbs)
Widget			w;
Widget			location;
XmListCallbackStruct	*cbs;
{
    char			*type = NULL;
    struct ldap_disptmpl	*dt;
    char			*str;

    if (cbs->item != NULL) {
	XmStringGetLtoR(cbs->item, XmFONTLIST_DEFAULT_TAG, &type);
	for (dt = ldap_first_disptmpl(templates); dt != NULL;
		dt = ldap_next_disptmpl(templates, dt)) {
	    if (LDAP_IS_DISPTMPL_OPTION_SET(dt, LDAP_DTMPL_OPT_ADDABLE) &&
		    !strcmp(type, dt->dt_name)) {
		if (dt->dt_defaddlocation == NULL) {
		    str = search_base;
		} else {
		    str = dt->dt_defaddlocation;
		}
	    break;
	    }
	}
    } else {
	str = "";
    }

    XmTextSetString(location, str);
    if (type != NULL) {
	XtFree((XtPointer) type);
    }
}



void
updateNEOK(w, nedi, cbs)
Widget			w;
newEntryDialogInfo	*nedi;
XmAnyCallbackStruct	*cbs;
{
    char	*s;

    if (*(s = XmTextGetString(nedi->newEntryNameTextWidget))) {
	XtSetSensitive(nedi->newEntryOKButton, True);
    } else {
	XtSetSensitive(nedi->newEntryOKButton, False);
    }
    XtFree(s);
}



void
destroyNewEntryDialog(w, nedi, cbs)
Widget			w;
newEntryDialogInfo	*nedi;
XmAnyCallbackStruct	*cbs;
{
    if (w == NULL) {
	return;
    }

    if (nedi != NULL) {
	if (nedi->newEntryTemplates != NULL) {
	    free(nedi->newEntryTemplates);
	}
	free(nedi);
    }
    XtDestroyWidget(GetTopShell(w));
    return;
}

