/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#define offset(x) XtOffsetOf (AppData, x)

static XtResource xresources[] = {
    /* Host name of LDAP server */
    {"ldapServer", "LdapServer", XmRString, sizeof (char *),
      offset(ldapServer), XmRString, DFLT_LDAP_SERVER},
    /* Port for LDAP server */
    {"ldapPort", "LdapPort", XmRInt, sizeof (int),
      offset(ldapPort), XmRImmediate, (XtPointer) 0},
    /* Default DN to bind as upon startup */
    {"ldapBindDN", "LdapBindDN", XmRString, sizeof (char *),
      offset(ldapBindDN), XmRImmediate, DFLT_BIND_DN},
    /* Default X.500 search base */
    {"searchBase", "SearchBase", XmRString, sizeof (char *),
      offset(searchBase), XmRString, DFLT_SEARCH_BASE },
    /* Default friendly name mapping file */
    {"friendlyMap", "FriendlyMap", XmRString, sizeof (char *),
      offset(friendlyMap), XmRString, DFLT_FM_FILE },
    /* Default LDAP filter configuration file */
    {"ldapFilterFile", "LdapFilterFile", XmRString, sizeof (char *),
      offset(ldapFilterFile), XmRString, DFLT_FILT_FILE },
    /* LDAP cache size in kbytes */
    {"ldapCacheSize", "LdapCacheSize", XmRInt, sizeof (int),
      offset(ldapCacheSize), XmRImmediate, (XtPointer) DFLT_CACHE_SIZE },
    /* LDAP cache timeout in minutes */
    {"ldapCacheTimeout", "LdapCacheTimeout", XmRInt, sizeof (int),
      offset(ldapCacheTimeout), XmRImmediate, (XtPointer) DFLT_CACHE_TIME},
    /* LDAP search timeout in seconds */
    {"ldapSearchTimeout", "LdapSearchTimeout", XmRInt, sizeof (int),
      offset(ldapSearchTimeout), XmRImmediate,
      (XtPointer) DFLT_LDAP_SEARCH_TIMEOUT},
    /* LDAP read timeout in seconds */
    {"ldapReadTimeout", "LdapReadTimeout", XmRInt, sizeof (int),
      offset(ldapReadTimeout), XmRImmediate,
      (XtPointer) DFLT_LDAP_READ_TIMEOUT},
    /* LDAP write timeout in seconds */
    {"ldapWriteTimeout", "LdapWriteTimeout", XmRInt, sizeof (int),
      offset(ldapWriteTimeout), XmRImmediate,
      (XtPointer) DFLT_LDAP_WRITE_TIMEOUT},
    /* Global templates file */
    {"ldapTemplatesFile", "LdapTemplatesFile", XmRString, sizeof (char *),
      offset(ldapTemplatesFile), XmRString, DFLT_LDAP_TEMPLATES_FILE },
    /* Global searchprefs file */
    {"ldapSearchprefsFile", "LdapSearchprefsFile", XmRString,
      sizeof (char *), offset(ldapSearchprefsFile), XmRString,
      DFLT_LDAP_SEARCHPREFS_FILE },
    /* User configuration file */
    {"configFile", "ConfigFile", XmRString, sizeof (char *),
      offset(configFile), XmRString, DFLT_CONFIG_FILE },
    /* Name of WWW Browser.  Only Mosaic is supported now */
    {"wwwBrowser", "WwwBrowser", XmRString, sizeof (char *),
      offset(wwwBrowser), XmRString, DFLT_WWW_BROWSER },
    /* Program which is used to view JPEG photos */
    {"jpegViewer", "JpegViewer", XmRString, sizeof (char *),
      offset(jpegViewer), XmRString, DFLT_JPEG_VIEWER },
    /* Program which is used to play audio data */
    {"audioViewer", "AudioViewer", XmRString, sizeof (char *),
      offset(audioViewer), XmRString, DFLT_AUDIO_VIEWER },
    /* Maximum length of string to place in selection list */
    {"maxSelLen", "MaxSelLen", XmRInt, sizeof (int),
      offset(maxSelLen), XmRImmediate, (XtPointer) DFLT_MAX_SEL_LEN },
    /* Use (or not) multiple instances of the browser */
    {"multipleBrowseWindows", "MultipleBrowseWindows", XmRBoolean, 
     sizeof(Boolean), offset(multipleBrowseWindows), XmRString, "False" },
    /* Display multiple entries simultaneously */
    {"multipleEntryWindows", "MultipleEntryWindows", XmRBoolean, 
     sizeof(Boolean), offset(multipleEntryWindows), XmRString, "True" },
    /* URL to open when "Retrieve xax500 News" menu chosen */
    {"xax500NewsURL", "Xax500NewsURL", XmRString, sizeof(char *),
      offset(xax500NewsURL), XmRString, DFLT_NEWS_URL },

    /*
     * Recources below mostly deal with internationalization, error
     * messages, and so on.  Typically, will not be set by users.
     */
    /* Start up with "More Choices" mode */
    {"moreChoices", "MoreChoices", XmRBoolean, sizeof(Boolean),
     offset(moreChoices), XmRString, "False" },
    /* Name of help file */
    {"helpFile", "HelpFile", XmRString, sizeof (char *),
      offset(helpFile), XmRString, DFLT_HELP_FILE },
    /* Label for "Fewer Choices" button */
    {"fewerButtonLabel", "FewerButtonLabel", XmRString, sizeof (char *),
      offset(fewerButtonLabel), XmRString, NULL },
    /* Label for "More Choices" button */
    {"moreButtonLabel", "MoreButtonLabel", XmRString, sizeof (char *),
      offset(moreButtonLabel), XmRString, NULL },
    /* "Look Within" preface for searchbase in main app window */
    {"lookWithinString", "LookWithinString", XmRString, sizeof (char *),
      offset(lookWithinString), XmRString, NULL },
    /* String to use in find window searchbase area */
    /* when searchbase is top of DIT */
    {"theWorldString", "TheWorldString", XmRString, sizeof (char *),
      offset(theWorldString), XmRString, NULL },
    {"boundAsString", "BoundAsString", XmRString, sizeof (char *),
      offset(boundAsString), XmRString, NULL },
    {"authFilter", "AuthFilter", XmRString, sizeof (char *),
      offset(authFilter), XmRString, DFLT_AUTH_FILTER },
    {"msgSortByFullName", "MsgSortByFullName", XmRString, sizeof (char *),
      offset(msgSortByFullName), XmRString, DFLT_MSG_SB_FULLNAME },
    {"msgSortBySurname", "MsgSortBySurname", XmRString, sizeof (char *),
      offset(msgSortBySurname), XmRString, DFLT_MSG_SB_SURNAME },


      /* --- Error message texts --- */
    {"emHelpTextNotFound", "EmHelpTextNotFound", XmRString, sizeof (char *),
      offset(emHelpTextNotFound), XmRString, DMHelpTextNotFound },
    {"emHelpTextNotFoundDlgLabel", "EmHelpTextNotFoundDlgLabel", XmRString,
      sizeof (char *), offset(emHelpTextNotFoundDlgLabel), XmRString,
      DMHelpTextNotFoundDlgLabel },
    {"emBlanksNotAllowed", "EmBlanksNotAllowed", XmRString, sizeof (char *),
      offset(emBlanksNotAllowed), XmRString, DMBlanksNotAllowed },
    {"emSearchTimedOut", "EmSearchTimedOut", XmRString, sizeof (char *),
      offset(emSearchTimedOut), XmRString, DMSearchTimedOut },
    {"emBlankURLNotAllowed", "EmBlankURLNotAllowed", XmRString, sizeof (char *),
      offset(emBlankURLNotAllowed), XmRString, DMBlankURLNotAllowed },
    {"emBadDNArgs", "EmBadDNArgs", XmRString, sizeof (char *),
      offset(emBadDNArgs), XmRString, DMBadDNArgs },
    {"emSizeLimit", "EmSizeLimit", XmRString, sizeof (char *),
      offset(emSizeLimit), XmRString, DMSizeLimit },
    {"emTimeLimit", "EmTimeLimit", XmRString, sizeof (char *),
      offset(emTimeLimit), XmRString, DMTimeLimit },
};

static XrmOptionDescRec options[] = {
    {"-ldapserver",	"*ldapServer", XrmoptionSepArg, NULL},
    {"-port",		"*ldapPort", XrmoptionSepArg, NULL},
    {"-binddn",		"*ldapBindDN", XrmoptionSepArg, NULL},
    {"-sb",		"*searchBase", XrmoptionSepArg, NULL},
    {"-fm",		"*friendlyMap", XrmoptionSepArg, NULL},
    {"-filt",		"*ldapFilterFile", XrmoptionSepArg, NULL},
    {"-cachesize",	"*ldapCacheSize", XrmoptionSepArg, NULL},
    {"-cachetime",	"*ldapCacheTimeout", XrmoptionSepArg, NULL},
    {"-config",		"*configFile", XrmoptionSepArg, NULL},
    {"-template",	"*ldapTemplatesFile", XrmoptionSepArg, NULL},
    {"-searchprefs",	"*ldapSearchprefsFile", XrmoptionSepArg, NULL},
    {"-mc",		"*moreChoices", XrmoptionNoArg, "True"},
    {"-sto",		"*ldapSearchTimeout", XrmoptionSepArg, NULL},
    {"-rto",		"*ldapReadTimeout", XrmoptionSepArg, NULL},
    {"-wto",		"*ldapWriteTimeout", XrmoptionSepArg, NULL},
    {"-helpfile",	"*helpFile", XrmoptionSepArg, NULL},
};
