/*
 * Copyright (c) 1993, 1994, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <Xm/DrawnB.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/DragDrop.h>
#include <Xm/ArrowB.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>

#include <stdio.h>
#include <sys/wait.h>
#include <netdb.h>

#include <portable.h>
#include <lber.h>
#include <ldap.h>
#include <disptmpl.h>

#include <errno.h>
#ifdef ultrix
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif /* ultrix */

#include "dtypes.h"
#include "globals.h"
#include "find.h"
#include "prefs.h"
#include "auth.h"
#include "browse.h"
#include "display.h"
#include "help.h"
#include "util.h"


/* Forward references */
#ifndef NEEDPROTOS
void	doNothing();
static struct ldap_adddeflist *attr_has_default_value();
struct berval **make_value_len();
#else /* !NEEDPROTOS */
void	doNothing();
static struct ldap_adddeflist *attr_has_default_value(struct ldap_disptmpl *dt,
	char *attrname);
struct berval **make_value_len(char *value);
#endif /* !NEEDPROTOS */

void cleanup_and_exit()
{
    int			i;
    struct forkedApps	*fa, *fan;
    displayData		*dd, *ddn;

    /* Check for any unsaved entries */
    dd = displayDlgData;
    while (dd != NULL) {
	/*
	 * Since we're going to free dd in the callback, save the di_next
	 * pointer before we call the callback.
	 */
	ddn = dd->di_next;
	XtCallCallbacks(dd->di_displayDismissButton, XmNactivateCallback, NULL);
	dd = ddn;
    }

#ifdef notdef
    for (dd = displayDlgData; dd != NULL; dd = dd->di_next) {
	XtCallCallbacks(dd->di_displayDismissButton, XmNactivateCallback, NULL);
    }
#endif

    if (ldp) {
	ldap_unbind(ldp);
    }
    /* clean up our temp files */
    fa = forkedAppsList;
    while (fa != NULL) {
	for (i = 0; fa->fa_files[i] != NULL; i++) {
	    if (unlink(fa->fa_files[i]) == -1) {
		if (errno != ENOENT) {
		    perror("unlink");
		}	
	    }
	}
	fan = fa->fa_next;
	free(fa);
	fa = fan;
    }

#ifdef notdef
    for (fa = forkedAppsList; fa != NULL; fa = fa->fa_next) {
	for (i = 0; fa->fa_files[i] != NULL; i++) {
	    if (unlink(fa->fa_files[i]) == -1) {
		if (errno != ENOENT) {
		    perror("unlink");
		}	
	    }
	}
    }
#endif

    if (searchprefs != NULL) {
	ldap_free_searchprefs(searchprefs);
    }

    freeHelp();

    exit(0);
}




Widget
create_pixmap_button(parent, pixmap_name, inactive_pixmap_name, wname)
Widget	parent;
char	*pixmap_name;
char	*inactive_pixmap_name;
char	*wname;
{
    Pixmap	pixmap, ipixmap;
    Widget	button;

    pixmap = XmGetPixmap(XtScreen(parent), pixmap_name, fg, bg);
    if (inactive_pixmap_name == NULL) {
	inactive_pixmap_name = pixmap_name;
    }
    ipixmap = XmGetPixmap(XtScreen(parent), inactive_pixmap_name, fg, bg);
    button = XtVaCreateWidget(wname,
	    xmDrawnButtonWidgetClass,
	    parent, 
	    XmNlabelType, XmPIXMAP,
	    XmNlabelPixmap, pixmap,
	    XmNlabelInsensitivePixmap, ipixmap,
	    XmNshadowType, XmSHADOW_ETCHED_OUT,
	    XmNpushButtonEnabled, True,
	    NULL);
    return(button);
}




Widget
create_find_button(parent)
Widget parent;
{
    Widget	w;

    w = create_pixmap_button(parent, "find_image", NULL, "findButton");
    return(w);
}
	

void
SetIcon(w, name)
Widget	w;
char	*name;
{
    Pixmap pixmap = XmGetPixmap(XtScreen(w), name, fg, bg);
    XtVaSetValues(w, XmNiconPixmap, pixmap, NULL);
}



#define TIGHTNESS 20

Widget
CreateActionArea(parent, actions, num_actions)
Widget parent;
ActionAreaItem *actions;
int num_actions;
{
    Widget action_area, widget;
    int i;

    action_area = XtVaCreateWidget("action_area", xmFormWidgetClass, parent,
        XmNfractionBase, TIGHTNESS*num_actions - 1,
        XmNleftOffset,   10,
        XmNrightOffset,  10,
        NULL);

    for (i = 0; i < num_actions; i++) {
        widget = XtVaCreateManagedWidget(actions[i].label,
            xmPushButtonWidgetClass, action_area,
            XmNleftAttachment,       i? XmATTACH_POSITION : XmATTACH_FORM,
            XmNleftPosition,         TIGHTNESS*i,
            XmNtopAttachment,        XmATTACH_FORM,
            XmNbottomAttachment,     XmATTACH_FORM,
            XmNrightAttachment,
                    i != num_actions-1? XmATTACH_POSITION : XmATTACH_FORM,
            XmNrightPosition,        TIGHTNESS*i + (TIGHTNESS-1),
            XmNshowAsDefault,        i == 0,
            XmNdefaultButtonShadowThickness, 1,
            NULL);
        if (actions[i].callback)
            XtAddCallback(widget, XmNactivateCallback,
                actions[i].callback, actions[i].data);
	if (actions[i].widget) {
	    *(actions[i].widget) = widget;
	}
        if (i == 0) {
            /* Set the action_area's default button to the first widget
             * created (or, make the index a parameter to the function
             * or have it be part of the data structure). Also, set the
             * pane window constraint for max and min heights so this
             * particular pane in the PanedWindow is not resizable.
             */
            Dimension height, h;
            XtVaGetValues(action_area, XmNmarginHeight, &h, NULL);
            XtVaGetValues(widget, XmNheight, &height, NULL);
            height += 2 * h;
            XtVaSetValues(action_area,
                XmNdefaultButton, widget,
                XmNpaneMaximum,   height,
                XmNpaneMinimum,   height,
                NULL);
        }
    }

    XtManageChild(action_area);

    return action_area;
}



int
is_leaf(dn, oclist)
char *dn;
char **oclist;
{
    int i;

    if ((oclist == NULL) || (dn == NULL)) {
	/* Who knows? */
	return 1;
    }
    /* XXX fix later, but for now look for objClass clues */
    for (i = 0; oclist[i] != NULL; i++) {
	if (!strcasecmp(oclist[i], "quipuNonLeafObject")) {
	    return 0;
	}
    }
    return 1;
}


Widget
GetTopShell(w)
Widget	w;
{
    while (w && !XtIsWMShell(w)) {
	w = XtParent(w);
    }
    return(w);
}


void
reap_child(sig, code, scp, addr)
int			sig, code;
void			*scp;
char			*addr;
{
    struct forkedApps	**fa;
    int			i;
    WAITSTATUSTYPE	status;
    pid_t		pid;

#ifdef USE_WAITPID
    while ((pid = waitpid(-1, &status, WNOHANG | WUNTRACED)) > 0) {
#else /* USE_WAITPID */
    while ((pid = wait3(&status, WNOHANG | WUNTRACED, NULL)) > 0) {
#endif /* USE_WAITPID */
	for (fa = &forkedAppsList; (*fa != NULL) && ((*fa)->fa_pid != pid);
		fa = &(*fa)->fa_next);
	if (*fa == NULL) {
	    fprintf(stderr, "warning: pid %d exited but no fa\n", pid);
	} else {
	    for (i = 0; (*fa)->fa_files[i] != NULL; i++) {
		if (unlink((*fa)->fa_files[i]) == -1) {
		    if (errno != ENOENT) {
			perror("unlink");
		    }
		}
		free((*fa)->fa_files[i]);
	    }
	    free((*fa)->fa_files);
	    *fa = (*fa)->fa_next;
	}
    }
}




/*
 * From motif FAQ (David Brooks, Systems Engineering, Open Software Foundation)
 */
void
ForceDialog(w)
Widget w;
{
    Widget		diashell, topshell;
    Window		diawindow, topwindow;
    Display		*dpy;
    XWindowAttributes	xwa;
    XEvent		event;
    XtAppContext	cxt;

    /*
     * Locate the shell we are interested in.  In a particular
     * instance, you may know these shells already.
     */

    for (diashell = w; !XtIsShell(diashell); diashell = XtParent(diashell));

    /* Locate its primary window's shell (which may be the same) */

    for (topshell = diashell; !XtIsTopLevelShell(topshell);
	    topshell = XtParent(topshell)) {
    }

    if (XtIsRealized(diashell) && XtIsRealized(topshell)) {
	dpy = XtDisplay(topshell);
	diawindow = XtWindow(diashell);
	topwindow = XtWindow(topshell);
	cxt = XtWidgetToApplicationContext(diashell);

	/*
	 * Wait for the dialog to be mapped. 
	 * It's guaranteed to become so unless...
	 */

	while (XGetWindowAttributes(dpy, diawindow, &xwa),
		xwa.map_state != IsViewable) {

	    /*
	     * ...if the primary is (or becomes) unviewable or
	     * unmapped, it's probably iconified, and nothing
	     * will happen.
	     */

	    if (XGetWindowAttributes(dpy, topwindow, &xwa),
		    xwa.map_state != IsViewable) {
		break;
	    }

	    /*
	     * At this stage, we are guaranteed there will be an
	     * event of some kind.  Beware; we are presumably
	     * in a callback, so this can recurse.
	     */

	    XtAppNextEvent(cxt, &event);
	    XtDispatchEvent(&event);
	}
    }

    /*
     * The next XSync() will get an expose event if the dialog
     * was unmapped.
     */

    XmUpdateDisplay(topshell);
}




/*
 * This nonsense is a workaround for a Motif bug.  If you call
 * XmProcessTraversal right after realizing a Widget, it might
 * not do anything.  Instead, set input focus by setting up an
 * XtAppTimeout to go off and set the input focus.
 */
static void
focusTO(data, id)
void		*data;
XtIntervalId	*id;
{
    XmProcessTraversal((Widget) data, XmTRAVERSE_CURRENT);
    XmProcessTraversal((Widget) data, XmTRAVERSE_CURRENT);
}

static void
focusCB(w, data, cbs)
Widget		w;
XtPointer	data;
XtPointer	cbs;
{
    XtRemoveCallback(w, XmNfocusCallback, focusCB, data);
    XtAppAddTimeOut(XtWidgetToApplicationContext(w), 0, focusTO, data);
}

void
SetFocus(parent, w)
Widget		parent;
Widget		w;
{
    XtAppAddTimeOut(XtWidgetToApplicationContext(w), 0, focusTO, w);
}


void
raiseFindWindow(w, client_data, cbs)
Widget			w;
XtPointer		client_data;
XmAnyCallbackStruct	*cbs;
{
    XtMapWidget(toplevel);
    XtPopup(toplevel, XtGrabNone);
}




void
freeXmStringList(items, count)
XmString	*items;
int		count;
{
    int	i;

    for (i = 0; i < count; i++) {
	XmStringFree(items[i]);
    }
    free(items);
}




void
configureSize(w, client_data, event)
Widget		w;
XtPointer	client_data;
XConfigureEvent	*event;
{
    if (event->type != ConfigureNotify) {
	return;
    }
    XtVaSetValues(w,
	    XmNminWidth, event->width,
	    XmNminHeight, event->height,
	    XmNmaxWidth, event->width,
	    XmNmaxHeight, event->height,
	    NULL);
    XtRemoveEventHandler(w, StructureNotifyMask, False, 
	    (XtEventHandler) configureSize, NULL);

}

/*
 * Allocate and initialize a valueData structure. Places structure at
 * the head of the valueData linked list for en.
 */
struct valueData *
get_valuedata(dd)
displayData	*dd;
{
    struct valueData	*vd;

    if (dd == NULL) {
	return NULL;
    }

    vd = (valueData *) malloc(sizeof(valueData));
    if (vd == NULL) {
	return NULL;
    }
    memset(vd, 0, sizeof(struct valueData));

    vd->vd_next = dd->di_valueDataList;
    dd->di_valueDataList = vd;
    return vd;
}


/*
 * Free a valueData structure
 */
void
free_valuedata(vd)
valueData	*vd;
{
    if (vd != NULL) {
	switch (vd->vd_tag) {
	    case VD_URL:
		/* FALLTHROUGH */
	    case VD_DN:
		if (vd->vd_value != NULL) {
		    free(vd->vd_value);
		}
		break;
	    default:
		fprintf(stderr, "free_valuedata: warning, illegal valueData");
	}
	free(vd);
    }
}



/*
 * Free an actiondata structure
 */
void
free_actiondata(actd)
actionData *actd;
{
    if (actd == NULL) {
	return;
    }
    ch_free(actd->ac_disableattr);
    switch (actd->ac_type) {
    case ACTION_TYPE_SEARCHACT: {
	searchactData *args;
	args = actd->ac_args;
	if (args != NULL) {
	    ch_free((char *) args->search_value);
	    ch_free((char *) args->search_filter);
	    ch_free((char *) args->search_disambig);
	    ch_free((char *) args->search_disambiglabel);
	}
	ch_free((char *) args);
	break;
    }
    case ACTION_TYPE_ADDDNACT:
	{
	    adddnactData *args;
	    args = actd->ac_args;
	    if (args != NULL) {
		ch_free((char *) args->adddn_value);
		ch_free((char *) args->adddn_dstdn);
		ch_free((char *) args->adddn_dstattr);
		ch_free((char *) args->adddn_disableattr);
		ch_free((char *) args->adddn_existsmsg);
		ch_free((char *) args->adddn_confmsg);
	    }
	    ch_free((char *) args);
	    break;
	}
    case ACTION_TYPE_VERIFYACT:
	{
	    verifyactData *args;
	    args = actd->ac_args;
	    fprintf(stderr, "XXX - must free verifyact args!!!\n");
	    break;
	}
    case ACTION_TYPE_LINKACT:
	{
	    linkactData *args;
	    args = actd->ac_args;
	    fprintf(stderr, "XXX - must free linkact args!!!\n");
	    break;
	}
    }
    free(actd);
    return;
}

/*
 * Allocate and initialize a searchData structure
 */
struct searchData *
get_searchdata()
{
    struct searchData	*f;
    f = (searchData *) malloc(sizeof(searchData));
    if (f == NULL) {
	return NULL;
    }
    memset(f, 0, sizeof(struct searchData));
    return f;
}




/*
 * Free a searchData struct
 */
void
free_searchdata(fp)
struct searchData	**fp;
{
    int	i;
    struct searchData *f = *fp;

    if (f) {
	if (f->ldfp != NULL) {
	    ldap_getfilter_free(f->ldfp);
	}
	if (f->search_string != NULL) {
	    free(f->search_string);
	}
	if (f->searchbase != NULL) {
	    free(f->searchbase);
	}
	if (f->newsb != NULL) {
	    free(f->newsb);
	}
	if (f->dn_list != NULL) {
	    ldap_value_free(f->dn_list);
	}
	if (f->sattrs != NULL) {
	    ldap_value_free(f->sattrs);
	}
	if (f->dattrs != NULL) {
	    ldap_value_free(f->dattrs);
	}
	if (f->allattrs != NULL) {
	    ldap_value_free(f->allattrs);
	}
	if (f->objcllist != NULL) {
	    for (i = 0; f->objcllist[i] != NULL; i++) {
		ldap_value_free(f->objcllist[i]);
	    }
	}
	if (f->filterspec != NULL) {
	    free(f->filterspec);
	}
	free(f);
	*fp = NULL;
    }
}




char **
copy_vals(vals)
char	**vals;
{
    char	**newvals;
    int		i, numvals;

    for (i = 0; vals[i] != NULL; i++);
    numvals = i;
    newvals = (char **) malloc(numvals * sizeof(char *));
    for (i = 0; vals[i] != NULL; i++) {
	newvals[i] = strdup(vals[i]);
    }
    return(newvals);
}


/*
 * Return the number of ldap_tmplitem arguments
 */
static int
tmpl_count_args(ti)
struct ldap_tmplitem *ti;
{
    int i;

    if ((ti == NULL) || (ti->ti_args == NULL)) {
	return(0);
    }
    for (i = 0; ti->ti_args[i] != NULL; i++);
    return(i);
}
	


int
fixup_templates(templates)
struct ldap_disptmpl	*templates;
{
    struct ldap_disptmpl	*dt;
    struct ldap_tmplitem	*ti, *ti2;

    if (templates == NULL) {
	return -1;
    }

    /*
     * Fix up template ti_appdata fields to point to our
     * syntax handlers. XXX - needs work.
     */
    for (dt = ldap_first_disptmpl(templates), numAddableTemplates = 0;
	    dt != NULL; dt = ldap_next_disptmpl(templates, dt)) {
	if (LDAP_IS_DISPTMPL_OPTION_SET(dt, LDAP_DTMPL_OPT_ADDABLE)) {
	    numAddableTemplates++;
	}
	for (ti = ldap_first_tmplrow(dt); ti != NULL;
		 ti = ldap_next_tmplrow(dt, ti)) {
	    for (ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULL;
		     ti2 = ldap_next_tmplcol(dt, ti, ti2)) {
		switch (ti2->ti_syntaxid) {
		case LDAP_SYN_CASEIGNORESTR :
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[0]);
		    break;
		case LDAP_SYN_MULTILINESTR :
		    /* 
		     * If single-valued, use alternate syntax to
		     * preserve order.
		     */
		    if (LDAP_IS_TMPLITEM_OPTION_SET(ti2,
			    LDAP_DITEM_OPT_SINGLEVALUED)) {
			LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[14]);
		    } else {
			LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[1]);
		    }
		    break;
		case LDAP_SYN_DN :
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[2]);
		    break;
		case LDAP_SYN_JPEGBUTTON:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[4]);
		    break;
		case LDAP_SYN_AUDIOBUTTON:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[3]);
		    break;
		case LDAP_SYN_BOOLEAN:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[5]);
		    break;
		case LDAP_SYN_LABELEDURL:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[6]);
		    break;
		case LDAP_SYN_TIME:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[7]);
		    break;
		case LDAP_SYN_DATE:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[8]);
		    break;
		case LDAP_SYN_RFC822ADDR:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[9]);
		    break;
		case LDAP_SYN_SEARCHACTION:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[10]);
		    /* Sanity-check number of arguments */
		    if (tmpl_count_args(ti2) != NUM_SEARCHACTION_ARGS) {
			fprintf(stderr, "Incorrect number of arguments for ");
			fprintf(stderr, "searchaction \"%s\"\n", ti2->ti_label);
			return(-1);
		    }
		    break;
		case LDAP_SYN_LINKACTION:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[11]);
		    /* Sanity-check number of arguments */
		    if (tmpl_count_args(ti2) != NUM_LINKACTION_ARGS) {
			fprintf(stderr, "Incorrect number of arguments for ");
			fprintf(stderr, "linkaction \"%s\"\n", ti2->ti_label);
			return(-1);
		    }
		    break;
		case LDAP_SYN_ADDDNACTION:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[12]);
		    /* Sanity-check number of arguments */
		    if (tmpl_count_args(ti2) != NUM_ADDDNACTION_ARGS) {
			fprintf(stderr, "Incorrect number of arguments for ");
			fprintf(stderr, "adddnaction \"%s\"\n", ti2->ti_label);
			return(-1);
		    }
		    break;
		case LDAP_SYN_VERIFYDNACTION:
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[13]);
		    /* Sanity-check number of arguments */
		    if (tmpl_count_args(ti2) != NUM_VERIFYDNACTION_ARGS) {
			fprintf(stderr, "Incorrect number of arguments for ");
			fprintf(stderr, "verifydnaction \"%s\"\n",
				ti2->ti_label);
			return(-1);
		    }
		    break;
		default:
		    fprintf(stderr,
			    "don't know how to handle syntax type 0x%x\n",
			    ti2->ti_syntaxid);
		    LDAP_SET_TMPLITEM_APPDATA(ti2, &syntaxes[1]);
		    break;
		}
	    }
	}
    }
    return 0;
}

#if defined( ultrix )
static int _gettemp();

int
mkstemp(path)
char *path;
{
    int fd;

    return (_gettemp(path, &fd) ? fd : -1);
}




static int
_gettemp(path, doopen)
char *path;
register int *doopen;
{
    extern int		errno;
    register char	*start, *trv;
    struct stat		sbuf;
    u_int		pid;

    pid = getpid();
    for (trv = path; *trv; ++trv);		/* extra X's get set to 0's */
    while (*--trv == 'X') {
	*trv = (pid % 10) + '0';
	pid /= 10;
    }

    /*
     * check the target directory; if you have six X's and it
     * doesn't exist this runs for a *very* long time.
     */
    for (start = trv + 1;; --trv) {
	if (trv <= path) {
	    break;
	}
	if (*trv == '/') {
	    *trv = '\0';
	    if (stat(path, &sbuf))
		return(0);
	    if (!S_ISDIR(sbuf.st_mode)) {
		errno = ENOTDIR;
		return(0);
	    }
	    *trv = '/';
	    break;
	}
    }

    for (;;) {
	if (doopen) {
	    if ((*doopen = open(path, O_CREAT|O_EXCL|O_RDWR, 0600)) >= 0) {
		return(1);
	    }
	    if (errno != EEXIST) {
		return(0);
	    }
	}
	else if (stat(path, &sbuf)) {
	    return(errno == ENOENT ? 1 : 0);
	}

	/* tricky little algorithm for backward compatibility */
	for (trv = start;;) {
	    if (!*trv) {
		return(0);
	    }
	    if (*trv == 'z') {
		*trv++ = 'a';
	    } else {
		if (isdigit(*trv)) {
		    *trv = 'a';
		} else {
		    ++*trv;
		}
		break;
	    }
	}
    }
	/*NOTREACHED*/
}
#endif /* defined( ultrix ) */


void
flushCache(w, data, cbs)
Widget		w;
XtPointer	data;
XtPointer	cbs;
{
    ldap_flush_cache(ldp);
    /* infoDlg(toplevel, "Cache flushed"); */ /* You're welcome, Lance */
}       


int
str2upper( s )
char        *s;
{
    char    *p;

    for ( p = s; *p != '\0'; ++p ) {
	if ( islower( *p )) {
	    *p = toupper( *p );
	}
    }
}


int
berval_compare(v1, v2)
struct berval	*v1;
struct berval	*v2;
{
    if (v1->bv_len != v2->bv_len) {
	return(0);
    } else {
	return(!strncasecmp(v1->bv_val, v2->bv_val, v1->bv_len));
    }
}

int
bervec_compare(b1, b2)
struct berval	**b1;
struct berval	**b2;
{
    int i,j;
    int found;

    if ((b1 == NULL) && (b2 == NULL)) {
	return(1);
    }

    /* Same # of bervals? */
    if (b1 == NULL) {
	i = -1;
    } else {
	for (i = 0; b1[i] != NULL; i++);
    }
    if (b2 == NULL) {
	j = -1;
    } else {
	for (j = 0; b2[j] != NULL; j++);
    }
    if (i != j) {
	return(0);
    }
    /* Are they the same?  O(n^2) search.  Hope there aren't a lot... */
    for (i = 0; b1[i] != NULL; i++) {
	found = 0;
	for (j = 0; b2[j] != NULL; j++) {
	    if (berval_compare(b1[i], b2[j])) {
		found = 1;
		break;
	    }
	}
	if (!found) {
	    return(0);
	}
    }
    return(1);
}


/*
 * Make a copy of a char **
 */
char **
ccopy(src)
char	**src;
{
    int	i;
    char	**new;

    if (src == NULL) {
	return(NULL);
    }

    for (i = 0; src[i] != NULL; i++);
    i++;

    new = (char **) malloc(i * sizeof(char *));
    memset(new, 0, i * sizeof(char *));
    for (i = 0; src[i] != NULL; i++) {
	new[i] = strdup(src[i]);
    }
    return(new);
}




/*
 * Add a string to an existing bervec
 */
int
berval_add(bv, str)
struct	berval	***bv;
char		*str;
{
    int 		i;

    if (str == NULL) {
	return 0;
    }

    if (*bv == NULL) {
	*bv = (struct berval **) malloc(2 * sizeof(struct berval *));
	i = 0;
    } else {
	for (i = 0; ((*bv)[i] != NULL); i++);
	*bv = (struct berval **) realloc(*bv,
		((i + 2) * sizeof(struct berval *)));
    }
    if (*bv == NULL) {
	return -1;
    }
    (*bv)[i + 1] = NULL;
    (*bv)[i] = (struct berval *) malloc(sizeof(struct berval));
    (*bv)[i]->bv_val = strdup(str);
    (*bv)[i]->bv_len = strlen(str);

    return 0;
}





/*
 * Delete a given value from a bervec
 */
int
berval_del(bv, str)
struct	berval	***bv;
char		*str;
{
    int i, found = 0;

    if ((bv == NULL) || (*bv == NULL)) {
	return -1;
    }

    for (i = 0; !found && (*bv)[i] != NULL; i++) {
	if (!strncmp((*bv)[i]->bv_val, str, (*bv)[i]->bv_len)) {
	    found = 1;
	    break;
	}
    }

    if (!found) {
	return -1;
    }

    free((*bv)[i]->bv_val);

    for (; (*bv)[i + 1] != NULL; i++) {
	(*bv)[i]->bv_val = (*bv)[i + 1]->bv_val;
	(*bv)[i]->bv_len = (*bv)[i + 1]->bv_len;
    }
    free((*bv)[i]);
    (*bv)[i] = NULL;
    *bv = (struct berval **) realloc(*bv, (i + 1) * sizeof(struct berval *));

    return 0;
}




/* 
 * Make a copy of a bervec
 */
struct berval **
bervec_copy(in)
struct berval **in;
{
    int	i;
    struct berval **out;

    for (i = 0; in[i] != NULL; i++);

    out = (struct berval **) malloc((i + 1) * sizeof(struct berval *));
    for (i = 0; in[i] != NULL; i++) {
	out[i] = (struct berval *) malloc(sizeof(struct berval));
	out[i]->bv_len = in[i]->bv_len;
	out[i]->bv_val = strdup(in[i]->bv_val);
    }
    out[i] = NULL;
    return(out);
}



	
/*
 * Return True if the widget is one of our on-screen display dialogs,
 * False otherwise.
 */
Boolean
isDisplayDialog(w)
Widget w;
{
    displayData	*di;

    for (di = displayDlgData; di != NULL; di = di->di_next) {
	if (di->di_displayDlg == w) {
	    return(True);
	}
    }
    return(False);
}


int
stcmp(a, b)
char **a, **b;
{
    return(strcasecmp(*a, *b));
}


/*
 * Given a buffer of text, sort the lines in lexicographical order.
 */
char *
sort_strings(in)
char *in;
{
    int		i, nl;
    char	*p, *s, *out;
    char	**lines;

    /* count number of lines */
    if ((in == NULL) || (*in == '\0')) {
	if (in) {
	    free(in);
	}
	return(strdup(""));
    }

    out = (char *) malloc(strlen(in) + 1);
    out[0] = '\0';

    p = in;
    nl = 1;
    while (*p != '\0') {
	if (*p == '\n') {
	    nl++;
	}
	p++;
    }
    lines = (char **) malloc(nl * sizeof(char *));

    p = in;
    lines[0] = in;
    i = 1;
    while (*p != '\0') {
	if (*p == '\n') {
	    lines[i] = p + 1;
	    *p = '\0';
	    i++;
	}
	p++;
    }

    qsort((char *) lines, nl, sizeof(char *), stcmp);

    for (i = 0, p = out; i < nl; i++) {
	s = lines[i];
	while (*s) {
	    *p++ = *s++;
	}
	*p++ = '\n';
    }
    *--p = '\0';
    free(lines);
    return(out);
}




void
initAtoms()
{
    Display	*dpy;
    Atom	TEXT, COMPOUND_TEXT, DN, LABELED_URL;

    dpy = XtDisplay(toplevel);
    TEXT = XmInternAtom(dpy, "TEXT", False);
    COMPOUND_TEXT = XmInternAtom(dpy, "COMPOUND_TEXT", False);
    DN = XmInternAtom(dpy, "DN", False);
    LABELED_URL = XmInternAtom(dpy, "LABELED_UR", False);
}



/*
 * Look through the children of this widget and return the first Text
 * widget found.
 */
Widget
getLabelGadgetChild(w)
Widget	w;
{
    WidgetList	children;
    Cardinal	numchildren;
    int		i;

    XtVaGetValues(w,
	XmNnumChildren, &numchildren,
	XmNchildren, &children,
	NULL);

    for (i = 0; i < numchildren; i++) {
	if (XmIsLabelGadget(children[i])) {
	    return(children[i]);
	}
    }
    return(NULL);
}




/*
 * Look through the children of this widget and return the first Text
 * widget found.
 */
Widget
getArrowButtonChild(w)
Widget	w;
{
    WidgetList	children;
    Cardinal	numchildren;
    int		i;

    XtVaGetValues(w,
	XmNnumChildren, &numchildren,
	XmNchildren, &children,
	NULL);

    for (i = 0; i < numchildren; i++) {
	if (XmIsArrowButton(children[i])) {
	    return(children[i]);
	}
    }
    return(NULL);
}



void
clearStatus(w, id)
Widget		w;
XtIntervalId	*id;
{
    XtIntervalId	oid;
    XmString		s;

    if (XtIsRealized(w)) {
	XtVaGetValues(w, XmNuserData, &oid, NULL);
	if (oid == *id) {
	    s = XmStringCreateLocalized("");
	    XtVaSetValues(w, XmNlabelString, s, NULL);
	    XmStringFree(s);
	}
    }
}



/*
 * Place the given text in a status window.  Keep it there at least
 * <delay> seconds, then clear it out.  Sets the widget's userData resource
 * to the the interval id returned from XtAppAddTimeOut, so that the
 * timer callback proc can know if someone else wrote to the status widget.
 */
void
timedStatusUpdate(w, text, delay)
Widget		w;
char		*text;
int		delay;
{
    XtIntervalId	t;
    XmString		s;

    if (!w) {
	return;
    }
    s = XmStringCreateSimple(text);
    t = XtAppAddTimeOut(app, delay * 1000, clearStatus, w);
    XtVaSetValues(w,
	XmNlabelString, s,
	XmNuserData, t,
	NULL);
    XmStringFree(s);
}




/*
 * Given any widget within a display window, return the displayData
 * for that window.
 */
displayData *
getDisplayDataByWidget(w)
Widget	w;
{
    Widget	t;
    displayData	*d;

    t = GetTopShell(w);
    for (d = displayDlgData; d != NULL; d = d->di_next) {
	if (d->di_displayDlg == t) {
	    return(d);
	}
    }
    return(NULL);
}



/* 
 * Allocate and initialize a displayData structure
 */
displayData *
getDisplayData()
{
    displayData	*d;

    if ((d = (displayData *) malloc(sizeof(displayData))) == NULL) {
	return NULL;
    }

    d->di_displayDlg = d->di_status = d->di_displaySWin = NULL;
    d->di_displaySaveButton = d->di_displayDismissButton = NULL;
    d->di_displayAsOption = d->di_displayDN = d->di_prevWidget = NULL;
    d->di_currAttr = d->di_prevAttr = NULL;
    d->di_displayAttrData = NULL;
    d->di_entryData = NULL;
    d->di_valueDataList = NULL;
    d->di_closeOnWrite = False;
    d->di_tmpl = 0;
    d->di_numMods = 0;
    d->di_nextDlgForEntry = d->di_next = NULL;

    return d;
}

/*
 * Given a displayData struct and a template item, return the
 * attrData for that template item.
 */
displayAttrData *
getDisplayAttrDataByTemplate(d, ti)
displayData		*d;
struct ldap_tmplitem	*ti;
{
    displayAttrData	*dad;

    if ((!d) || (!ti) || (!d->di_entryData)) {
	return(NULL);
    }

    for (dad = d->di_displayAttrData; dad != NULL; dad = dad->dad_next) {
	if (ti == dad->dad_ti) {
	    return(dad);
	}
    }

    return(NULL);
}


/*
 * Given a label widget, return the displayAttrData struct associated
 * with that label, e.g. you have the Widget label
 * "See Also" and you want the displayAttrData for the see also attribute.
 */
displayAttrData *
getDisplayAttrDataByLabelWidget(w)
Widget	w;
{
    displayData		*dd;
    displayAttrData	*dad;

    dd = getDisplayDataByWidget(w);
    if (dd == NULL) {
	return NULL;
    }
    for (dad = dd->di_displayAttrData; dad != NULL; dad = dad->dad_next) {
	if (dad->dad_labelWidget == w) {
	    return dad;
	}
    }
    return NULL;
}
    



/*
 * Return True if the string looks like a DN, False otherwise.
 */
Boolean
is_dn(str)
char	*str;
{
    return(strchr(str, '=') && strchr(str, ','));
}





/*
 * Return True if the string looks like a UFN, False otherwise.
 *
 * Since both DNs and UFNs contain commas, always test for DN first.
 */
Boolean
is_ufn(str)
char	*str;
{
    return(strchr(str, ',') != NULL);
}



void
initTranslations()
{
    static char	dragTranslations[] =
	"#override <Btn2Down>: displayStartDrag()";
    static XtActionsRec dragActions[] =
	{ {"displayStartDrag", (XtActionProc) displayStartDrag} };

    static char	noDragTranslations[] =
	"#override <Btn2Down>: doNothing()";
    static XtActionsRec noDragActions[] =
	{ {"doNothing", (XtActionProc) doNothing} };

    parsedDragTranslations = XtParseTranslationTable(dragTranslations);
    XtAppAddActions(app, dragActions, XtNumber(dragActions));
    parsedNoDragTranslations = XtParseTranslationTable(noDragTranslations);
    XtAppAddActions(app, noDragActions, XtNumber(noDragActions));
}



/*
 * Given a labeled URL, find the start of the label and url, and return
 * pointers to them.  Alters the given string.
 */
int
split_labeled_url(lurl, label, url)
char	*lurl;
char	**label;
char	**url;
{
    /* Break apart the URL label and URL */
    *url = strchr(lurl, '$');
    if (*url == NULL) {
	/* Either new format URL, or malformed */
	*url = *label = lurl;
	while (**label && !(isspace(**label))) {
	    (*label)++;
	}
	if (!(**label)) {
	    /* malformed, just use URL as label */
	    *label = *url;
	    return(-1);
	} else {
	    **label = '\0';	/* null-terminate URL */
	    (*label)++;
	    while (**label && isspace(**label)) {
		(*label)++;
	    }
	    return(0);
	}
    } else {
	/* old format URL */
	*label = lurl;
	**url = '\0';
	(*url)++;
	/* find start of url */
	while (**url && isspace(**url)) {
	    (*url)++;
	}
	return(0);
    }
}


int
init_ldap(p)
LDAP	**p;
{
    char 		*ldap_host = appConfig.ldapServer;
    struct servent	*sv;
    int			ldap_port;
    int			rc;

    if ((ldap_port = appConfig.ldapPort) == 0) {
	if ((sv = getservbyname("tcp", "ldap")) != NULL) {
	    ldap_port = sv->s_port;
	} else {
	    /* fallback */
	    ldap_port = LDAP_PORT;
	}
    }

    if ((*p = ldap_open(ldap_host, ldap_port)) == NULL) {
	return(LDAP_SERVER_DOWN);
    }

#define LDC NULL
#ifdef notdef
    rc = ldap_simple_bind_s(*p, appConfig.ldapBindDN, LDC);
#endif
    rc = start_async_bind(NULL);

    if (appConfig.ldapCacheSize) {
	ldap_enable_cache(*p, appConfig.ldapCacheTimeout * 60,
		appConfig.ldapCacheSize * 1024);
    }
    return(rc);
}



/*
 * Take a look at the LDAP return code.  If it looks like something
 * we can fix by re-connecting to the LDAP server, do so.  Otherwise,
 * just return.
 */
int
reconnect_ldap(rc)
int	rc;
{
    switch (rc) {
    case LDAP_SERVER_DOWN:
	if (init_ldap(&ldp)) {
	    errorDlg(toplevel, "LDAP server failure: ldap_init failed");
	    return(-1);
	}
	bound_as[0] = '\0';
	sticky_bind = False;
	updateBoundAsText(bound_as);
	return(0);
	break;
    default:
	return(0);
    }
}


/*
 * Functions used for testing if a given attrData struct is of a given
 * type.
 */
Boolean
isDNAttrData(dad)
displayAttrData	*dad;
{
    return(dad->dad_ti->ti_syntaxid == LDAP_SYN_DN);
}


Boolean
isURLAttrData(dad)
displayAttrData	*dad;
{
    return(dad->dad_ti->ti_syntaxid == LDAP_SYN_LABELEDURL);
}


Boolean
isAudioAttrData(dad)
displayAttrData	*dad;
{
    return(dad->dad_ti->ti_syntaxid == LDAP_SYN_AUDIOBUTTON);
}


Boolean
isJPEGButtonAttrData(dad)
displayAttrData	*dad;
{
    return(dad->dad_ti->ti_syntaxid == LDAP_SYN_JPEGBUTTON);
}


Boolean
isJPEGImageAttrData(dad)
displayAttrData	*dad;
{
    return(dad->dad_ti->ti_syntaxid == LDAP_SYN_JPEGIMAGE);
}

/*
 * Routine used as a dragAction when weneed to disable a drag source.
 */
void
doNothing()
{
}

/*
 * Disallow this widget from being a drag source.
 */
void
disableDragSource(w)
Widget	w;
{
    XtVaSetValues(w, XmNtranslations, parsedNoDragTranslations, NULL);
}


/*
 * Return True if "ad" points to a "member" attribute and "value" is
 * the same as the DN of the entry.  Used to disallow people from adding
 * groups as members of themselves.
 */
static int
X500MemberIsSameAsDn(dad, value)
displayAttrData	*dad;
char		*value;
{
    if (strcasecmp(dad->dad_ti->ti_attrname, "member")) {
	return 0;
    } else {
	return(!strcasecmp(dad->dad_entryData->en_dn, value));
    }
}


/*
 * Check to see if it's OK to add the DN specified by "value" to the
 * attribute spcified by "ad".  Has intimate knowledge of contents
 * of templates config file, which is probably bad.
 */
int
okToAddValuetoAttribute(dd, dad, value)
displayData	*dd;
displayAttrData	*dad;
char		*value;
{
    /*
     * Don't allow a group's own DN to be put into its "Errors to",
     * "Requests to", or "X.500 Members" attributes
     */
    if (!strcasecmp(dd->di_tmpl->dt_name, "group")) {
	if (!strcasecmp(dad->dad_ti->ti_attrname, "member") ||
		!strcasecmp(dad->dad_ti->ti_attrname, "errorsTo") ||
		!strcasecmp(dad->dad_ti->ti_attrname, "requestsTo")) {
	    if (!strcasecmp(dad->dad_entryData->en_dn, value)) {
		return 0;
	    }
	}
    }
    return 1;
}



/*
 * Clear the text of the given label
 */
static void
clearLabel(w, id)
Widget		w;
XtIntervalId	*id;
{
    XmString	str;

    str = XmStringCreateLocalized("");
    XtVaSetValues(w, XmNlabelString, str, NULL);
    XmStringFree(str);
}


/*
 * Set the labelString of the given widget to "msg" for "time" seconds.
 */
void
displayLabelTimed(w, msg, time)
Widget	w;
char	*msg;
int	time;
{
    XmString	str;

    if (!XmIsLabel(w) && !XmIsLabelGadget(w)) {
	return;
    }
    str = XmStringCreateLocalized(msg);
    XtVaSetValues(w, XmNlabelString, str, NULL);
    XmStringFree(str);
    XtAppAddTimeOut(app, time * 1000, clearLabel, w);
}


/*
 * Convert magic placeholder strings: "-dnt" is converted to the value of
 * "entryDN", "-dnb" is converted to the DN we're bound as.  If a string
 * other that "-dnt" or "-dnb" is given, NULL is returned.  The caller
 * is responsible for freeing the storage allocated here.
 */
char *
convertDNEscape(str, entryDN)
char	*str;
char	*entryDN;
{
    if (!strcmp(str, "-dnt")) {
	return(strdup(entryDN));
    }
    if (!strcmp(str, "-dnb")) {
	return(strdup(bound_as));
    }
    return(NULL);
}


/*
 * Convert a message with embedded escape sequences of the form
 * "%-dnt" and "%-dnb".  These are replaced with, respectively, the
 * leftmost component of "entryDN" and the leftmost component 
 * of the DN we're bound as.  Used when handling an adddn action
 * button - for messages like "Ok to add woof to group foo?"
 * XXX can overflow buf.
 */
char *
convertMessageDNEscape(str, entryDN)
char	*str;
char	*entryDN;
{
    char	buf[1024];
    char	*s, *d, *p;
    char	**xdn;
    int		is_esc;

    s = str;
    d = buf;
    while ((s != NULL) && (*s != '\0')) {
	if (*s != '%') {
	    *d++ = *s++;
	} else {
	    s++;
	    is_esc = 0;
	    if (!strncmp(s, "-dnt", 4)) {
		xdn = ldap_explode_dn(entryDN, 1);
		is_esc = 1;
	    } else if (!strncmp(s, "-dnb", 4)) {
		xdn = ldap_explode_dn(bound_as, 1);
		is_esc = 1;
	    } else {
		/* not an escape sequence we recognize.  leave it alone */
		*d++ = '%';
	    }
	    if (is_esc) {
		p = xdn[0];
		while (*p) {
		    *d++ = *p++;
		}
		s += 4;
	    } else {
		*d++ = *s++;
	    }
	}
    }
    *d = '\0';
    return(strdup(buf));
}



attrData *
getAttrDataByName(ad, attr)
attrData	*ad;
char		*attr;
{
    attrData	*p;

    if (attr == NULL) {
	return(NULL);
    }

    for (p = ad; p != NULL; p = p->ad_next) {
	if ((p->ad_attrname != NULL) && !strcmp(p->ad_attrname, attr)) {
	    return(p);
	}
    }

    return(NULL);
}


/*
 * Given an ldap message (a result from a search) produce an entryData
 * structure, with a linked list of attrData nodes.
 */
entryData *
ldapMessage2entryData(ld, e, dt)
LDAP			*ld;
LDAPMessage		*e;
struct ldap_disptmpl	*dt;
{
    entryData			*entry;
    attrData			**ad;
    char			**xdn;
    int				row, col;
    struct ldap_tmplitem	*ti, *ti2;

    if ((entry = (entryData *) malloc(sizeof(entryData))) == NULL) {
	fprintf(stderr, "out of memory, malloc entryData\n");
	exit(1);
    }
    entry->en_dn = ldap_get_dn(ldp, e);
    xdn = ldap_explode_dn(entry->en_dn, 1);
    entry->en_rdn = strdup(xdn[0]);
    ldap_value_free(xdn);
    entry->en_oclist = ldap_get_values(ld, e, "objectClass");
    entry->en_isnew = False;
    entry->en_modified = False;
    entry->en_refcnt = 1;
#ifdef KERBEROS
    entry->en_krbNames = ldap_get_values(ld, e, "krbName");
#endif
    ad = &(entry->en_attrData);
    
    /*
     * Construct a linked list of attrData structs and hang it off the
     * attrData field in the entry struct.
     *
     * For each template item, retrieve all the values in the
     * LDAPMessage.
     */
    for (row = 0, ti = ldap_first_tmplrow(dt); ti != NULLTMPLITEM;
	    row++, ti = ldap_next_tmplrow(dt, ti)) {
	for (col = 0, ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULLTMPLITEM;
		col++, ti2 = ldap_next_tmplcol(dt, ti, ti2)) {
	    *ad = (attrData *) malloc(sizeof(attrData));
	    (*ad)->ad_attrname = ti2->ti_attrname;
	    (*ad)->ad_flags = 0;
	    if (LDAP_GET_SYN_TYPE(ti2->ti_syntaxid) == LDAP_SYN_TYPE_ACTION) {
		/* Action types have no values, so don't retrieve them */
		(*ad)->ad_vals = NULL;
		(*ad)->ad_nvals = 0;
	    } else {
		(*ad)->ad_vals = ldap_get_values_len(ld, e, ti2->ti_attrname);
		(*ad)->ad_nvals = ldap_count_values_len((*ad)->ad_vals);
	    }
	    (*ad)->ad_syndata = LDAP_GET_TMPLITEM_APPDATA(ti2, syntaxdata *);
	    ad = &((*ad)->ad_next);
	}
    }
    *ad = NULL;
    return(entry);
}

/*
 * Given a display template, produce an entryData structure, with a
 * linked list of attrData nodes filled in as specified by the template
 * instructions.
 */
entryData *
ldapTemplate2entryData(dt, addersdn)
struct ldap_disptmpl	*dt;
char			*addersdn;
{
    entryData			*entry;
    attrData			**ad;
    int				row, col;
    struct ldap_tmplitem	*ti, *ti2;
    struct ldap_adddeflist	*addef;

    if ((entry = (entryData *) malloc(sizeof(entryData))) == NULL) {
	fprintf(stderr, "out of memory, malloc entryData\n");
	exit(1);
    }
    entry->en_dn = NULL;
    entry->en_rdn = NULL;
    entry->en_oclist = NULL;
    entry->en_isnew = True;
    entry->en_modified = True;
    entry->en_refcnt = 1;
#ifdef KERBEROS
    entry->en_krbNames = NULL;
#endif
    ad = &(entry->en_attrData);
    
    /*
     * Construct a linked list of attrData structs and hang it off the
     * attrData field in the entry struct.
     */
    for (row = 0, ti = ldap_first_tmplrow(dt); ti != NULLTMPLITEM;
	    row++, ti = ldap_next_tmplrow(dt, ti)) {
	for (col = 0, ti2 = ldap_first_tmplcol(dt, ti); ti2 != NULLTMPLITEM;
		col++, ti2 = ldap_next_tmplcol(dt, ti, ti2)) {
	    *ad = (attrData *) malloc(sizeof(attrData));
	    (*ad)->ad_flags = 0;
	    (*ad)->ad_attrname = ti2->ti_attrname;

	    /*
	     * Do we need to fill in a value for this attribute?
	     */
	    if ((addef = attr_has_default_value(dt, ti2->ti_attrname))
		    != NULL) {
		if (addef->ad_source == LDAP_ADSRC_CONSTANTVALUE) {
		    (*ad)->ad_vals = make_value_len(addef->ad_value);
		    (*ad)->ad_flags = AD_DIRTY;
		} else if (addef->ad_source == LDAP_ADSRC_ADDERSDN) {
		    (*ad)->ad_vals = make_value_len(addersdn);
		    (*ad)->ad_flags = AD_DIRTY;
		} else {
		    (*ad)->ad_vals = NULL;
		}
		/* Action types have no values, so don't retrieve them */
		(*ad)->ad_nvals = 0;
	    } else {
		(*ad)->ad_vals = NULL;
		(*ad)->ad_nvals = 0;
	    }
	    (*ad)->ad_syndata = LDAP_GET_TMPLITEM_APPDATA(ti2, syntaxdata *);
	    ad = &((*ad)->ad_next);
	}
    }
    *ad = NULL;
    return(entry);
}


static struct ldap_adddeflist *
attr_has_default_value(dt, attrname)
struct ldap_disptmpl	*dt;
char *attrname;
{
    struct ldap_adddeflist	*addef;

    if ((dt == NULL) || (attrname == NULL) || (attrname[0] == '\0')) {
	return NULL;
    }

    for (addef = dt->dt_adddeflist; addef != NULL; addef = addef->ad_next) {
	if (!strcasecmp(addef->ad_attrname, attrname)) {
	    return(addef);
	}
    }
    return NULL;
}



/*
 * Create a single-valued berval with value "value"
 */
struct berval **
make_value_len(value)
char *value;
{
    struct berval **bv;

    if (value == NULL) {
	return NULL;
    }

    /* XXX ch_malloc */
    bv = (struct berval **) malloc(2 * sizeof(struct berval *));
    bv[0] = (struct berval *) malloc(sizeof(struct berval));
    bv[1] = NULL;
    bv[0]->bv_val = strdup(value);
    bv[0]->bv_len = strlen(value);

    return(bv);
}


/*
 * Just like malloc, except we check the returned value and exit
 * if anything goes wrong.
 */
char *
ch_malloc(size)
unsigned long	size;
{
	char	*new;

	if ( (new = (char *) malloc( size )) == NULL ) {
		fprintf( stderr, "malloc of %d bytes failed\n", size );
		exit( 1 );
	}

	return( new );
}

/*
 * Just like realloc, except we check the returned value and exit
 * if anything goes wrong.
 */
char *
ch_realloc(block, size)
char		*block;
unsigned long	size;
{
	char	*new;

	if ( block == NULL ) {
		return( ch_malloc( size ) );
	}

	if ( (new = (char *) realloc( block, size )) == NULL ) {
		fprintf( stderr, "realloc of %d bytes failed\n", size );
		exit( 1 );
	}

	return( new );
}

/*
 * Just like calloc, except we check the returned value and exit
 * if anything goes wrong.
 */
char *
ch_calloc(nelem, size)
unsigned long	nelem;
unsigned long	size;
{
	char	*new;

	if ( (new = (char *) calloc( nelem, size )) == NULL ) {
		fprintf( stderr, "calloc of %d elems of %d bytes failed\n",
		    nelem, size );
		exit( 1 );
	}

	return( new );
}


void
ch_free(p)
char *p;
{
    if (p != NULL) {
	free(p);
    }
}
